@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2016-2022 Intel Corporation
+* Copyright 2016-2023 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -160,7 +160,7 @@ inline typename remove_reference<T>::type zero() {
 }
 
 template <typename T, typename... Args>
-std::unique_ptr<T> make_unique(Args &&... args) {
+std::unique_ptr<T> make_unique(Args &&...args) {
     return std::unique_ptr<T>(new T(std::forward<Args>(args)...));
 }
 
@@ -241,6 +241,11 @@ inline R array_product(const T *arr, size_t size) {
     return prod;
 }
 
+template <typename T, typename R = T>
+inline R array_product(const std::vector<T> &v) {
+    return array_product<T, R>(v.data(), v.size());
+}
+
 template <typename T, typename R = T>
 inline R array_min(const T *arr, size_t size) {
     R min = std::numeric_limits<R>::max();
@@ -362,7 +367,7 @@ inline T nd_iterator_init(T start) {
     return start;
 }
 template <typename T, typename U, typename W, typename... Args>
-inline T nd_iterator_init(T start, U &x, const W &X, Args &&... tuple) {
+inline T nd_iterator_init(T start, U &x, const W &X, Args &&...tuple) {
     start = nd_iterator_init(start, utils::forward<Args>(tuple)...);
     x = start % X;
     return start / X;
@@ -372,7 +377,7 @@ inline bool nd_iterator_step() {
     return true;
 }
 template <typename U, typename W, typename... Args>
-inline bool nd_iterator_step(U &x, const W &X, Args &&... tuple) {
+inline bool nd_iterator_step(U &x, const W &X, Args &&...tuple) {
     if (nd_iterator_step(utils::forward<Args>(tuple)...)) {
         if (++x - X == 0) {
             x = 0;
@@ -398,7 +403,7 @@ inline bool nd_iterator_jump(U &cur, const U end, W &x, const Y &X) {
 }
 template <typename U, typename W, typename Y, typename... Args>
 inline bool nd_iterator_jump(
-        U &cur, const U end, W &x, const Y &X, Args &&... tuple) {
+        U &cur, const U end, W &x, const Y &X, Args &&...tuple) {
     if (nd_iterator_jump(cur, end, utils::forward<Args>(tuple)...)) {
         if (++x - X == 0) {
             x = 0;
@@ -413,7 +418,7 @@ constexpr T pick(size_t i, const T &x0) {
     return x0;
 }
 template <typename T, typename... Args>
-constexpr T pick(size_t i, const T &x0, Args &&... args) {
+constexpr T pick(size_t i, const T &x0, Args &&...args) {
     return i == 0 ? x0 : pick(i - 1, utils::forward<Args>(args)...);
 }
 
@@ -505,7 +510,7 @@ std::string format_impl(const char *fmt, Args... args) {
 }
 
 template <typename... Args>
-std::string format(const char *fmt, Args &&... args) {
+std::string format(const char *fmt, Args &&...args) {
     return format_impl(fmt, format_cvt_impl(std::forward<Args>(args))...);
 }
 
@@ -722,6 +727,32 @@ inline bool is_native_runtime(runtime_kind_t kind) {
             runtime_kind::tbb, runtime_kind::threadpool);
 }
 
+// Convenience wrapper to choose at compile-time between std::unique_ptr's
+// default deleter and a no-op one.
+//
+// This is useful for static pointers to objects with non-trivial destructors.
+// In some environments (e.g. tests where not all threads are joined at exit
+// time) these destructors can result in sanitizer failures (e.g. races in
+// thread sanitizer) when destructing unique_ptr's, but not with raw pointers.
+// Of course in a shared library environment using raw pointers (that are
+// therefore never freed) would result in memory leaks; this is why
+// DNNL_MAYBE_UNIQUE_PTR_IS_UNIQUE defaults to 1.
+#ifndef DNNL_MAYBE_UNIQUE_PTR_IS_UNIQUE
+#define DNNL_MAYBE_UNIQUE_PTR_IS_UNIQUE 1
+#endif
+
+#if DNNL_MAYBE_UNIQUE_PTR_IS_UNIQUE
+template <typename T>
+using maybe_unique_ptr = std::unique_ptr<T>;
+#else
+struct nop_deleter_t {
+    template <typename T>
+    void operator()(T const &) const noexcept {}
+};
+template <typename T>
+using maybe_unique_ptr = std::unique_ptr<T, nop_deleter_t>;
+#endif // DNNL_MAYBE_UNIQUE_PTR_IS_UNIQUE
+
 } // namespace impl
 } // namespace dnnl
 
