@@ -102,6 +102,8 @@ public:
         _op_mxcsr = 4u,
     };
 
+    const uint64_t cpu_sveLen = get_sve_length();
+
     const Xbyak_aarch64::WReg W_TMP_0 = w23;
     const Xbyak_aarch64::WReg W_TMP_1 = w24;
     const Xbyak_aarch64::WReg W_TMP_2 = w25;
@@ -215,21 +217,99 @@ public:
         L(label);
     }
 
+    template <typename T>
+    Xbyak_aarch64::XReg addr_off(const Xbyak_aarch64::XReg &base, const T off,
+            const Xbyak_aarch64::XReg &addr, const Xbyak_aarch64::XReg &x_tmp) {
+        if (off == 0) return base;
+
+        add_imm(addr, base, off, x_tmp);
+        return addr;
+    }
+
+    template <typename PRegBHSD, typename T>
+    void set_preg(const PRegBHSD &p, T tail_size,
+            const Xbyak_aarch64::XReg x_tmp0 = Xbyak_aarch64::XReg(DUMMY_IDX),
+            const Xbyak_aarch64::XReg x_tmp1 = Xbyak_aarch64::XReg(DUMMY_IDX)) {
+        using namespace Xbyak_aarch64;
+
+        assert(tail_size <= 64); // Implemented only for "SVE size <=  512"
+
+        switch (tail_size) {
+            case 0: pfalse(PRegB(p.getIdx())); return;
+            case 1: ptrue(p, VL1); return;
+            case 2: ptrue(p, VL2); return;
+            case 3: ptrue(p, VL3); return;
+            case 4: ptrue(p, VL4); return;
+            case 5: ptrue(p, VL5); return;
+            case 6: ptrue(p, VL6); return;
+            case 7: ptrue(p, VL7); return;
+            case 8: ptrue(p, VL8); return;
+            case 16: ptrue(p, VL16); return;
+            case 32: ptrue(p, VL32); return;
+            case 64: ptrue(p, VL64); return;
+        }
+
+        assert(x_tmp0.getIdx() != DUMMY_IDX && x_tmp1.getIdx() != DUMMY_IDX);
+        mov_imm(x_tmp0, 0);
+        mov_imm(x_tmp1, tail_size);
+        whilelt(p, x_tmp0, x_tmp1);
+    }
+
+    template <typename T>
+    void uni_add(const T &x1, const T &x2, const T &op) {
+        add(x1, x2, op);
+    }
+
+    void uni_add(const Xbyak_aarch64::VReg4S &x1,
+            const Xbyak_aarch64::VReg4S &x2, const Xbyak_aarch64::VReg4S &op) {
+        add(x1, x2, op);
+    }
+
+    void uni_add(const Xbyak_aarch64::ZReg &x1, const Xbyak_aarch64::ZReg &x2,
+            const Xbyak_aarch64::ZReg &op) {
+        add(Xbyak_aarch64::ZRegS(x1.getIdx()),
+                Xbyak_aarch64::ZRegS(x2.getIdx()),
+                Xbyak_aarch64::ZRegS(op.getIdx()));
+    }
+
+    template <typename T>
+    void udiv_mod(const T &q, const T &r, const T &divend, const T &divisor) {
+        assert(q.getIdx() != divisor.getIdx());
+        assert(q.getIdx() != divend.getIdx());
+        assert(r.getIdx() != divend.getIdx());
+
+        udiv(q, divend, divisor);
+        mul(r, q, divisor);
+        sub(r, divend, r);
+    }
+
+    template <typename T>
+    void umod(const T &r, const T &divend, const T &divisor) {
+        assert(r.getIdx() != divend.getIdx());
+        assert(r.getIdx() != divisor.getIdx());
+
+        udiv(r, divend, divisor);
+        mul(r, r, divisor);
+        sub(r, divend, r);
+    }
+
     void uni_clear(const Xbyak_aarch64::VReg &dst) { eor(dst.b, dst.b, dst.b); }
 
     void uni_clear(const Xbyak_aarch64::ZReg &dst) { eor(dst.d, dst.d, dst.d); }
 
-    template <typename TReg>
-    void uni_fdiv(const TReg &dst, const TReg &src, const TReg &src2) {
-        fdiv(dst, src, src2);
+    template <typename T>
+    void uni_fadd(const T &dst, const T &src, const T &src2) {
+        fadd(dst, src, src2);
     }
 
-    void uni_fdiv(const Xbyak_aarch64::VReg4S &dst,
-            const Xbyak_aarch64::VReg4S &src, const Xbyak_aarch64::VReg4S &src2,
-            const Xbyak_aarch64::VReg4S &tmp, const Xbyak_aarch64::PReg &pred) {
-        UNUSED(tmp);
-        UNUSED(pred);
-        fdiv(dst, src, src2);
+    void uni_fcvtzs(
+            const Xbyak_aarch64::VReg4S &d, const Xbyak_aarch64::VReg4S &s) {
+        fcvtzs(d, s);
+    }
+
+    void uni_fcvtzs(
+            const Xbyak_aarch64::ZRegS &d, const Xbyak_aarch64::ZRegS &s) {
+        fcvtzs(d, P_ALL_ONE / Xbyak_aarch64::T_z, s);
     }
 
     template <typename TReg>
@@ -254,11 +334,62 @@ public:
         }
     }
 
+    template <typename TReg>
+    void uni_fdiv(const TReg &dst, const TReg &src, const TReg &src2) {
+        fdiv(dst, src, src2);
+    }
+
+    void uni_fdiv(const Xbyak_aarch64::VReg4S &dst,
+            const Xbyak_aarch64::VReg4S &src, const Xbyak_aarch64::VReg4S &src2,
+            const Xbyak_aarch64::VReg4S &tmp, const Xbyak_aarch64::PReg &pred) {
+        UNUSED(tmp);
+        UNUSED(pred);
+        fdiv(dst, src, src2);
+    }
+
+    template <typename T>
+    void uni_fmax(const T &dst, const T &src, const T &src2) {
+        uint32_t dstIdx = dst.getIdx();
+        uint32_t srcIdx = src.getIdx();
+        if (dstIdx != srcIdx)
+            mov(Xbyak_aarch64::ZRegD(dstIdx), Xbyak_aarch64::ZRegD(srcIdx));
+        fmax(dst, P_ALL_ONE / Xbyak_aarch64::T_m, src2);
+    }
+
+    template <typename T>
+    void uni_fmin(const T &dst, const T &src, const T &src2) {
+        uint32_t dstIdx = dst.getIdx();
+        uint32_t srcIdx = src.getIdx();
+        if (dstIdx != srcIdx)
+            mov(Xbyak_aarch64::ZRegD(dstIdx), Xbyak_aarch64::ZRegD(srcIdx));
+        fmin(dst, P_ALL_ONE / Xbyak_aarch64::T_m, src2);
+    }
+
+    template <typename T>
+    void uni_fmul(const T &dst, const T &src, const T &src2) {
+        fmul(dst, src, src2);
+    }
+
+    void uni_frinti(
+            const Xbyak_aarch64::VReg4S &d, const Xbyak_aarch64::VReg4S &s) {
+        frinti(d, s);
+    }
+
+    void uni_frinti(
+            const Xbyak_aarch64::ZRegS &d, const Xbyak_aarch64::ZRegS &s) {
+        frinti(d, P_ALL_ONE / Xbyak_aarch64::T_m, s);
+    }
+
     void uni_fsub(const Xbyak_aarch64::VReg4S &v1,
             const Xbyak_aarch64::VReg4S &v2, const Xbyak_aarch64::VReg4S &v3) {
         fsub(v1, v2, v3);
     }
 
+    template <typename T>
+    void uni_fsub(const T &dst, const T &src, const T &src2) {
+        fsub(dst, src, src2);
+    }
+
     void uni_fsub(const Xbyak_aarch64::ZRegS &z1,
             const Xbyak_aarch64::ZRegS &z2, const Xbyak_aarch64::ZRegS &z3) {
         fsub(z1, z2, z3);
@@ -288,6 +419,56 @@ public:
         ldr(dst, ptr(addr));
     }
 
+    template <typename T>
+    void uni_ldr(const Xbyak_aarch64::ZReg &r, const Xbyak_aarch64::XReg &base,
+            const T off) {
+        const int off_mod = off % cpu_sveLen;
+        const int off_mul_vl = off / cpu_sveLen;
+
+        if (off_mod == 0 && -256 <= off_mul_vl && off_mul_vl <= 255) {
+            ldr(r, Xbyak_aarch64::ptr(base, off_mul_vl, Xbyak_aarch64::MUL_VL));
+        } else {
+            const int offset = off_mod * 0x10 * (cpu_sveLen / 16);
+            add_imm(X_DEFAULT_ADDR, base, offset, X_TMP_0);
+            ldr(r, Xbyak_aarch64::ptr(X_DEFAULT_ADDR));
+        }
+    }
+
+    template <typename T>
+    void uni_ldr(const Xbyak_aarch64::VReg &r, const Xbyak_aarch64::XReg &base,
+            const T off) {
+        const int off_mod = off % 16;
+        const int off_mul_vl = off / 16;
+
+        if (off_mod == 0 && 0 <= off_mul_vl && off_mul_vl <= 65520) {
+            ldr(Xbyak_aarch64::QReg(r.getIdx()), Xbyak_aarch64::ptr(base, off));
+        } else {
+            add_imm(X_DEFAULT_ADDR, base, off_mod * 4, X_TMP_0);
+            ldr(Xbyak_aarch64::QReg(r.getIdx()),
+                    Xbyak_aarch64::ptr(X_DEFAULT_ADDR));
+        }
+    }
+
+    void uni_orr(const Xbyak_aarch64::VReg &d, const Xbyak_aarch64::VReg &s0,
+            const Xbyak_aarch64::VReg &s1) {
+        orr(d.b16, s0.b16, s1.b16);
+    }
+
+    void uni_orr(const Xbyak_aarch64::ZReg &d, const Xbyak_aarch64::ZReg &s0,
+            const Xbyak_aarch64::ZReg &s1) {
+        orr(d.d, s0.d, s1.d);
+    }
+
+    void uni_scvtf(
+            const Xbyak_aarch64::VReg4S &t0, const Xbyak_aarch64::VReg4S &t1) {
+        scvtf(t0, t1);
+    }
+
+    void uni_scvtf(
+            const Xbyak_aarch64::ZRegS &t0, const Xbyak_aarch64::ZRegS &t1) {
+        scvtf(t0, P_ALL_ONE / Xbyak_aarch64::T_m, t1);
+    }
+
     void uni_str(
             const Xbyak_aarch64::VReg &src, const Xbyak_aarch64::XReg &addr) {
         str(Xbyak_aarch64::QReg(src.getIdx()), ptr(addr));
@@ -298,6 +479,11 @@ public:
         str(src, ptr(addr));
     }
 
+    template <typename T>
+    void uni_sub(const T &x1, const T &x2, const T &op) {
+        sub(x1, x2, op);
+    }
+
     /*
       Saturation facility functions. enable to prepare the register
       holding the saturation upperbound and apply the saturation on
@@ -376,6 +562,46 @@ public:
             assert(!"unreachable");
     }
 
+    /* A utility function to process f32 tail (load, store or other) depending
+     * on tail size, stored in Reg64. Tail size must be value from 0 to 3/7
+     * (Xmm/Ymm). Tail process functions require integer as argument to specify
+     * behavior for each tail size.
+     *
+     * Only supported for Xmm and Ymm.
+     */
+    template <cpu_isa_t isa>
+    void runtime_tail_process(const Xbyak_aarch64::XReg &reg_tail,
+            const Xbyak_aarch64::XReg &reg_tmp,
+            const std::function<void(int)> &tail_process) {
+        constexpr int simd_w_ymm = 8;
+        constexpr int f32_bits = sizeof(float) * 8;
+        const auto simd_w = cpu_isa_traits<isa>::vlen * 8 / f32_bits;
+        assert(simd_w != cpu_isa_traits<isa>::vlen * 8 / f32_bits);
+
+        Xbyak_aarch64::Label label_tbl, label_tbl_end;
+        Xbyak_aarch64::Label l_case[simd_w_ymm];
+
+        adr(reg_tmp, label_tbl);
+        mov_imm(X_TMP_0, sizeof(void *));
+        madd(X_DEFAULT_ADDR, reg_tail, X_TMP_0, reg_tmp);
+        br(X_DEFAULT_ADDR);
+
+        // create jump table
+        L(label_tbl);
+        for (size_t i = 0; i < simd_w; i++)
+            putL(l_case[i]);
+
+        // cases for each tail size - from 0 to 3/7
+        L(l_case[0]);
+        b(label_tbl_end);
+        for (size_t i = 1; i < simd_w; i++) {
+            L(l_case[i]);
+            tail_process(i);
+            b(label_tbl_end);
+        }
+        L(label_tbl_end);
+    }
+
     DNNL_DISALLOW_COPY_AND_ASSIGN(jit_generator);
 
 public:
