@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2021-2022 Intel Corporation
+* Copyright 2021-2023 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -21,6 +21,7 @@
 #include "common/utils.hpp"
 
 #include "cpu/platform.hpp"
+#include "cpu/x64/brgemm/brgemm.hpp"
 #include "cpu/x64/brgemm/brgemm_types.hpp"
 #include "cpu/x64/injectors/jit_uni_postops_injector.hpp"
 #include "cpu/x64/jit_generator.hpp"
@@ -38,7 +39,7 @@ using namespace Xbyak;
 
 struct jit_brgemm_amx_uker_base_t : public jit_generator {
     jit_brgemm_amx_uker_base_t(const brgemm_t &abrg)
-        : jit_generator(nullptr, MAX_CODE_SIZE, true, avx512_core)
+        : jit_generator(jit_name(), nullptr, MAX_CODE_SIZE, true, avx512_core)
         , brg(abrg)
         , postops_injector_(nullptr) {
 
@@ -61,7 +62,7 @@ struct jit_brgemm_amx_uker_base_t : public jit_generator {
                             broadcasting_strategy_t::no_broadcast};
             const binary_injector::rhs_arg_static_params_t rhs_sp {
                     static_cast<size_t>(Xbyak::Zmm(1).getIdx()), this->r14,
-                    this->r15, preserve_gpr, preserve_vmm,
+                    this->r15, this->r13, preserve_gpr, preserve_vmm,
                     GET_OFF(post_ops_binary_rhs_arg_vec), GET_OFF(data_C_ptr_),
                     dst_md_wrapper, static_cast<size_t>(brg.ldb_tail),
                     ld_tail_mask, use_exact_tail_scalar_bcast};
@@ -104,11 +105,18 @@ private:
             postops_injector_;
 
     using reg64_t = const Xbyak::Reg64;
+    enum {
+        simd_w = 16,
+        zmm_width_in_bytes = cpu_isa_traits<avx512_core>::vlen,
+        tile_size = 1024
+    };
+    enum matrix_kind_t { matrix_A, matrix_B };
 
     // Register decomposition
     const reg64_t param1 = abi_param1;
 
     const reg64_t reg_addr_batch = r13;
+    const reg64_t reg_aux1_batch = rbp;
     const reg64_t reg_aux_A = r11;
     const reg64_t reg_aux_B = r10;
     const reg64_t reg_stride_lda = r14;
@@ -117,6 +125,8 @@ private:
     const reg64_t reg_D = r12;
 
     const reg64_t reg_buf = r8;
+    const reg64_t reg_BS = rbx;
+    const reg64_t reg_BS_loop = r9;
     const reg64_t reg_bias = rbx;
     const reg64_t reg_scales = rbx;
 
@@ -130,6 +140,7 @@ private:
     const reg64_t reg_zp_comp_b = rbx;
     const reg64_t reg_zp_c_values = rbx;
     const reg64_t reg_ptr_sum_zp = r9;
+    const reg64_t reg_bf32_stride = rsi;
 
     constexpr static int abi_param1_offs_ = 0;
     constexpr static int reg_zp_comp_a_offs_ = 8;
@@ -139,6 +150,7 @@ private:
 
     bool are_post_ops_applicable_ = false;
     bool need_to_apply_alpha_beta_ = false;
+    bool may_load_accumulators_ = false;
 
     bool handle_binary_po_offset_ = false;
     bool with_binary_per_oc_bcast_ = false;
@@ -156,25 +168,30 @@ private:
     size_t *adj_bd_mask_buffer_ptr_ = nullptr;
     std::vector<size_t> skipped_bd_mask_buffer_;
     size_t *skipped_bd_mask_buffer_ptr_ = nullptr;
+    palette_config_t palette_;
+    // used to store offsets within wsp buffer where the data is
+    // transformed(downconverted), to reuse when needed.
+    std::unordered_map<std::string, size_t> transform_buf_map_A_;
+    std::unordered_map<std::string, size_t> transform_buf_map_B_;
 
-    size_t LDA_size_;
-    size_t LDB_size_;
-    size_t LDC_size_;
-    size_t LDD_size_;
-    size_t ld_block_B_size_;
-    size_t ld_block_C_size_;
-    size_t ld_block_D_size_;
-    size_t ld_block_bias_size_;
-    size_t ld_block_scales_size_;
-    size_t ld_block_zp_size_;
+    size_t LDA_size_ = 0, LDA2_size_ = 0;
+    size_t LDB_size_ = 0, LDB2_size_ = 0;
+    size_t LDC_size_ = 0, LDC2_size_M_ = 0, LDC2_size_N_ = 0;
+    size_t LDD_size_ = 0;
+    size_t ld_block_B_size_ = 0;
+    size_t ld_block_C_size_ = 0;
+    size_t ld_block_D_size_ = 0;
+    size_t ld_block_bias_size_ = 0;
+    size_t ld_block_scales_size_ = 0;
+    size_t ld_block_zp_size_ = 0;
 
-    size_t ldb_tail_B_size_;
-    size_t ldb_tail_C_size_;
-    size_t ldb_tail_D_size_;
-    size_t ldb_tail_zp_size_;
+    size_t ldb_tail_B_size_ = 0;
+    size_t ldb_tail_C_size_ = 0;
+    size_t ldb_tail_D_size_ = 0;
+    size_t ldb_tail_zp_size_ = 0;
 
     // variables of bd loop
-    size_t inp_bd_;
+    size_t inp_bd_ = 0;
 
     // interleave stores
     bool use_ils_ = false;
@@ -188,10 +205,11 @@ private:
         ils_bd_step_ = 0;
     int pfo_vec_ = 0, pfo_vecs_per_store_ = 0;
 
-    bool dt_requires_saturation_;
+    bool dt_requires_saturation_ = false;
 
     Xbyak::Opmask ld_full_mask = Xbyak::Opmask(2);
     Xbyak::Opmask ld_tail_mask = Xbyak::Opmask(3);
+    Xbyak::Opmask bf32_col_mask = Xbyak::Opmask(4);
 
     int adjusted_bd_block(int bdb) const noexcept {
         return (brg.is_M_tail && bdb == brg.bd_block2 - 1) ? brg.bdb_tail
@@ -217,6 +235,7 @@ private:
         // zmm9, zmm10, zmm11
         return Xbyak::Zmm(9 + ldb);
     }
+    const Xbyak::Zmm zmm_bf32_pemute = zmm12;
     const Xbyak::Zmm zmm_zp_comp_a = zmm12;
     const Xbyak::Zmm zmm_zp_c = zmm13;
     const Xbyak::Zmm zmm_lbound = zmm14;
@@ -232,7 +251,8 @@ private:
             Xbyak::Opmask ktail_mask);
 
     void read_params();
-    void load_accumulators(int bd_block2, int ld_block, bool is_ld_tail);
+    void load_accumulators(
+            int bd_block2, int ld_block, bool is_ld_tail, int ldb_ind);
 
     void maybe_saturation(Xbyak::Zmm &zmm);
     void apply_alpha_beta_to_vector(
@@ -260,13 +280,31 @@ private:
     void prefetch_output_data(int ldb_ind, int bd_block2, int ld_block2);
     void interleave_store(bool store_all);
 
-    void store_accumulators(int bd_block2, int ld_block, int l_step,
-            bool is_ld_tail, size_t c_offset, size_t d_offset, int ldb_ind,
-            bool apply_post_ops);
+    void store_accumulators(int bd_block2, int ld_block, bool is_ld_tail,
+            int ldb_ind, bool apply_post_ops);
 
     void set_A_B_matrices(int bs);
+    void set_A_B_matrices();
 
-    void gemm_microkernel_amx(int bd_block2, int ldb_ind, int ld_block2,
+    void bf32_downconvert(int num_rows, int tile_num_col_bytes,
+            reg64_t reg_data, int offset, reg64_t reg_data_stride,
+            reg64_t reg_buf, bool is_rd_tail);
+
+    void bf32_downconvert_to_vnni(int num_rows, int tile_num_col_bytes,
+            reg64_t reg_data, int offset, reg64_t reg_data_stride,
+            reg64_t reg_buf, bool is_rd_tail);
+
+    void maybe_pre_process_data(const Tmm &t1, reg64_t reg_base, size_t offset,
+            reg64_t reg_stride, int bs, int rbd, matrix_kind_t mk,
+            bool is_rd_tail);
+
+    void maybe_tileloadd_nt(matrix_kind_t mk, int bs, int xdb, int rdb,
+            size_t offset, bool is_rd_tail);
+
+    void tdpbxxd(int bdb_idx, int ldb_idx, int bd_block2, int ld_block2,
+            int ldb_ind);
+
+    void gemm_microkernel_amx(int bs, int bd_block2, int ldb_ind, int ld_block2,
             bool is_rd_tail, bool is_ld_tail);
 
     void ldb_loop(int bd_block2, int ld_block, int ldb_loop_length,
@@ -283,8 +321,11 @@ private:
 
     size_t B_offset(int ldb) const noexcept;
     size_t C_offset(int bd, int ldb) const noexcept;
+    size_t C_block_offset(int bd, int ldb) const noexcept;
     size_t D_offset(int bd, int ldb) const noexcept;
 
+    size_t lda() const noexcept;
+    size_t ldb() const noexcept;
     size_t rdb_A_offset() const noexcept;
     size_t rdb_B_offset() const noexcept;
 
@@ -333,22 +374,35 @@ int jit_brgemm_amx_uker_base_t::skipped_bd_mask(int inp_bd) noexcept {
         return skipped_bd_mask_buffer_ptr_[inp_bd];
 }
 
-size_t jit_brgemm_amx_uker_base_t::A_offset(int bd) const noexcept {
-    return bd * LDA_size_;
+size_t jit_brgemm_amx_uker_base_t::A_offset(int bdb) const noexcept {
+    return bdb * LDA2_size_;
 }
 
 size_t jit_brgemm_amx_uker_base_t::B_offset(int ldb) const noexcept {
-    return brg.rd_step * ldb * ld_block_B_size_;
+    return (brg.is_blocked ? 1 : brg.rd_step) * ldb * ld_block_B_size_;
 }
 
 size_t jit_brgemm_amx_uker_base_t::C_offset(int bd, int ldb) const noexcept {
     return bd * LDC_size_ + ldb * ld_block_C_size_;
 }
 
+size_t jit_brgemm_amx_uker_base_t::C_block_offset(
+        int bd, int ldb) const noexcept {
+    return (size_t)bd * LDC2_size_M_ + (size_t)ldb * LDC2_size_N_;
+}
+
 size_t jit_brgemm_amx_uker_base_t::D_offset(int bd, int ldb) const noexcept {
     return bd * LDD_size_ + ldb * ld_block_D_size_;
 }
 
+size_t jit_brgemm_amx_uker_base_t::lda() const noexcept {
+    return LDA_size_;
+}
+
+size_t jit_brgemm_amx_uker_base_t::ldb() const noexcept {
+    return LDB_size_ * brg.rd_step;
+}
+
 size_t jit_brgemm_amx_uker_base_t::rdb_A_offset() const noexcept {
     return brg.typesize_A * brg.rd_block;
 }
@@ -446,6 +500,8 @@ void jit_brgemm_amx_uker_base_t::read_params() {
 
     mov(reg_C, ptr[param1 + GET_OFF(ptr_C)]);
     mov(reg_D, ptr[param1 + GET_OFF(ptr_D)]);
+    mov(reg_BS, ptr[param1 + GET_OFF(BS)]);
+
     mov(reg_addr_batch, ptr[param1 + GET_OFF(batch)]);
 
     mov(reg_buf, ptr[param1 + GET_OFF(ptr_buf)]);
@@ -467,13 +523,26 @@ void jit_brgemm_amx_uker_base_t::read_params() {
 }
 
 void jit_brgemm_amx_uker_base_t::load_accumulators(
-        int bd_block2, int ld_block2, bool is_ld_tail) {
+        int bd_block2, int ld_block2, bool is_ld_tail, int ldb_ind) {
+    assert(IMPLICATION(is_ld_tail, ld_block2 == 1));
+    if (may_load_accumulators_) mov(reg_stride_ld_block, LDC_size_);
+
+    int bd_inp_bdb = inp_bd_;
+
     for (int bdb = 0; bdb < bd_block2; bdb++) {
-        if (is_ld_tail)
-            tilezero(Tmm(brg.get_C_tensor(bdb, brg.ld_block2)));
-        else
-            for (int ldb = 0; ldb < ld_block2; ldb++)
-                tilezero(Tmm(brg.get_C_tensor(bdb, ldb)));
+        for (int ldb = 0; ldb < ld_block2; ldb++) {
+            int idx = (is_ld_tail) ? brg.ld_block2 : ldb;
+            if (may_load_accumulators_) {
+                const auto bd_out_bdb = get_out_bd(bd_inp_bdb, 0);
+                const auto c_offset = C_block_offset(bd_out_bdb, ldb_ind + ldb);
+                tileloadd(Tmm(brg.get_C_tensor(bdb, idx)),
+                        ptr[reg_C + c_offset + reg_stride_ld_block]);
+            } else {
+                tilezero(Tmm(brg.get_C_tensor(bdb, idx)));
+            }
+        }
+        bd_inp_bdb += brg.bd_block;
+        bd_inp_bdb = skipped_bd_mask(bd_inp_bdb);
     }
 }
 
@@ -494,12 +563,12 @@ void jit_brgemm_amx_uker_base_t::apply_alpha_beta_to_vector(
 
     if (apply_beta && !use_vadd_for_beta) {
         mov(reg_tmp_gpr, float2int(static_cast<float>(brg.beta)));
-        movq(Xmm(zmm_beta.getIdx()), reg_tmp_gpr);
+        vmovq(Xmm(zmm_beta.getIdx()), reg_tmp_gpr);
         vbroadcastss(zmm_beta, Xmm(zmm_beta.getIdx()));
     }
     if (apply_alpha) {
         mov(reg_tmp_gpr, float2int(static_cast<float>(brg.alpha)));
-        movq(Xmm(zmm_alpha.getIdx()), reg_tmp_gpr);
+        vmovq(Xmm(zmm_alpha.getIdx()), reg_tmp_gpr);
         vbroadcastss(zmm_alpha, Xmm(zmm_alpha.getIdx()));
     }
     if (dq2ps_required) vcvtdq2ps(zmm, zmm);
@@ -688,7 +757,7 @@ void jit_brgemm_amx_uker_base_t::prepare_post_ops_registers(
         for (int ldb = 0; ldb < ld_block2; ldb++) {
             auto scales_ptr = EVEX_compress_addr(
                     reg_scales, scales_offset(ldb_ind + ldb));
-            vmovups(zmm_scales(ldb), scales_ptr);
+            vmovups(zmm_scales(ldb) | k_mask | T_z, scales_ptr);
         }
     }
 }
@@ -976,8 +1045,7 @@ void jit_brgemm_amx_uker_base_t::interleave_store(bool store_all) {
 }
 
 void jit_brgemm_amx_uker_base_t::store_accumulators(int bd_block2,
-        int ld_block2, int l_step, bool is_ld_tail, size_t c_offset,
-        size_t d_offset, int ldb_ind, bool apply_post_ops) {
+        int ld_block2, bool is_ld_tail, int ldb_ind, bool apply_post_ops) {
 
     const bool need_to_apply_post_ops
             = are_post_ops_applicable_ && apply_post_ops;
@@ -1038,7 +1106,7 @@ void jit_brgemm_amx_uker_base_t::store_accumulators(int bd_block2,
                 }
             } else {
                 const auto bd_out_bdb = get_out_bd(bd_inp_bdb, 0);
-                const auto c_offset = C_offset(bd_out_bdb, ldb_ind + ldb);
+                const auto c_offset = C_block_offset(bd_out_bdb, ldb_ind + ldb);
                 tilestored(ptr[reg_C + c_offset + reg_stride_ld_block],
                         Tmm(brg.get_C_tensor(bdb, idx)));
             }
@@ -1049,6 +1117,7 @@ void jit_brgemm_amx_uker_base_t::store_accumulators(int bd_block2,
 }
 
 void jit_brgemm_amx_uker_base_t::set_A_B_matrices(int bs) {
+    assert(brg.type == brgemm_addr);
     auto batch_offset = (size_t)bs * sizeof(brgemm_batch_element_t);
     if (brg.layout == brgemm_row_major) {
         mov(reg_aux_A,
@@ -1067,42 +1136,247 @@ void jit_brgemm_amx_uker_base_t::set_A_B_matrices(int bs) {
     }
 }
 
-void jit_brgemm_amx_uker_base_t::gemm_microkernel_amx(int bd_block2,
-        int ldb_ind, int ld_block2, bool is_rd_tail, bool is_ld_tail) {
-    auto tdpbxxd = [=](int bdb_idx, int ldb_idx) {
-        prefetch_output_data(ldb_ind, bd_block2, ld_block2);
-        const Tmm &x1 = Tmm(brg.get_C_tensor(bdb_idx, ldb_idx));
-        const Tmm &x2 = Tmm(brg.get_A_tensor(bdb_idx));
-        const Tmm &x3 = Tmm(brg.get_B_tensor(ldb_idx));
-        if (brg.dt_a == data_type::bf16 && brg.dt_b == data_type::bf16) {
-            tdpbf16ps(x1, x2, x3);
-        } else if (brg.dt_a == data_type::u8 && brg.dt_b == data_type::u8) {
-            tdpbuud(x1, x2, x3);
-        } else if (brg.dt_a == data_type::u8 && brg.dt_b == data_type::s8) {
-            tdpbusd(x1, x2, x3);
-        } else if (brg.dt_a == data_type::s8 && brg.dt_b == data_type::u8) {
-            tdpbsud(x1, x2, x3);
-        } else if (brg.dt_a == data_type::s8 && brg.dt_b == data_type::s8) {
-            tdpbssd(x1, x2, x3);
+void jit_brgemm_amx_uker_base_t::set_A_B_matrices() {
+    assert(brg.type == brgemm_addr);
+    assert(brg.brgattr.var_bs);
+
+    if (brg.layout == brgemm_row_major) {
+        mov(reg_aux_A, ptr[reg_aux1_batch + GET_OFF_BATCH_ELEMENT(ptr.A)]);
+        mov(reg_aux_B, ptr[reg_aux1_batch + GET_OFF_BATCH_ELEMENT(ptr.B)]);
+    } else {
+        mov(reg_aux_A, ptr[reg_aux1_batch + GET_OFF_BATCH_ELEMENT(ptr.B)]);
+        mov(reg_aux_B, ptr[reg_aux1_batch + GET_OFF_BATCH_ELEMENT(ptr.A)]);
+    }
+}
+
+void jit_brgemm_amx_uker_base_t::maybe_tileloadd_nt(matrix_kind_t mk, int bs,
+        int xdb, int rdb, size_t offset, bool is_rd_tail) {
+    const bool try_load_nt_A
+            = (brg.brgattr.hint_innermost_loop == brgemm_bd_loop_innermost);
+    const bool try_load_nt_B
+            = (brg.brgattr.hint_innermost_loop == brgemm_ld_loop_innermost);
+    const bool is_A = mk == matrix_kind_t::matrix_A;
+    bool try_load_nt = is_A ? try_load_nt_A : try_load_nt_B;
+    try_load_nt = try_load_nt
+            && (static_cast<size_t>(brg.typesize_A)
+                               * brg.brgattr.hint_expected_A_size
+                       + static_cast<size_t>(brg.typesize_B)
+                               * brg.brgattr.hint_expected_B_size
+                       + static_cast<size_t>(brg.typesize_C)
+                               * brg.brgattr.hint_expected_C_size)
+                    >= platform::get_per_core_cache_size(1);
+
+    auto t1 = Tmm(is_A ? brg.get_A_tensor(xdb) : brg.get_B_tensor(xdb));
+    auto reg_base = is_A ? reg_aux_A : reg_aux_B;
+    auto reg_stride = is_A ? reg_stride_lda : reg_stride_ldb;
+
+    if (brg.is_bf32)
+        // try_load_nt is not supported in maybe_pre_process_data as there is
+        // no guarantee that the data is cacheline aligned.
+        maybe_pre_process_data(
+                t1, reg_base, offset, reg_stride, bs, rdb, mk, is_rd_tail);
+    else if (try_load_nt)
+        tileloaddt1(t1, ptr[reg_base + offset + reg_stride]);
+    else
+        tileloadd(t1, ptr[reg_base + offset + reg_stride]);
+}
+
+void jit_brgemm_amx_uker_base_t::tdpbxxd(
+        int bdb_idx, int ldb_idx, int bd_block2, int ld_block2, int ldb_ind) {
+    prefetch_output_data(ldb_ind, bd_block2, ld_block2);
+    const Tmm &x1 = Tmm(brg.get_C_tensor(bdb_idx, ldb_idx));
+    const Tmm &x2 = Tmm(brg.get_A_tensor(bdb_idx));
+    const Tmm &x3 = Tmm(brg.get_B_tensor(ldb_idx));
+    if (brg.is_bf32
+            || (brg.dt_a == data_type::bf16 && brg.dt_b == data_type::bf16)) {
+        tdpbf16ps(x1, x2, x3);
+    } else if (brg.dt_a == data_type::u8 && brg.dt_b == data_type::u8) {
+        tdpbuud(x1, x2, x3);
+    } else if (brg.dt_a == data_type::u8 && brg.dt_b == data_type::s8) {
+        tdpbusd(x1, x2, x3);
+    } else if (brg.dt_a == data_type::s8 && brg.dt_b == data_type::u8) {
+        tdpbsud(x1, x2, x3);
+    } else if (brg.dt_a == data_type::s8 && brg.dt_b == data_type::s8) {
+        tdpbssd(x1, x2, x3);
+    } else {
+        assert(!"unsupported combination");
+    }
+    interleave_store(false);
+}
+
+// This method down-converts the data from f32 to bf16 and saves at reg_buf.
+// Generally used by matrix_A, where no vnni transformation of data is needed.
+void jit_brgemm_amx_uker_base_t::bf32_downconvert(int num_rows,
+        int tile_num_col_bytes, reg64_t reg_data, int offset,
+        reg64_t reg_data_stride, reg64_t reg_buf, bool is_rd_tail) {
+    const int rd_block = is_rd_tail ? brg.rdb_tail : brg.rd_block;
+    const int max_num_cols
+            = nstl::min<int>(tile_num_col_bytes / sizeof(bfloat16_t), rd_block);
+    const int col_tail = max_num_cols % simd_w;
+    auto zmm_1 = zmm_tmp_1();
+    auto zmm_2 = zmm_tmp_2();
+    auto zmm_2_masked = col_tail ? zmm_2 | bf32_col_mask | T_z : zmm_2;
+
+    assert(max_num_cols > 0);
+
+    if (col_tail) {
+        const int tail_mask = (1 << col_tail) - 1;
+        auto reg_tmp_32 = reg_tmp_gpr.cvt32();
+        mov(reg_tmp_32, tail_mask);
+        kmovw(bf32_col_mask, reg_tmp_32);
+    }
+
+    // Note: using the same register used in col_tail, so order is important
+    const auto reg_data_aux = reg_tmp_gpr;
+    lea(reg_data_aux, ptr[reg_data + offset]);
+
+    for (int r = 0; r < num_rows; ++r) {
+        if (max_num_cols > 16) {
+            vmovups(zmm_1, ptr[reg_data_aux]);
+            vmovups(zmm_2_masked, ptr[reg_data_aux + zmm_width_in_bytes]);
+            vcvtne2ps2bf16(zmm_1, zmm_2, zmm_1);
+            // we assume enough padding space is available.
+            vmovups(ptr[reg_buf + r * zmm_width_in_bytes], zmm_1);
         } else {
-            assert(!"unsupported combination");
+            auto ymm_1 = Ymm(zmm_1.getIdx());
+            auto ymm_1_masked
+                    = max_num_cols == 16 ? ymm_1 : ymm_1 | bf32_col_mask | T_z;
+            vcvtneps2bf16(ymm_1_masked, ptr[reg_data_aux]);
+            vmovups(ptr[reg_buf + r * zmm_width_in_bytes], ymm_1);
         }
-        interleave_store(false);
+        add(reg_data_aux, reg_data_stride);
+    }
+}
+
+// This method down-converts and transforms the data from f32 to bf16_vnni
+// format. Generally used by matrix_B.
+void jit_brgemm_amx_uker_base_t::bf32_downconvert_to_vnni(int num_rows,
+        int tile_num_col_bytes, reg64_t reg_data, int offset,
+        reg64_t reg_data_stride, reg64_t reg_buf, bool is_rd_tail) {
+    const int num_cols_ele = tile_num_col_bytes / sizeof(bfloat16_t);
+    const int num_N = num_cols_ele / sizeof(bfloat16_t);
+    const int col_tail = num_N % simd_w;
+    const auto zmm_1 = zmm_tmp_1();
+    const auto zmm_2 = zmm_tmp_2();
+
+    assert(num_N > 0);
+
+    auto load = [&](Zmm zmm, Address addr) {
+        if (col_tail)
+            vmovups(zmm | bf32_col_mask | T_z, addr);
+        else
+            vmovups(zmm, addr);
     };
 
-    auto maybe_tileloadd_nt = [=](const Tmm &t1, reg64_t base, size_t offset,
-                                      reg64_t stride, bool try_load_nt) {
-        if (try_load_nt
-                && static_cast<size_t>(
-                           brg.typesize_A * brg.brgattr.hint_expected_A_size
-                           + brg.typesize_B * brg.brgattr.hint_expected_B_size
-                           + brg.typesize_C * brg.brgattr.hint_expected_C_size)
-                        >= platform::get_per_core_cache_size(1))
-            tileloaddt1(t1, ptr[base + offset + stride]);
-        else
-            tileloadd(t1, ptr[base + offset + stride]);
+    if (col_tail) {
+        const int tail_mask = (1 << col_tail) - 1;
+        auto reg_tmp_32 = reg_tmp_gpr.cvt32();
+        mov(reg_tmp_32, tail_mask);
+        kmovw(bf32_col_mask, reg_tmp_32);
+    }
+
+    // Note: using the same register used in col_tail, so order is important
+    const auto reg_data_aux = reg_tmp_gpr;
+    lea(reg_data_aux, ptr[reg_data + offset]);
+
+    const int rd_block = is_rd_tail ? brg.rdb_tail : brg.rd_block;
+    const int vnni_granularity
+            = data_type_vnni_granularity(data_type_t::dnnl_bf16);
+    const int r_end
+            = nstl::min(utils::div_up(rd_block, vnni_granularity), num_rows);
+
+    for (int r = 0; r < r_end; ++r) {
+        load(zmm_1, ptr[reg_data_aux]);
+
+        if (r * vnni_granularity + 1 >= rd_block) {
+            vpxord(zmm_2, zmm_2, zmm_2);
+        } else {
+            load(zmm_2, ptr[reg_data_aux + reg_data_stride]);
+        }
+
+        vcvtne2ps2bf16(zmm_1, zmm_2, zmm_1);
+        vpermw(zmm_1, zmm_bf32_pemute, zmm_1);
+        vmovups(ptr[reg_buf + r * zmm_width_in_bytes], zmm_1);
+        lea(reg_data_aux,
+                ptr[reg_data_aux + vnni_granularity * reg_data_stride]);
+    }
+
+    // zero rest of the tile data
+    if (r_end < num_rows) {
+        vpxord(zmm_2, zmm_2, zmm_2);
+        for (int r = r_end; r < num_rows; ++r)
+            vmovups(ptr[reg_buf + r * zmm_width_in_bytes], zmm_2);
+    }
+}
+
+void jit_brgemm_amx_uker_base_t::maybe_pre_process_data(const Tmm &t1,
+        reg64_t reg_base, size_t offset, reg64_t reg_stride, int bs, int rdb,
+        matrix_kind_t mk, bool is_rd_tail) {
+
+    auto should_save_transform = [&](matrix_kind_t mk) {
+        // save if there is a reuse
+        if (mk == matrix_A) {
+            return brg.ldb + (brg.ldb_tail != 0) > brg.ld_block2;
+        } else {
+            return brg.bdb + (brg.bdb_tail != 0) > brg.bd_block2;
+        }
     };
 
+    const bool is_A = mk == matrix_A;
+    auto &transform_buf = is_A ? transform_buf_map_A_ : transform_buf_map_B_;
+
+    const int transform_offset
+            = use_ils_ ? brg.get_num_C_tiles() * tile_size : 0;
+    const int max_bdb2 = brg.bd_block2;
+    const int max_rdb = brg.rdb + (brg.rdb_tail != 0);
+    const int matrix_a_offset = transform_offset;
+    const int matrix_b_offset = transform_offset
+            + tile_size
+                    * (nstl::max<int>(should_save_transform(mk),
+                            should_save_transform(matrix_A) * brg.brgattr.max_bs
+                                    * max_bdb2 * max_rdb));
+    const int matrix_offset = is_A ? matrix_a_offset : matrix_b_offset;
+    const std::string key = std::to_string(bs) + "_" + std::to_string(offset);
+
+    if (transform_buf.find(key) != transform_buf.end()) {
+        auto buf_idx = transform_buf[key];
+        auto offt = matrix_offset + buf_idx * tile_size;
+        tileloadd(t1, ptr[reg_buf + reg_bf32_stride + offt]);
+        return;
+    }
+
+    int buf_offt = matrix_offset;
+    // save offset of the transformation if required.
+    if (should_save_transform(mk)) {
+        auto buf_idx = transform_buf.size();
+        buf_offt = matrix_offset + buf_idx * tile_size;
+        transform_buf[key] = buf_idx;
+    }
+
+    if (buf_offt) add(reg_buf, buf_offt);
+    mov(reg_bf32_stride, zmm_width_in_bytes);
+
+    assert(t1.getIdx() >= 0 && t1.getIdx() < 16);
+    const int num_rows = palette_.rows[t1.getIdx()];
+    const int num_col_bytes = palette_.cols[t1.getIdx()];
+    if (is_A) {
+        bf32_downconvert(num_rows, num_col_bytes, reg_base, offset, reg_stride,
+                reg_buf, is_rd_tail);
+    } else {
+        bf32_downconvert_to_vnni(num_rows, num_col_bytes, reg_base, offset,
+                reg_stride, reg_buf, is_rd_tail);
+    }
+
+    // load into tmm from the transformed data.
+    tileloadd(t1, ptr[reg_buf + reg_bf32_stride]);
+
+    // reset buf pointer.
+    if (buf_offt) sub(reg_buf, buf_offt);
+}
+
+void jit_brgemm_amx_uker_base_t::gemm_microkernel_amx(int bs, int bd_block2,
+        int ldb_ind, int ld_block2, bool is_rd_tail, bool is_ld_tail) {
+
     int rbd_block = 0;
     size_t a_offset = 0, b_offset = 0;
     if (is_rd_tail) {
@@ -1116,11 +1390,6 @@ void jit_brgemm_amx_uker_base_t::gemm_microkernel_amx(int bd_block2,
         b_offset = a_offset = 0;
     }
 
-    const bool try_load_nt_A
-            = (brg.brgattr.hint_innermost_loop == brgemm_bd_loop_innermost);
-    const bool try_load_nt_B
-            = (brg.brgattr.hint_innermost_loop == brgemm_ld_loop_innermost);
-
     for (int rdb = 0; rdb < rbd_block; rdb++) {
         int bd_inp_bdb = inp_bd_;
         const size_t rdb_A_off
@@ -1129,30 +1398,29 @@ void jit_brgemm_amx_uker_base_t::gemm_microkernel_amx(int bd_block2,
                 + b_offset_ + b_offset;
 
         for (int bdb = 0; bdb < bd_block2; bdb++) {
-            maybe_tileloadd_nt(Tmm(brg.get_A_tensor(bdb)), reg_aux_A,
-                    rdb_A_off + A_offset(bd_inp_bdb), reg_stride_lda,
-                    try_load_nt_A);
+            maybe_tileloadd_nt(matrix_kind_t::matrix_A, bs, bdb, rdb,
+                    rdb_A_off + A_offset(bd_inp_bdb), is_rd_tail);
             bd_inp_bdb += brg.bd_block;
             bd_inp_bdb = skipped_bd_mask(bd_inp_bdb);
             for (int ldb = 0; ldb < ld_block2; ldb++) {
                 if (bdb == 0) {
                     const int ldb_idx = (is_ld_tail) ? brg.ld_block2 : ldb;
-                    maybe_tileloadd_nt(Tmm(brg.get_B_tensor(ldb_idx)),
-                            reg_aux_B, rdb_B_off + B_offset(ldb),
-                            reg_stride_ldb, try_load_nt_B);
+                    maybe_tileloadd_nt(matrix_kind_t::matrix_B, bs, ldb_idx,
+                            rdb, rdb_B_off + B_offset(ldb), is_rd_tail);
                 }
                 if (ldb == 0) {
                     if (bdb > 0) {
                         const int ldb_idx
                                 = (is_ld_tail) ? brg.ld_block2 : ld_block2 - 1;
                         const int bdb_idx = bdb - 1;
-                        tdpbxxd(bdb_idx, ldb_idx);
+                        tdpbxxd(bdb_idx, ldb_idx, bd_block2, ld_block2,
+                                ldb_ind);
                     }
                 } else {
                     const int ldb_idx
                             = (is_ld_tail) ? brg.ld_block2 : (ldb - 1);
                     const int bdb_idx = bdb;
-                    tdpbxxd(bdb_idx, ldb_idx);
+                    tdpbxxd(bdb_idx, ldb_idx, bd_block2, ld_block2, ldb_ind);
                 }
             }
         }
@@ -1160,7 +1428,7 @@ void jit_brgemm_amx_uker_base_t::gemm_microkernel_amx(int bd_block2,
         {
             const int ldb_idx = (is_ld_tail) ? brg.ld_block2 : (ld_block2 - 1);
             const int bdb_idx = bd_block2 - 1;
-            tdpbxxd(bdb_idx, ldb_idx);
+            tdpbxxd(bdb_idx, ldb_idx, bd_block2, ld_block2, ldb_ind);
         }
     }
 }
@@ -1177,33 +1445,53 @@ void jit_brgemm_amx_uker_base_t::ldb_loop(int bd_block2, int ld_block2,
         ils_vecs_per_store_ = (calc_ops) ? div_up(ils_store_ops_, calc_ops) : 0;
         pfo_vecs_per_store_ = ils_vecs_per_store_;
 
-        size_t l_c_offset = (is_ld_tail) ? ldb_C_offset(1, true)
-                                         : ldb_C_offset(ld_block2);
-        l_c_offset *= l_ldb;
-        l_c_offset += c_offset;
-        size_t l_d_offset = (is_ld_tail) ? ldb_D_offset(1, true)
-                                         : ldb_D_offset(ld_block2);
-        l_d_offset *= l_ldb;
-        l_d_offset += d_offset;
-
         const auto l_ldb_ind = l_ldb * (is_ld_tail ? brg.ldb_tail : ld_block2);
 
-        load_accumulators(bd_block2, ld_block2, is_ld_tail);
+        load_accumulators(
+                bd_block2, ld_block2, is_ld_tail, ldb_ind + l_ldb_ind);
 
         if (brg.alpha != 0.f) {
-            for (int bs = 0; bs < brg.brgattr.max_bs; bs++) {
-                set_A_B_matrices(bs);
-                gemm_microkernel_amx(
-                        bd_block2, ldb_ind, ld_block2, false, is_ld_tail);
+            if (brg.brgattr.var_bs) {
+                Label BS_loop_label, end_BS_loop_label;
 
-                if (brg.rdb_tail != 0) {
-                    gemm_microkernel_amx(
-                            bd_block2, ldb_ind, ld_block2, true, is_ld_tail);
+                mov(reg_BS_loop, reg_BS);
+                cmp(reg_BS_loop, 0);
+                jz(end_BS_loop_label, T_NEAR);
+                mov(reg_aux1_batch, reg_addr_batch);
+
+                L_aligned(BS_loop_label, 64);
+                {
+                    set_A_B_matrices();
+                    add(reg_aux1_batch, sizeof(brgemm_batch_element_t));
+                    prefetcht0(ptr[reg_aux1_batch]);
+                    gemm_microkernel_amx(0, bd_block2, ldb_ind, ld_block2,
+                            false, is_ld_tail);
+
+                    if (brg.rdb_tail != 0) {
+                        gemm_microkernel_amx(0, bd_block2, ldb_ind, ld_block2,
+                                true, is_ld_tail);
+                    }
+                    dec(reg_BS_loop);
+                    cmp(reg_BS_loop, 0);
+                    jg(BS_loop_label, T_NEAR);
+                }
+                L(end_BS_loop_label);
+
+            } else {
+                for (int bs = 0; bs < brg.brgattr.max_bs; bs++) {
+                    set_A_B_matrices(bs);
+                    gemm_microkernel_amx(bs, bd_block2, ldb_ind, ld_block2,
+                            false, is_ld_tail);
+
+                    if (brg.rdb_tail != 0) {
+                        gemm_microkernel_amx(bs, bd_block2, ldb_ind, ld_block2,
+                                true, is_ld_tail);
+                    }
                 }
             }
         }
-        store_accumulators(bd_block2, ld_block2, l_ldb, is_ld_tail, l_c_offset,
-                l_d_offset, ldb_ind + l_ldb_ind, apply_post_ops);
+        store_accumulators(bd_block2, ld_block2, is_ld_tail,
+                ldb_ind + l_ldb_ind, apply_post_ops);
 
         b_offset_ += (is_ld_tail) ? ldb_B_offset(1, true)
                                   : ldb_B_offset(ld_block2);
@@ -1255,6 +1543,9 @@ void jit_brgemm_amx_uker_base_t::bdb_loop(bool apply_post_ops) {
         size_t c_offset = 0;
         size_t d_offset = 0;
         int ldb_ind = 0;
+        // clear the transform cache for A, as the existing data is invalid as
+        // we move to next bdb2 block.
+        transform_buf_map_A_.clear();
         if (brg.ldb2 > 0) {
             const bool is_ld_reg_tail = false;
             const bool is_ld_tail = false;
@@ -1311,7 +1602,14 @@ void jit_brgemm_amx_uker_base_t::generate() {
     LDB_size_ = brg.typesize_B * brg.LDB;
     LDC_size_ = brg.typesize_C * brg.LDC;
     LDD_size_ = brg.typesize_D * brg.LDD;
-    ld_block_B_size_ = brg.typesize_B * brg.ld_block;
+
+    LDA2_size_ = brg.typesize_A * brg.LDA2;
+    LDB2_size_ = brg.typesize_B * brg.LDB2;
+    LDC2_size_M_ = brg.typesize_C * brg.LDC2_M;
+    LDC2_size_N_ = brg.typesize_C * brg.LDC2_N;
+
+    ld_block_B_size_ = brg.typesize_B
+            * ((brg.brgattr.LDB2 != 0) ? brg.brgattr.LDB2 : brg.ld_block);
     ld_block_C_size_ = brg.typesize_C * brg.ld_block;
     ld_block_D_size_ = brg.typesize_D * brg.ld_block;
     ld_block_bias_size_ = brg.typesize_bias * brg.ld_block;
@@ -1322,13 +1620,36 @@ void jit_brgemm_amx_uker_base_t::generate() {
     ldb_tail_D_size_ = brg.typesize_D * brg.ldb_tail;
     ldb_tail_zp_size_ = sizeof(int32_t) * brg.ldb_tail;
 
-    need_to_apply_alpha_beta_ = brg.beta != 0.f || brg.alpha != 1.f;
+    // if beta == 1 and C datatype is f32 it is better to perform addition by
+    // reading tiles directly from C instead of by reading/writing by vectors
+    may_load_accumulators_
+            = (brg.beta == 1.f && brg.dt_c == data_type::f32 && !brg.is_bf32)
+            && brg.brgattr.bd_mask_level == 0;
+    need_to_apply_alpha_beta_
+            = (brg.beta != 0.f && !may_load_accumulators_) || brg.alpha != 1.f;
     const bool has_zero_points = !everyone_is(brgemm_broadcast_t::none,
             brg.zp_type_a, brg.zp_type_b, brg.zp_type_c);
     are_post_ops_applicable_ = one_of(true, brg.with_eltwise, brg.with_binary,
             brg.with_scales, brg.with_bias, brg.with_sum, brg.dt_d != brg.dt_c,
             has_zero_points);
 
+    // second level blocking eligible only if we don't use store by vectors for now
+
+    assert(IMPLICATION(are_post_ops_applicable_ || need_to_apply_alpha_beta_
+                    || brg.brgattr.bd_mask_level,
+            !brg.is_blocked && !brg.brgattr.var_bs));
+
+    //! 'maybe_pre_process_data' for 'is_bf32' uses 'bs' value,
+    //! so we can't support var_bs for is_bf32 for this moment
+    assert(IMPLICATION(brg.brgattr.var_bs, !brg.is_bf32));
+
+    Label permute_index_table;
+    if (brg.is_bf32) {
+        brgemm_init_tiles(brg, (char *)(&palette_));
+        mov(reg_tmp_gpr, permute_index_table);
+        vmovups(zmm_bf32_pemute, ptr[reg_tmp_gpr]);
+    }
+
     reg64_t reg_mask = rax;
 
     mov(reg_mask, full_mask);
@@ -1336,8 +1657,8 @@ void jit_brgemm_amx_uker_base_t::generate() {
     mov(reg_mask, tail_mask);
     kmovq(ld_tail_mask, reg_mask);
 
-    mov(reg_stride_lda, LDA_size_);
-    mov(reg_stride_ldb, brg.rd_step * LDB_size_);
+    mov(reg_stride_lda, lda());
+    mov(reg_stride_ldb, ldb());
 
     read_params();
 
@@ -1351,6 +1672,8 @@ void jit_brgemm_amx_uker_base_t::generate() {
         jz(label_store_without_post_ops, T_NEAR);
         bdb_loop(true);
         jmp(label_to_ret, T_NEAR);
+        transform_buf_map_A_.clear();
+        transform_buf_map_B_.clear();
         L(label_store_without_post_ops);
     }
     bdb_loop(false);
@@ -1361,6 +1684,16 @@ void jit_brgemm_amx_uker_base_t::generate() {
     postamble();
 
     if (brg.with_eltwise) postops_injector_->prepare_table();
+
+    if (brg.is_bf32) {
+        align(64);
+        L(permute_index_table);
+        const uint16_t _idx[32] = {0, 16, 1, 17, 2, 18, 3, 19, 4, 20, 5, 21, 6,
+                22, 7, 23, 8, 24, 9, 25, 10, 26, 11, 27, 12, 28, 13, 29, 14, 30,
+                15, 31};
+        for (size_t i = 0; i < 32; ++i)
+            dw(_idx[i]);
+    }
 }
 
 brgemm_amx_uker_t::brgemm_amx_uker_t(const brgemm_t abrd) {
