@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2020-2022 Intel Corporation
+* Copyright 2020-2023 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -153,11 +153,9 @@ status_t brgemm_inner_product_fwd_t<isa>::execute_forward(
             c_buffer = c_buffer_global + c_buffer_off;
         }
 
-        char *wsp_tile = is_amx ? wsp_tile_base
-                        + ithr
-                                * jit_brgemm_primitive_conf_t::
-                                        tile_wsp_per_thread
-                                : nullptr;
+        char *wsp_tile = is_amx
+                ? wsp_tile_base + ithr * jbgp.amx_buf_size_per_thread
+                : nullptr;
         int icb = icc * jbgp.nb_ic_blocking;
         int ic = icb * jbgp.ic_block;
 
@@ -171,8 +169,8 @@ status_t brgemm_inner_product_fwd_t<isa>::execute_forward(
                 = (jbgp.use_buffer_a ? utils::rnd_up(jbgp.ic, jbgp.ic_block)
                                      : jbgp.ic)
                 - ic;
-        const int gemm_batch = nstl::min(
-                jbgp.gemm_batch_size, remaining_ic_blks / jbgp.ic_block);
+        const int gemm_batch
+                = nstl::min(jbgp.gemm_batch_size, remaining_ic_blks / jbgp.K);
 
         auto is_bs_tail = (gemm_batch != jbgp.gemm_batch_size);
         int brg_ker_idx = brgemm_inner_product_utils::get_brg_kernel_index(
@@ -180,6 +178,7 @@ status_t brgemm_inner_product_fwd_t<isa>::execute_forward(
         auto brg_kernel = brg_kernels_[brg_ker_idx].get();
 
         if (copy_buffer_a) {
+            assert(!jbgp.is_bf32);
             auto src_ptr = src + get_blk_off(src_d, jbgp.src_dt, n, ic);
             copy_data_chunk(copy_src_kernel_, a_buffer, src_ptr,
                     is_os_tail ? jbgp.mb - n : jbgp.os_block, is_last_ic_chunk);
@@ -187,15 +186,17 @@ status_t brgemm_inner_product_fwd_t<isa>::execute_forward(
         if (gemm_batch > 0 && brg_kernel != nullptr) {
             if (is_amx && (is_os_tail || is_oc_tail))
                 amx_tile_configure(&brg_kernel_palettes_[brg_ker_idx][0]);
+            const int ic_blocks_per_batch = jbgp.K / jbgp.ic_block;
             for (int b = 0; b < gemm_batch; b++) {
                 auto A_ptr = jbgp.use_buffer_a
-                        ? (a_buffer + src_dt_size * b * jbgp.ic_block)
+                        ? (a_buffer + src_dt_size * b * jbgp.K)
                         : (src
                                 + get_blk_off(src_d, jbgp.src_dt, n,
-                                        ic + b * jbgp.ic_block));
+                                        ic + b * jbgp.K));
                 addr_batch[b].ptr.A = A_ptr;
                 addr_batch[b].ptr.B = weights
-                        + get_blk_off(weights_d, jbgp.wei_dt, ocb, icb + b);
+                        + get_blk_off(weights_d, jbgp.wei_dt, ocb,
+                                icb + b * ic_blocks_per_batch);
             }
 
             auto ptr_D = dst + dst_off;
@@ -326,7 +327,8 @@ status_t brgemm_inner_product_fwd_t<isa>::execute_forward(
             int icc = 0, osb = 0, ocb = 0;
 
             // If buffer is required, then inner-most loop will be over icc_work
-            const bool ocb_inner_most = is_f32 && !jbgp.use_buffer;
+            const bool ocb_inner_most
+                    = is_f32 && !(jbgp.is_bf32 || jbgp.use_buffer);
             if (ocb_inner_most)
                 nd_iterator_init(
                         0, icc, icc_work, osb, osb_work, ocb, ocb_work);
@@ -434,9 +436,7 @@ status_t brgemm_inner_product_fwd_t<isa>::execute_forward(
                                     + get_dst_reduced_off(0, osb, ocb);
 
                             char *wsp_tile = is_amx ? wsp_tile_base
-                                            + ithr
-                                                    * jit_brgemm_primitive_conf_t::
-                                                            tile_wsp_per_thread
+                                            + ithr * jbgp.amx_buf_size_per_thread
                                                     : nullptr;
 
                             void *scratch = is_amx
@@ -535,7 +535,8 @@ void brgemm_inner_product_bwd_data_t<isa>::execute_backward_data(
     const int num_threads = (work_amount == 1 ? 1 : jbgp.nthr);
 
     const auto get_weights_ptr = [&](int icb, int ocb) {
-        int fwd_ic_block = (is_amx) ? 2 * jbgp.simd_w : jbgp.simd_w;
+        int fwd_ic_block
+                = (is_amx && !jbgp.is_bf32) ? 2 * jbgp.simd_w : jbgp.simd_w;
         int fwd_oc_block = 0;
         switch (jbgp.wei_tag) {
             case OI16i64o:
@@ -626,11 +627,9 @@ void brgemm_inner_product_bwd_data_t<isa>::execute_backward_data(
         char *a_buffer = jbgp.use_buffer_a
                 ? a_buffer_global + ithr * a_buffer_size_per_thr
                 : diff_dst;
-        char *wsp_tile = is_amx ? wsp_tile_base
-                        + ithr
-                                * jit_brgemm_primitive_conf_t::
-                                        tile_wsp_per_thread
-                                : nullptr;
+        char *wsp_tile = is_amx
+                ? wsp_tile_base + ithr * jbgp.amx_buf_size_per_thread
+                : nullptr;
 
         bool kernel_init = do_init;
 
@@ -888,8 +887,6 @@ struct brgemm_inner_product_bwd_weights_t<isa>::thread_info_t {
 
     const memory_tracking::grantor_t scratchpad;
 
-    char *buffer_a = nullptr;
-    char *buffer_b = nullptr;
     char *buffer_c = nullptr;
     char *buffer_bias = nullptr;
     char *wsp_tile_base = nullptr;
@@ -926,11 +923,62 @@ struct brgemm_inner_product_bwd_weights_t<isa>::thread_info_t {
                 ? scratchpad.template get<char>(key_iprod_bias_bf16_convert_wsp)
                 : nullptr;
 
-        buffer_a = scratchpad.template get<char>(key_brgemm_primitive_buffer_a);
-        buffer_b = jbgp.use_buffer_b
+        buffer_a_
+                = scratchpad.template get<char>(key_brgemm_primitive_buffer_a);
+        buffer_b_ = jbgp.use_buffer_b
                 ? scratchpad.template get<char>(key_brgemm_primitive_buffer_b)
                 : nullptr;
 
+        thread_local_input_buffers_
+                = jbgp.ip_bwd_w_local_buffers_for_input_tensors;
+        int ic_chunks = utils::div_up(jbgp.nb_ic, jbgp.nb_ic_blocking);
+        int os_chunks = utils::div_up(jbgp.nb_os, jbgp.nb_os_blocking);
+        nb_ic_blocking_ = jbgp.nb_ic_blocking;
+        nb_oc_blocking_ = jbgp.nb_oc_blocking;
+        const size_t os_chunks_per_thr = utils::div_up(os_chunks, jbgp.nthr_mb);
+        const size_t num_os_chunks_per_thread
+                = thread_local_input_buffers_ ? 1 : os_chunks_per_thr;
+
+        if (jbgp.use_buffer_a) {
+            const size_t src_dt = types::data_type_size(jbgp.src_dt);
+            const size_t ic_chunks_per_thr
+                    = utils::div_up(ic_chunks, jbgp.nthr_ic_b);
+            const size_t num_ic_chunks_per_thread
+                    = thread_local_input_buffers_ ? 1 : ic_chunks_per_thr;
+            const size_t block_A_size = src_dt * jbgp.LDA * jbgp.M;
+            const size_t os_chunk_A_buffer
+                    = jbgp.gemm_batch_size * block_A_size;
+            const size_t ic_os_chunk_A_buffer
+                    = jbgp.nb_ic_blocking * os_chunk_A_buffer;
+
+            buffer_a_icb_shift_ = os_chunk_A_buffer;
+            buffer_a_osb_shift_ = block_A_size;
+            buffer_a_osc_shift_ = thread_local_input_buffers_
+                    ? 0
+                    : ic_chunks_per_thr * ic_os_chunk_A_buffer;
+            const size_t buffer_a_thread_shift = num_ic_chunks_per_thread
+                    * num_os_chunks_per_thread * ic_os_chunk_A_buffer;
+
+            buffer_a_ = buffer_a_ + ithr * buffer_a_thread_shift;
+        }
+
+        if (jbgp.use_buffer_b) {
+            const size_t dst_dt = types::data_type_size(jbgp.dst_dt);
+            const size_t block_B_size = dst_dt * jbgp.LDB * jbgp.K;
+            const size_t os_chunk_B_buffer
+                    = jbgp.gemm_batch_size * block_B_size;
+            buffer_b_ocb_shift_ = dst_dt * jbgp.oc_block
+                    * data_type_vnni_granularity(jbgp.dst_dt);
+            buffer_b_osb_shift_ = block_B_size;
+            buffer_b_osc_shift_
+                    = thread_local_input_buffers_ ? 0 : os_chunk_B_buffer;
+
+            const size_t buffer_b_thread_shift
+                    = num_os_chunks_per_thread * os_chunk_B_buffer;
+
+            buffer_b_ = buffer_b_ + ithr * buffer_b_thread_shift;
+        }
+
         wsp_tile_base = is_amx
                 ? ctx.get_scratchpad_grantor().template get<char>(
                         key_conv_amx_tile_buffer)
@@ -946,8 +994,6 @@ struct brgemm_inner_product_bwd_weights_t<isa>::thread_info_t {
         ithr_os_c = ithr / nthr_ic_c / nthr_oc_c;
 
         int oc_chunks = utils::div_up(jbgp.nb_oc, jbgp.nb_oc_blocking);
-        int ic_chunks = utils::div_up(jbgp.nb_ic, jbgp.nb_ic_blocking);
-        int os_chunks = utils::div_up(jbgp.nb_os, jbgp.nb_os_blocking);
 
         /* reduction dimension */
         balance211(os_chunks, nthr_os_c, ithr_os_c, os_c_start, os_c_end);
@@ -963,6 +1009,45 @@ struct brgemm_inner_product_bwd_weights_t<isa>::thread_info_t {
             barrier_ctx = scratchpad.template get<simple_barrier::ctx_t>(
                     key_conv_wei_bia_reduction_bctx);
     }
+
+    char *get_buffer_a_ptr(int icb, int osc) const {
+        if (!buffer_a_) return (char *)nullptr;
+
+        const int icb_idx = thread_local_input_buffers_
+                ? icb % nb_ic_blocking_
+                : icb - ic_c_start * nb_ic_blocking_;
+        const int osc_idx = thread_local_input_buffers_ ? 0 : osc - os_c_start;
+
+        return buffer_a_ + osc_idx * buffer_a_osc_shift_
+                + icb_idx * buffer_a_icb_shift_;
+    }
+
+    char *get_buffer_b_ptr(int ocb, int osc) const {
+        if (!buffer_b_) return (char *)nullptr;
+
+        const int ocb_idx = ocb % nb_oc_blocking_;
+        const int osc_idx = thread_local_input_buffers_ ? 0 : osc - os_c_start;
+
+        return buffer_b_ + osc_idx * buffer_b_osc_shift_
+                + ocb_idx * buffer_b_ocb_shift_;
+    }
+
+    size_t get_buffer_a_osb_shift() const { return buffer_a_osb_shift_; }
+    size_t get_buffer_b_osb_shift() const { return buffer_b_osb_shift_; }
+
+private:
+    char *buffer_a_ = nullptr;
+    char *buffer_b_ = nullptr;
+
+    bool thread_local_input_buffers_ = false;
+    int nb_ic_blocking_ = 1;
+    int nb_oc_blocking_ = 1;
+    size_t buffer_a_icb_shift_ = 0;
+    size_t buffer_a_osc_shift_ = 0;
+    size_t buffer_a_osb_shift_ = 0;
+    size_t buffer_b_ocb_shift_ = 0;
+    size_t buffer_b_osc_shift_ = 0;
+    size_t buffer_b_osb_shift_ = 0;
 };
 
 template <cpu_isa_t isa>
@@ -1135,8 +1220,6 @@ void brgemm_inner_product_bwd_weights_t<isa>::compute_diff_weights_and_bias(
                     key_brgemm_primitive_batch);
 
     constexpr bool is_amx_bf16 = (isa == avx512_core_bf16_amx_bf16);
-    char *a_buffer_global = ti->buffer_a;
-    char *b_buffer_global = ti->buffer_b;
     char *wsp_tile_global = (is_amx_bf16) ? ti->wsp_tile_base : nullptr;
     int os_chunks = utils::div_up(jbgp.nb_os, jbgp.nb_os_blocking);
 
@@ -1154,46 +1237,26 @@ void brgemm_inner_product_bwd_weights_t<isa>::compute_diff_weights_and_bias(
         }
     };
 
-    const auto ker = [&](const int osc, const int icb, const int ocb) {
-        dim_t os_chunks_per_thr = utils::div_up(os_chunks, jbgp.nthr_mb);
-        int ic_chunks = utils::div_up(jbgp.nb_ic, jbgp.nb_ic_blocking);
-        dim_t ic_chunks_per_thr = utils::div_up(ic_chunks, jbgp.nthr_ic_b);
-
-        int osc_l_idx = osc - ti->os_c_start;
-        int icb_l_idx = icb - ti->ic_c_start * jbgp.nb_ic_blocking;
-        int ocb_l_idx = ocb - ti->oc_c_start * jbgp.nb_oc_blocking;
-        dim_t a_buf_idx = osc_l_idx * ic_chunks_per_thr * jbgp.nb_ic_blocking
-                + icb_l_idx;
-        int b_buf_idx = osc_l_idx;
-
-        constexpr int tile_size = 1024;
+    const auto a_buf_osb_shift = ti->get_buffer_a_osb_shift();
+    const auto b_buf_osb_shift = ti->get_buffer_b_osb_shift();
 
+    const auto ker = [&](const int osc, const int icb, const int ocb,
+                             const int osc_prev, const int icc_prev,
+                             const int occ_prev) {
         brgemm_batch_element_t *addr_batch
                 = addr_batch_global + ti->ithr * jbgp.adjusted_batch_size;
-        const int size_A = jbgp.LDA * jbgp.M;
-        const int size_B = jbgp.LDB * rnd_up(jbgp.K, 2);
-        char *a_buffer = a_buffer_global
-                + types::data_type_size(jbgp.src_dt)
-                        * ((ti->ithr * os_chunks_per_thr * ic_chunks_per_thr
-                                           * jbgp.nb_ic_blocking
-                                   + a_buf_idx)
-                                * jbgp.gemm_batch_size * jbgp.os_block
-                                * jbgp.ic_block);
-        char *b_buffer = b_buffer_global ? b_buffer_global
-                        + types::data_type_size(jbgp.dst_dt)
-                                * ((ti->ithr * os_chunks_per_thr + b_buf_idx)
-                                                * jbgp.gemm_batch_size
-                                                * jbgp.os_block * jbgp.LDB
-                                        + (ocb_l_idx % jbgp.nb_oc_blocking)
-                                                * jbgp.oc_block)
-                                         : nullptr;
-
         char *wsp_tile = is_amx_bf16
-                ? wsp_tile_global + ti->ithr * 4 * tile_size
+                ? wsp_tile_global + ti->ithr * jbgp.amx_buf_size_per_thread
                 : nullptr;
         int ic = icb * jbgp.ic_block;
         int oc = ocb * jbgp.oc_block;
-        int n = osc * jbgp.nb_os_blocking * jbgp.os_block;
+        int osb = osc * jbgp.nb_os_blocking;
+        int n = osb * jbgp.os_block;
+
+        // Base buffer pointers for the current kernel iteration,
+        // x_buf_osb_shift values are used for shifting wrt osb iter variable
+        char *a_buffer = ti->get_buffer_a_ptr(icb, osc);
+        char *b_buffer = ti->get_buffer_b_ptr(ocb, osc);
 
         bool kernel_init = (osc == ti->os_c_start);
 
@@ -1209,6 +1272,20 @@ void brgemm_inner_product_bwd_weights_t<isa>::compute_diff_weights_and_bias(
         const bool transform_weights_to_vnni = jbgp.wei_dt == bf16
                 && (jbgp.nthr_mb == 1 || os_chunks == 1)
                 && osc == (os_chunks - 1);
+        const bool transform_b = jbgp.ip_bwd_w_local_buffers_for_input_tensors
+                ? jbgp.use_buffer_b && icb % jbgp.nb_ic_blocking == 0
+                        && ocb % jbgp.nb_oc_blocking == 0
+                        && IMPLICATION(osc_prev == osc,
+                                occ_prev != ocb / jbgp.nb_oc_blocking)
+                : jbgp.use_buffer_b
+                        && icb == ti->ic_c_start * jbgp.nb_ic_blocking
+                        && ocb % jbgp.nb_oc_blocking == 0;
+        const bool transform_a = jbgp.ip_bwd_w_local_buffers_for_input_tensors
+                ? jbgp.use_buffer_a && ocb % jbgp.nb_oc_blocking == 0
+                        && IMPLICATION(osc_prev == osc,
+                                icc_prev != icb / jbgp.nb_ic_blocking)
+                : jbgp.use_buffer_a
+                        && ocb == ti->oc_c_start * jbgp.nb_oc_blocking;
 
         auto nb_os_b = is_os_tail ? (jbgp.mb - n) / jbgp.os_block
                                   : jbgp.nb_os_blocking;
@@ -1226,18 +1303,18 @@ void brgemm_inner_product_bwd_weights_t<isa>::compute_diff_weights_and_bias(
         if (nb_os_b > 0 && brg_kernel != nullptr) {
             if (is_amx_bf16)
                 amx_tile_configure(&brg_kernel_palettes_[brg_ker_idx][0]);
-            if (jbgp.use_buffer_a && ocb_l_idx == 0) {
+            if (transform_a) {
                 const memory_desc_wrapper src_d(pd()->src_md());
                 auto src_ptr = ti->src
                         + types::data_type_size(jbgp.src_dt)
                                 * src_d.blk_off(n, ic);
+
                 transform_matrix_a_chunk(a_buffer, src_ptr, nb_os_b,
                         is_ic_tail ? jbgp.ic % jbgp.ic_block : jbgp.ic_block,
                         jbgp.os_block);
             }
 
-            if (jbgp.use_buffer_b && icb_l_idx == 0
-                    && ocb_l_idx % jbgp.nb_oc_blocking == 0) {
+            if (transform_b) {
                 auto diff_dst_ptr = diff_dst
                         + types::data_type_size(jbgp.dst_dt)
                                 * diff_dst_d.blk_off(n, oc);
@@ -1246,18 +1323,14 @@ void brgemm_inner_product_bwd_weights_t<isa>::compute_diff_weights_and_bias(
             }
 
             for (int os_block = 0; os_block < nb_os_b; os_block++) {
-                auto a_ptr = a_buffer
-                        + types::data_type_size(jbgp.src_dt) * os_block
-                                * size_A;
+                auto a_ptr = a_buffer + os_block * a_buf_osb_shift;
                 addr_batch[os_block].ptr.A = a_ptr;
                 auto diff_dst_ptr = diff_dst
                         + types::data_type_size(jbgp.dst_dt)
                                 * diff_dst_d.blk_off(
                                         n + os_block * jbgp.os_block, oc);
                 if (jbgp.use_buffer_b) {
-                    auto b_ptr = b_buffer
-                            + types::data_type_size(jbgp.dst_dt) * os_block
-                                    * size_B;
+                    auto b_ptr = b_buffer + os_block * b_buf_osb_shift;
                     addr_batch[os_block].ptr.B = b_ptr;
                 } else {
                     addr_batch[os_block].ptr.B = diff_dst_ptr;
@@ -1284,10 +1357,9 @@ void brgemm_inner_product_bwd_weights_t<isa>::compute_diff_weights_and_bias(
 
         if (is_os_tail) {
             int os_block = nb_os_b;
-            auto a_ptr = a_buffer
-                    + types::data_type_size(jbgp.src_dt) * os_block
-                            * jbgp.ic_block * jbgp.os_block;
-            if (jbgp.use_buffer_a && ocb_l_idx == 0) {
+            auto a_ptr = a_buffer + os_block * a_buf_osb_shift;
+
+            if (transform_a) {
                 const memory_desc_wrapper src_d(pd()->src_md());
                 auto src_ptr = ti->src
                         + types::data_type_size(jbgp.src_dt)
@@ -1304,10 +1376,9 @@ void brgemm_inner_product_bwd_weights_t<isa>::compute_diff_weights_and_bias(
                             * diff_dst_d.blk_off(
                                     n + os_block * jbgp.os_block, oc);
             if (jbgp.use_buffer_b) {
-                auto b_ptr = b_buffer
-                        + types::data_type_size(jbgp.dst_dt) * os_block
-                                * jbgp.os_block * jbgp.LDB;
-                if (icb_l_idx == 0 && ocb_l_idx % jbgp.nb_oc_blocking == 0)
+                auto b_ptr = b_buffer + os_block * b_buf_osb_shift;
+
+                if (transform_b)
                     transform_matrix_b_chunk(b_ptr, diff_dst_ptr, 1,
                             curr_oc_chunk_sz, jbgp.mb % jbgp.os_block);
                 addr_batch[0].ptr.B = b_ptr;
@@ -1359,15 +1430,26 @@ void brgemm_inner_product_bwd_weights_t<isa>::compute_diff_weights_and_bias(
     const auto loop_end = occ_work * icc_work * osc_work;
 
     int occ_idx = 0, icc_idx = 0, osc_idx = 0;
-    //TODO: Introduce loop order enum for brgemm-based implementations
-    const bool osc_loop_outermost = jbgp.harness == harness_mb_reduction;
-    if (osc_loop_outermost)
-        nd_iterator_init(loop_idx, osc_idx, osc_work, occ_idx, occ_work,
-                icc_idx, icc_work);
-    else
-        nd_iterator_init(loop_idx, occ_idx, occ_work, icc_idx, icc_work,
-                osc_idx, osc_work);
+    loop_order_t loop_order = jbgp.ip_bwd_w_local_buffers_for_input_tensors
+            ? loop_order_t::osc_icc_occ
+            : jbgp.harness == harness_mb_reduction ? loop_order_t::osc_occ_icc
+                                                   : loop_order_t::occ_icc_osc;
 
+    switch (loop_order) {
+        case loop_order_t::osc_icc_occ:
+            nd_iterator_init(loop_idx, osc_idx, osc_work, icc_idx, icc_work,
+                    occ_idx, occ_work);
+            break;
+        case loop_order_t::osc_occ_icc:
+            nd_iterator_init(loop_idx, osc_idx, osc_work, occ_idx, occ_work,
+                    icc_idx, icc_work);
+            break;
+        case loop_order_t::occ_icc_osc:
+            nd_iterator_init(loop_idx, occ_idx, occ_work, icc_idx, icc_work,
+                    osc_idx, osc_work);
+    };
+
+    int osc_prev = -1, icc_prev = -1, occ_prev = -1;
     while (loop_idx < loop_end) {
         const int occ = ti->oc_c_start + occ_idx;
         const int icc = ti->ic_c_start + icc_idx;
@@ -1381,16 +1463,28 @@ void brgemm_inner_product_bwd_weights_t<isa>::compute_diff_weights_and_bias(
         for_(int ocb = 0; ocb < ocb_work; ocb++)
         for (int icb = 0; icb < icb_work; icb++) {
             ker(osc, icc * jbgp.nb_ic_blocking + icb,
-                    occ * jbgp.nb_oc_blocking + ocb);
+                    occ * jbgp.nb_oc_blocking + ocb, osc_prev, icc_prev,
+                    occ_prev);
         }
+        osc_prev = osc;
+        icc_prev = icc;
+        occ_prev = occ;
 
         ++loop_idx;
-        if (osc_loop_outermost)
-            nd_iterator_step(
-                    osc_idx, osc_work, occ_idx, occ_work, icc_idx, icc_work);
-        else
-            nd_iterator_step(
-                    occ_idx, occ_work, icc_idx, icc_work, osc_idx, osc_work);
+
+        switch (loop_order) {
+            case loop_order_t::osc_icc_occ:
+                nd_iterator_step(osc_idx, osc_work, icc_idx, icc_work, occ_idx,
+                        occ_work);
+                break;
+            case loop_order_t::osc_occ_icc:
+                nd_iterator_step(osc_idx, osc_work, occ_idx, occ_work, icc_idx,
+                        icc_work);
+                break;
+            case loop_order_t::occ_icc_osc:
+                nd_iterator_step(occ_idx, occ_work, icc_idx, icc_work, osc_idx,
+                        osc_work);
+        };
     }
     if (is_amx_bf16) amx_tile_release();
 }
@@ -1408,8 +1502,10 @@ void brgemm_inner_product_bwd_weights_t<
     const bool is_bf16_out = jbgp.wei_dt == data_type::bf16;
     const int icb_scale = (!is_bf16_out) ? jbgp.ic_block / jbgp.simd_w : 1;
 
-    const int icb_work = ti->ic_c_work * jbgp.nb_ic_blocking;
-    const int ocb_work = ti->oc_c_work * jbgp.nb_oc_blocking;
+    const int icb_work = nstl::min(ti->ic_c_work * jbgp.nb_ic_blocking,
+            jbgp.nb_ic - ti->ic_c_start * jbgp.nb_ic_blocking);
+    const int ocb_work = nstl::min(ti->oc_c_work * jbgp.nb_oc_blocking,
+            jbgp.nb_oc - ti->oc_c_start * jbgp.nb_oc_blocking);
     const int work = ocb_work * icb_work;
 
     int os_chunks = utils::div_up(jbgp.nb_os, jbgp.nb_os_blocking);
