@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2020-2022 Intel Corporation
+* Copyright 2020-2023 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -47,6 +47,7 @@ int get_os_block(const jit_brgemm_primitive_conf_t &jbgp, bool try_to_adjust,
     const bool is_amx_bf16 = jbgp.isa == avx512_core_bf16_amx_bf16;
     const bool is_avx512_bf16 = jbgp.isa == avx512_core_bf16;
     const bool is_f32 = everyone_is(f32, jbgp.src_dt, jbgp.wei_dt, jbgp.dst_dt);
+    const bool is_bf32 = jbgp.is_bf32;
 
     int max_os_block = 0;
     int min_os_block = 0;
@@ -70,7 +71,7 @@ int get_os_block(const jit_brgemm_primitive_conf_t &jbgp, bool try_to_adjust,
         //
         // For f32 data type our objective is to determine the optimal value
         // of os_block such that the work amount per thread ~ 2
-        if (is_f32) {
+        if (is_f32 && !is_bf32) {
             const bool small_work_amt_per_thread
                     = div_up(jbgp.os, max_os_block) * jbgp.nb_oc
                     < 1.8f * jbgp.nthr;
@@ -164,7 +165,7 @@ std::vector<format_tag_t> get_desired_weights_tag(
 int get_oc_block(const jit_brgemm_primitive_conf_t &jbgp, bool try_to_adjust) {
     const bool amx_bf16_bwd_d_noadjust = !try_to_adjust
             && jbgp.prop_kind == backward_data
-            && jbgp.isa == avx512_core_bf16_amx_bf16;
+            && jbgp.isa == avx512_core_bf16_amx_bf16 && !jbgp.is_bf32;
     if (amx_bf16_bwd_d_noadjust) {
         constexpr int amx_bf16_row = 64;
         return amx_bf16_row;
@@ -220,7 +221,8 @@ bool ip_fwd_adjust_thread_balance(const jit_brgemm_primitive_conf_t &jbgp) {
 }
 
 int ip_fwd_get_adjusted_oc_block(const jit_brgemm_primitive_conf_t &jbgp) {
-    const bool is_amx_bf16 = jbgp.isa == avx512_core_bf16_amx_bf16;
+    const bool is_amx_bf16
+            = jbgp.isa == avx512_core_bf16_amx_bf16 && !jbgp.is_bf32;
 
     // we can't change block size on forward and weights update (external)
     // if layout is set by user, for backward data it can be choosen different
@@ -228,7 +230,7 @@ int ip_fwd_get_adjusted_oc_block(const jit_brgemm_primitive_conf_t &jbgp) {
     const bool not_adjustable_oc_block_size
             = !jbgp.is_wei_layout_any && jbgp.prop_kind != backward_data;
 
-    if (IMPLICATION(is_amx_bf16, not_adjustable_oc_block_size))
+    if (IMPLICATION(is_amx_bf16 || jbgp.is_bf32, not_adjustable_oc_block_size))
         return get_oc_block(jbgp);
 
     int oc_block = get_oc_block(jbgp, true);
@@ -267,14 +269,15 @@ bool post_ops_ok(jit_brgemm_primitive_conf_t &jbgp,
             {sum, eltwise, binary}, post_ops, &dst_d,
             false /*sum_at_pos_0_only*/, false /*sum_requires_scale_one*/,
             true /*sum_requires_zp_zero*/,
-            {broadcasting_strategy_t::per_oc,
-                    broadcasting_strategy_t::scalar}));
+            {broadcasting_strategy_t::per_oc, broadcasting_strategy_t::scalar,
+                    broadcasting_strategy_t::no_broadcast}));
 }
 
 status_t init_ip_conf_fwd(jit_brgemm_primitive_conf_t &jbgp,
         const primitive_attr_t &attr, const memory_desc_wrapper &dst_d) {
     const bool is_amx_int8 = jbgp.isa == avx512_core_bf16_amx_int8;
-    const bool is_amx_bf16 = jbgp.isa == avx512_core_bf16_amx_bf16;
+    const bool is_amx_bf16
+            = jbgp.isa == avx512_core_bf16_amx_bf16 && !jbgp.is_bf32;
     const bool is_int8 = one_of(jbgp.src_dt, u8, s8) && jbgp.wei_dt == s8;
     const bool is_f32 = everyone_is(f32, jbgp.src_dt, jbgp.wei_dt, jbgp.dst_dt);
 
@@ -300,11 +303,12 @@ status_t init_ip_conf_fwd(jit_brgemm_primitive_conf_t &jbgp,
     constexpr int amx_int8_row = 64;
     constexpr int amx_bf16_row = 32;
     jbgp.ic_block = (is_amx_int8) ? amx_int8_row
-                                  : (is_amx_bf16) ? amx_bf16_row : jbgp.simd_w;
+            : (is_amx_bf16)       ? amx_bf16_row
+                                  : jbgp.simd_w;
     jbgp.nb_ic = div_up(jbgp.ic, jbgp.ic_block);
 
     // gemm-based inner product performs better when oc = 1
-    if (is_f32 && jbgp.oc == 1) return status::unimplemented;
+    if (is_f32 && !jbgp.is_bf32 && jbgp.oc == 1) return status::unimplemented;
 
     jbgp.oc_block = ip_fwd_get_adjusted_oc_block(jbgp);
     jbgp.nb_oc = div_up(jbgp.oc, jbgp.oc_block);
@@ -320,7 +324,7 @@ status_t init_ip_conf_fwd(jit_brgemm_primitive_conf_t &jbgp,
     //   * 1 <= nb_os_blocking <= 8 AND nb_os_blocking <= nb_os
     //   * Work amount per thread ~ 2
     //   * NOTE: here nb_oc_blocking = 1 as os is large
-    if (jbgp.os > 256 && is_f32) {
+    if (jbgp.os > 256 && is_f32 && !jbgp.is_bf32) {
         jbgp.nb_os_blocking = saturate(1, nstl::min(8, jbgp.nb_os),
                 nstl::min(nstl::max(jbgp.oc / jbgp.os / 2, 1),
                         div_up(jbgp.nb_os * jbgp.nb_oc, 2 * jbgp.nthr)));
@@ -341,9 +345,9 @@ status_t init_ip_conf_fwd(jit_brgemm_primitive_conf_t &jbgp,
     //  * number of threads > 1
     //  * not a "gigantic shape" since it already has a lot of parallelism
     //      in mb and oc dimensions w/o enabling IC parallelism
-    const bool use_parallel_ic_reduction = is_f32 && jbgp.ic > 1024
-            && num_work_to_parallel < jbgp.nb_ic && jbgp.nthr > 1
-            && !is_gigantic_shape;
+    const bool use_parallel_ic_reduction = is_f32 && !jbgp.is_bf32
+            && jbgp.ic > 1024 && num_work_to_parallel < jbgp.nb_ic
+            && jbgp.nthr > 1 && !is_gigantic_shape;
 
     // For os > 256, compute all os blocks as a single chunk when performing
     // IC reduction. Note that this condition is empirical
@@ -352,6 +356,7 @@ status_t init_ip_conf_fwd(jit_brgemm_primitive_conf_t &jbgp,
 
     jbgp.nb_ic_blocking = 1;
     jbgp.nthr_ic_b = 1;
+    const int k_blk = jbgp.is_bf32 ? amx_bf16_row : jbgp.ic_block;
     const int max_nb_ic_blocking = nstl::min(64, jbgp.nb_ic);
     if (IMPLICATION(!is_int8, jbgp.ic <= max_nb_ic_blocking * jbgp.ic_block)
             && everyone_is(1, jbgp.kw, jbgp.kh, jbgp.kd)
@@ -360,8 +365,7 @@ status_t init_ip_conf_fwd(jit_brgemm_primitive_conf_t &jbgp,
         // brgemm kernel with K = ic & batch = 1
         // (K = rnd_dn(ic, ic_block), K_tail = ic % ic_block & batch = 1)
         // instead of K = ic_block & batch = nb_ic_blocking
-        jbgp.K = jbgp.ic <= jbgp.ic_block ? jbgp.ic
-                                          : rnd_dn(jbgp.ic, jbgp.ic_block);
+        jbgp.K = jbgp.ic <= jbgp.ic_block ? jbgp.ic : rnd_dn(jbgp.ic, k_blk);
         jbgp.nb_ic_blocking = jbgp.nb_ic;
         jbgp.gemm_batch_size = 1;
     } else if (!jbgp.use_buffer_a && use_parallel_ic_reduction) {
@@ -381,20 +385,25 @@ status_t init_ip_conf_fwd(jit_brgemm_primitive_conf_t &jbgp,
         jbgp.gemm_batch_size = jbgp.nb_ic_blocking;
         jbgp.K = jbgp.ic_block;
     } else {
-        jbgp.nb_ic_blocking = max_div(jbgp.nb_ic, max_nb_ic_blocking);
-        const bool small_nb_ic = jbgp.nb_ic <= max_nb_ic_blocking;
-        if (small_nb_ic && jbgp.nb_ic_blocking == 1)
-            jbgp.nb_ic_blocking = max_nb_ic_blocking;
+        // Note: Here, ic divided into K_blocks of gemm_batch
+        const int ic_blks_per_k = div_up(k_blk, jbgp.ic_block);
+        const int nb_k_blk = div_up(jbgp.ic, k_blk);
+        const int max_nb_k_blocking = div_up(max_nb_ic_blocking, ic_blks_per_k);
+        int nb_k_blocking = max_div(nb_k_blk, max_nb_k_blocking);
+        const bool small_nb_k_blk = nb_k_blk <= max_nb_k_blocking;
+        if (small_nb_k_blk && nb_k_blocking == 1)
+            nb_k_blocking = max_nb_k_blocking;
 
         // For non small_nb_ic [i.e. that has nb_ic > 64] shape that has
         // gcd(nb_ic, 64) < 16, we manually set nb_ic_blocking = 64
         // the coefficients 64 [used in max_nb_ic_blocking] and 16 are empirical
-        const int min_nb_ic_blocking = small_nb_ic ? 1 : 16;
-        if (jbgp.nb_ic_blocking < min_nb_ic_blocking)
-            jbgp.nb_ic_blocking = max_nb_ic_blocking;
+        const int min_nb_k_blocking = small_nb_k_blk ? 1 : 16;
+        if (nb_k_blocking < min_nb_k_blocking)
+            nb_k_blocking = max_nb_k_blocking;
 
-        jbgp.gemm_batch_size = jbgp.nb_ic_blocking;
-        jbgp.K = jbgp.ic_block;
+        jbgp.nb_ic_blocking = nb_k_blocking * ic_blks_per_k;
+        jbgp.K = k_blk;
+        jbgp.gemm_batch_size = nb_k_blocking;
     }
 
     // to avoid cache concurrent write access from different threads
@@ -402,7 +411,7 @@ status_t init_ip_conf_fwd(jit_brgemm_primitive_conf_t &jbgp,
     jbgp.adjusted_batch_size
             = div_up(rnd_up(jbgp.gemm_batch_size * sc_size, 4096), sc_size);
 
-    if (is_amx_bf16) {
+    if (is_amx_bf16 || jbgp.is_bf32) {
         if (ip_fwd_adjust_thread_balance(jbgp)) {
             // Adjust oc_block to improve thread balancing
             jbgp.oc_block = ip_fwd_get_adjusted_oc_block(jbgp);
@@ -427,7 +436,7 @@ status_t init_ip_conf_fwd(jit_brgemm_primitive_conf_t &jbgp,
 
     jbgp.N = jbgp.oc_block;
     jbgp.N_tail = jbgp.oc % jbgp.oc_block;
-    jbgp.K_tail = jbgp.use_buffer_a ? 0 : jbgp.ic % jbgp.ic_block;
+    jbgp.K_tail = jbgp.use_buffer_a ? 0 : jbgp.ic % jbgp.K;
 
     jbgp.LDA = jbgp.use_buffer_a ? jbgp.K * jbgp.gemm_batch_size
                                  : jbgp.ic_without_padding;
@@ -435,11 +444,23 @@ status_t init_ip_conf_fwd(jit_brgemm_primitive_conf_t &jbgp,
     jbgp.LDD = jbgp.oc_without_padding;
     jbgp.LDC = (jbgp.use_buffer && jbgp.nthr_ic_b == 1) ? jbgp.N : jbgp.LDD;
 
+    if (jbgp.is_bf32) {
+        const float M = static_cast<float>(jbgp.M);
+        const float N = nstl::min<float>(jbgp.N, jbgp.oc);
+        const float K
+                = nstl::min<float>(jbgp.K * jbgp.gemm_batch_size, jbgp.ic);
+        const float tmul_efficiency = (M / 16) * (N / 16) * (K / 32);
+        // TODO: Adjust blocking such that bigger M, N, K are generated.
+        if (one_of(true, M <= 8, K <= 8, N < 16, tmul_efficiency <= 2.25))
+            return status::unimplemented;
+    }
+
     return status::success;
 }
 
 status_t init_ip_conf_bwd_d(jit_brgemm_primitive_conf_t &jbgp) {
-    const bool is_amx_bf16 = jbgp.isa == avx512_core_bf16_amx_bf16;
+    const bool is_amx_bf16
+            = jbgp.isa == avx512_core_bf16_amx_bf16 && !jbgp.is_bf32;
     const bool is_avx512_bf16 = jbgp.isa == avx512_core_bf16;
     const bool is_f32 = everyone_is(f32, jbgp.src_dt, jbgp.wei_dt, jbgp.dst_dt);
     const bool is_bf16 = everyone_is(bf16, jbgp.wei_dt, jbgp.dst_dt);
@@ -457,12 +478,14 @@ status_t init_ip_conf_bwd_d(jit_brgemm_primitive_conf_t &jbgp) {
     //   os <= 128 && max(ic, oc) <= 2048 && min(ic, oc) <= 1000
     //
     // TODO: Will the optimization be useful for bf16 data type
-    const bool avoid_max_ic_block = is_f32 && jbgp.os <= 128
+    const bool avoid_max_ic_block = is_f32 && !jbgp.is_bf32 && jbgp.os <= 128
             && nstl::max(jbgp.ic, jbgp.oc) <= 2048
             && nstl::min(jbgp.ic, jbgp.oc) <= 1000;
-    jbgp.ic_block = !avoid_max_ic_block && jbgp.ic >= (is_f32 ? 512 : 64)
+    jbgp.ic_block = !avoid_max_ic_block
+                    && jbgp.ic >= (is_f32 && !jbgp.is_bf32 ? 512 : 64)
             ? 64
-            : jbgp.ic >= 32 ? 32 : 16;
+            : jbgp.ic >= 32 ? 32
+                            : 16;
 
     jbgp.nb_ic = div_up(jbgp.ic, jbgp.ic_block);
     jbgp.nb_ic_blocking = 1;
@@ -479,7 +502,7 @@ status_t init_ip_conf_bwd_d(jit_brgemm_primitive_conf_t &jbgp) {
             break;
         }
 
-    if (is_amx_bf16) {
+    if (is_amx_bf16 || jbgp.is_bf32) {
         const int os_chunks = div_up(jbgp.nb_os, jbgp.nb_os_blocking);
         const int work_amount = jbgp.nb_ic * os_chunks;
         float wb_ratio = (float)work_amount / (float)jbgp.nthr;
@@ -505,7 +528,7 @@ status_t init_ip_conf_bwd_d(jit_brgemm_primitive_conf_t &jbgp) {
     //   * very large output channels
     //   * small work amount available to each thread
     if ((num_work_to_parallel < 2 * jbgp.nthr
-                || jbgp.oc > (is_bf16 ? 4096 : 1024))) {
+                || jbgp.oc > (is_bf16 || jbgp.is_bf32 ? 4096 : 1024))) {
         const int min_chunck_sz = (is_avx512_bf16) ? 32 : 16;
         const int num_min_chunk_sz = div_up(jbgp.nb_oc, min_chunck_sz);
         float reduce_work = 0.5f * num_min_chunk_sz * jbgp.nb_os
@@ -513,7 +536,9 @@ status_t init_ip_conf_bwd_d(jit_brgemm_primitive_conf_t &jbgp) {
 
         // optimization for transformer_lt on CPX/SKX
         const int max_nthr_oc_b
-                = (!is_amx_bf16 && jbgp.oc > 32000) ? jbgp.nthr / 2 : 4;
+                = (!is_amx_bf16 && !jbgp.is_bf32 && jbgp.oc > 32000)
+                ? jbgp.nthr / 2
+                : 4;
         jbgp.nthr_oc_b = saturate(1, nstl::min(max_nthr_oc_b, num_min_chunk_sz),
                 int(reduce_work));
         jbgp.nthr_oc_b = nstl::min(jbgp.nthr_oc_b, jbgp.nthr);
@@ -545,37 +570,53 @@ status_t init_ip_conf_bwd_d(jit_brgemm_primitive_conf_t &jbgp) {
     jbgp.LDD = jbgp.ic_without_padding;
     jbgp.LDC = jbgp.use_buffer && jbgp.nthr_oc_b == 1 ? jbgp.N : jbgp.LDD;
 
+    if (jbgp.is_bf32) {
+        const float M = static_cast<float>(jbgp.M);
+        const float N = nstl::min<float>(jbgp.N, jbgp.ic);
+        const float K
+                = nstl::min<float>(jbgp.K * jbgp.gemm_batch_size, jbgp.oc);
+        const float tmul_efficiency = (M / 16) * (N / 16) * (K / 32);
+        // TODO: Adjust blocking such that bigger M, N, K are generated.
+        if (one_of(true, M <= 8, K <= 8, N < 16, tmul_efficiency <= 2.25))
+            return status::unimplemented;
+    }
+
     return status::success;
 }
 
 void thread_balance(const jit_brgemm_primitive_conf_t &j, int &nb_os_blocking_,
-        int &nthr_, int &nthr_mb_, int &nthr_oc_b_, int &nthr_ic_b_) {
+        int &nb_oc_blocking_, int &nb_ic_blocking_, int &nthr_, int &nthr_mb_,
+        int &nthr_oc_b_, int &nthr_ic_b_) {
     nthr_ = nthr_mb_ = nthr_oc_b_ = nthr_ic_b_ = 1;
     nb_os_blocking_ = j.nb_os_blocking;
+    nb_oc_blocking_ = j.nb_oc_blocking;
+    nb_ic_blocking_ = j.nb_ic_blocking;
 
     const bool is_f32 = everyone_is(f32, j.src_dt, j.wei_dt, j.dst_dt);
     const bool is_bf16 = everyone_is(bf16, j.src_dt, j.dst_dt);
 
     const int max_threads = j.nthr;
     const int nthr = max_threads;
-    int ic_chunks = j.nb_ic / j.nb_ic_blocking;
-    int oc_chunks = j.nb_oc / j.nb_oc_blocking;
-    auto calc_mem_cost = [=](int nb_os_blocking, int nthr_mb, int nthr_oc,
+    auto calc_mem_cost = [=](int nb_os_blocking, int nb_oc_blocking,
+                                 int nb_ic_blocking, int nthr_mb, int nthr_oc,
                                  int nthr_ic) {
-        int src_size = j.ic * j.mb;
-        int dst_size = j.oc * j.mb;
-        int wei_size = j.ic * j.oc;
+        float src_size = static_cast<float>(j.ic) * j.mb;
+        float dst_size = static_cast<float>(j.oc) * j.mb;
+        float wei_size = static_cast<float>(j.ic) * j.oc;
         int os_chunks = div_up(j.nb_os, nb_os_blocking);
+        int oc_chunks = div_up(j.nb_oc, nb_oc_blocking);
+        int ic_chunks = div_up(j.nb_ic, nb_ic_blocking);
+
         float wei_compensation_scale = 0.5f * (dst_size + src_size) / wei_size;
 
         float oi_channels_ratio = 0;
         if (is_bf16) {
             oi_channels_ratio = ((j.oc > 3 * j.ic && os_chunks > 1)
                                         || (os_chunks == 1 && j.ic > j.oc))
-                    ? (float)src_size / dst_size
-                    : (float)dst_size / src_size;
+                    ? src_size / dst_size
+                    : dst_size / src_size;
         } else {
-            oi_channels_ratio = (float)src_size / dst_size;
+            oi_channels_ratio = src_size / dst_size;
         }
 
         auto get_src_coef = [=]() {
@@ -611,13 +652,18 @@ void thread_balance(const jit_brgemm_primitive_conf_t &j, int &nb_os_blocking_,
                         4.0f - j.mb / 2048 * wei_compensation_scale, 1.0f);
             }
 
-            return nstl::max(wei_compensation_scale, 1.0f);
+            // limit the range of coefficient values to have more stable behavior
+            // for extreme cases
+            const float low_limit = 1.0f;
+            const float upper_limit = 1024.0f;
+            return utils::saturate(
+                    low_limit, upper_limit, wei_compensation_scale);
         };
 
         float src_tr = 0.0f;
         if (j.use_buffer_a && !is_f32) {
             int src_tr_oc_par_work = div_up(os_chunks, nthr_mb)
-                    * div_up(ic_chunks, nthr_ic) * j.nb_ic_blocking;
+                    * div_up(ic_chunks, nthr_ic) * nb_ic_blocking;
             src_tr = get_src_coef() * div_up(src_tr_oc_par_work, nthr_oc)
                     * nb_os_blocking * j.os_block * j.ic_block;
         }
@@ -625,29 +671,29 @@ void thread_balance(const jit_brgemm_primitive_conf_t &j, int &nb_os_blocking_,
         float dst_tr = 0.0f;
         if (j.use_buffer_b && !is_f32) {
             int dst_tr_ic_par_work = div_up(os_chunks, nthr_mb)
-                    * div_up(oc_chunks, nthr_oc) * j.nb_oc_blocking;
+                    * div_up(oc_chunks, nthr_oc) * nb_oc_blocking;
             dst_tr = get_dst_coef() * div_up(dst_tr_ic_par_work, nthr_ic)
                     * nb_os_blocking * j.os_block * j.oc_block;
         }
 
         float src_v = get_src_coef() * div_up(os_chunks, nthr_mb)
                 * div_up(ic_chunks, nthr_ic) * nb_os_blocking * j.os_block
-                * j.nb_ic_blocking * j.ic_block;
+                * nb_ic_blocking * j.ic_block;
         float dst_v = get_dst_coef() * div_up(os_chunks, nthr_mb)
                 * div_up(oc_chunks, nthr_oc) * nb_os_blocking * j.os_block
-                * j.nb_oc_blocking * j.oc_block;
+                * nb_oc_blocking * j.oc_block;
 
         auto acc_dt_sz = types::data_type_size(j.acc_dt);
         float wei_v = get_wei_coef() * acc_dt_sz * div_up(oc_chunks, nthr_oc)
-                * div_up(ic_chunks, nthr_ic) * j.nb_oc_blocking * j.oc_block
-                * j.nb_ic_blocking * j.ic_block;
+                * div_up(ic_chunks, nthr_ic) * nb_oc_blocking * j.oc_block
+                * nb_ic_blocking * j.ic_block;
 
         float wei_r = 0;
         if (nthr_mb > 1) {
             auto wei_dt_sz = types::data_type_size(j.wei_dt);
             int wei_r_mb_par_work = div_up(oc_chunks, nthr_oc)
-                    * div_up(ic_chunks, nthr_ic) * j.nb_oc_blocking
-                    * j.nb_ic_blocking;
+                    * div_up(ic_chunks, nthr_ic) * nb_oc_blocking
+                    * nb_ic_blocking;
             wei_r = get_wei_coef() * div_up(wei_r_mb_par_work, nthr_mb)
                     * j.oc_block * j.ic_block
                     * (wei_dt_sz
@@ -658,8 +704,31 @@ void thread_balance(const jit_brgemm_primitive_conf_t &j, int &nb_os_blocking_,
         return src_tr + dst_tr + src_v + dst_v + wei_v + wei_r;
     };
 
-    float best_mem_cost
-            = calc_mem_cost(nb_os_blocking_, nthr_mb_, nthr_oc_b_, nthr_ic_b_);
+    float best_mem_cost = calc_mem_cost(nb_os_blocking_, nb_oc_blocking_,
+            nb_ic_blocking_, nthr_mb_, nthr_oc_b_, nthr_ic_b_);
+
+    /* Set range of values for nb_oc_blocking/nb_ic_blocking parameters to try.
+       Use powers-of-2 values to avoid potential issues on converting to
+       blocked weights layout stage
+    */
+    auto get_blk_values
+            = [](int max_blk_value, int init_blk, int dim_blk_limit) {
+                  int val_1st = rnd_up_pow2(init_blk);
+                  int val_end = nstl::min(max_blk_value, dim_blk_limit);
+                  std::vector<int> values;
+                  for (int val = val_1st; val <= val_end; val <<= 1)
+                      values.push_back(val);
+                  return values;
+              };
+
+    const int max_nb_oc_blocking_pow
+            = j.ip_bwd_w_local_buffers_for_input_tensors ? 4 : j.nb_oc_blocking;
+    auto nb_oc_blocking_values
+            = get_blk_values(max_nb_oc_blocking_pow, j.nb_oc_blocking, j.nb_oc);
+    const int max_nb_ic_blocking_pow
+            = j.ip_bwd_w_local_buffers_for_input_tensors ? 4 : j.nb_ic_blocking;
+    auto nb_ic_blocking_values
+            = get_blk_values(max_nb_ic_blocking_pow, j.nb_ic_blocking, j.nb_ic);
 
     /* find the best thread distribution with lowest memory cost */
     const int min_osb_chunk = is_f32 ? 32 : is_bf16 ? 8 : 1;
@@ -678,17 +747,26 @@ void thread_balance(const jit_brgemm_primitive_conf_t &j, int &nb_os_blocking_,
         }
 
         const int nthr_par = nthr / nthr_mb;
-        const int nthr_oc_b_max = nstl::min(nthr_par, oc_chunks);
-        for (int nthr_oc_b = 1; nthr_oc_b <= nthr_oc_b_max; ++nthr_oc_b) {
-            int nthr_ic_b = nstl::min(nthr_par / nthr_oc_b, ic_chunks);
-            float mem_cost = calc_mem_cost(
-                    nb_os_blocking, nthr_mb, nthr_oc_b, nthr_ic_b);
-            if (mem_cost <= best_mem_cost) {
-                best_mem_cost = mem_cost;
-                nb_os_blocking_ = nb_os_blocking;
-                nthr_mb_ = nthr_mb;
-                nthr_oc_b_ = nthr_oc_b;
-                nthr_ic_b_ = nthr_ic_b;
+
+        for (auto nb_oc_blocking : nb_oc_blocking_values) {
+            int num_oc_chunks = div_up(j.nb_oc, nb_oc_blocking);
+            const int nthr_oc_b_max = nstl::min(nthr_par, num_oc_chunks);
+            for_(int nthr_oc_b = 1; nthr_oc_b <= nthr_oc_b_max; ++nthr_oc_b)
+            for (auto nb_ic_blocking : nb_ic_blocking_values) {
+                int num_ic_chunks = div_up(j.nb_ic, nb_ic_blocking);
+
+                int nthr_ic_b = nstl::min(nthr_par / nthr_oc_b, num_ic_chunks);
+                float mem_cost = calc_mem_cost(nb_os_blocking, nb_oc_blocking,
+                        nb_ic_blocking, nthr_mb, nthr_oc_b, nthr_ic_b);
+                if (mem_cost <= best_mem_cost) {
+                    best_mem_cost = mem_cost;
+                    nb_os_blocking_ = nb_os_blocking;
+                    nb_oc_blocking_ = nb_oc_blocking;
+                    nb_ic_blocking_ = nb_ic_blocking;
+                    nthr_mb_ = nthr_mb;
+                    nthr_oc_b_ = nthr_oc_b;
+                    nthr_ic_b_ = nthr_ic_b;
+                }
             }
         }
     }
@@ -697,16 +775,17 @@ void thread_balance(const jit_brgemm_primitive_conf_t &j, int &nb_os_blocking_,
 }
 
 status_t init_ip_conf_bwd_w(jit_brgemm_primitive_conf_t &jbgp) {
-    const bool is_amx_bf16 = jbgp.isa == avx512_core_bf16_amx_bf16;
+    const bool is_amx_bf16
+            = jbgp.isa == avx512_core_bf16_amx_bf16 && !jbgp.is_bf32;
     const bool is_f32 = everyone_is(f32, jbgp.src_dt, jbgp.wei_dt, jbgp.dst_dt);
     const bool has_weights_buffer = jbgp.wei_dt != jbgp.acc_dt;
 
     const int amx_bf16_row = 64;
     const bool big_ic_blk_ok
             = is_f32 && jbgp.ic % (4 * jbgp.simd_w) == 0 && jbgp.mb <= 128;
-    jbgp.ic_block = big_ic_blk_ok && !is_amx_bf16
-            ? 4 * jbgp.simd_w
-            : (is_amx_bf16 && has_weights_buffer) ? amx_bf16_row : jbgp.simd_w;
+    jbgp.ic_block = big_ic_blk_ok && !is_amx_bf16 ? 4 * jbgp.simd_w
+            : (is_amx_bf16 && has_weights_buffer) ? amx_bf16_row
+                                                  : jbgp.simd_w;
     jbgp.ic_block_ext = (jbgp.wei_dt == dnnl::impl::data_type::bf16) ? 32 : 16;
 
     jbgp.oc_block = has_weights_buffer ? get_oc_block(jbgp)
@@ -754,16 +833,24 @@ status_t init_ip_conf_bwd_w(jit_brgemm_primitive_conf_t &jbgp) {
     const bool is_oc_big_2_pow = jbgp.oc >= 512 && math::is_pow2(jbgp.oc);
     const bool is_huge_oc = jbgp.oc >= 4 * 1024;
     jbgp.use_buffer_b = jbgp.dst_dt == bf16 || is_oc_big_2_pow || is_huge_oc;
-    jbgp.harness = jbgp.os >= 5 * (jbgp.ic + jbgp.oc) && jbgp.nb_os >= 256
+    const bool os_dim_dominating = jbgp.os >= 5 * (jbgp.ic + jbgp.oc);
+    const int big_nb_os_threshold = is_amx_bf16 ? 64 : 256;
+    jbgp.ip_bwd_w_local_buffers_for_input_tensors
+            = is_amx_bf16 && jbgp.nb_os >= big_nb_os_threshold;
+    jbgp.harness = os_dim_dominating && jbgp.nb_os >= big_nb_os_threshold
             ? harness_mb_reduction
             : harness_2d_reduction;
 
-    int nb_os_blocking, nthr, nthr_mb, nthr_oc, nthr_ic;
+    int nb_os_blocking, nb_oc_blocking, nb_ic_blocking, nthr, nthr_mb, nthr_oc,
+            nthr_ic;
     // Caution: thread_balance requires `use_buffer_a` and `use_buffer_b`
     // fields of jbgp to be properly set
-    thread_balance(jbgp, nb_os_blocking, nthr, nthr_mb, nthr_oc, nthr_ic);
+    thread_balance(jbgp, nb_os_blocking, nb_oc_blocking, nb_ic_blocking, nthr,
+            nthr_mb, nthr_oc, nthr_ic);
 
     jbgp.nb_os_blocking = nb_os_blocking;
+    jbgp.nb_oc_blocking = nb_oc_blocking;
+    jbgp.nb_ic_blocking = nb_ic_blocking;
     jbgp.nthr = nthr;
     jbgp.nthr_mb = nthr_mb;
     jbgp.nthr_oc_b = nthr_oc;
@@ -782,6 +869,17 @@ status_t init_ip_conf_bwd_w(jit_brgemm_primitive_conf_t &jbgp) {
                                    : jbgp.oc_without_padding;
     jbgp.LDC = jbgp.LDD = jbgp.N;
 
+    if (jbgp.is_bf32) {
+        const float M = static_cast<float>(jbgp.M);
+        const float N = nstl::min<float>(jbgp.N, jbgp.oc);
+        const float K
+                = nstl::min<float>(jbgp.K * jbgp.gemm_batch_size, jbgp.os);
+        const float tmul_efficiency = (M / 16) * (N / 16) * (K / 32);
+        // TODO: Adjust blocking such that bigger M, N, K are generated.
+        if (one_of(true, M <= 8, K <= 8, N < 16, tmul_efficiency <= 2.25))
+            return status::unimplemented;
+    }
+
     return status::success;
 }
 
@@ -855,6 +953,8 @@ status_t init_ip_conf(cpu_isa_t isa, jit_brgemm_primitive_conf_t &jbgp,
                     everyone_is(bf16, jbgp.src_dt, jbgp.dst_dt)
                             && jbgp.wei_dt == f32);
     const bool is_f32 = everyone_is(f32, jbgp.src_dt, jbgp.wei_dt, jbgp.dst_dt);
+    jbgp.is_bf32 = is_f32 && attr.fpmath_mode_ == fpmath_mode::bf16
+            && isa == avx512_core_bf16_amx_bf16;
 
     if (!IMPLICATION(is_int8,
                 one_of(isa, avx512_core_vnni, avx512_core_bf16,
@@ -863,7 +963,8 @@ status_t init_ip_conf(cpu_isa_t isa, jit_brgemm_primitive_conf_t &jbgp,
     if (!IMPLICATION(is_bf16,
                 one_of(isa, avx512_core_bf16, avx512_core_bf16_amx_bf16)))
         return status::unimplemented;
-    if (!IMPLICATION(is_f32, isa == avx512_core)) return status::unimplemented;
+    if (!IMPLICATION(is_f32, jbgp.is_bf32 || (isa == avx512_core)))
+        return status::unimplemented;
 
     if (is_int8) {
         jbgp.acc_dt = s32;
@@ -1001,13 +1102,20 @@ void init_scratchpad(memory_tracking::registrar_t &scratchpad,
                 types::data_type_size(jbgp.acc_dt));
     }
     if (jbgp.use_buffer_a && jbgp.prop_kind == dnnl_backward_weights) {
-        const dim_t ic_chunks = div_up(
-                div_up(jbgp.nb_ic, jbgp.nb_ic_blocking), jbgp.nthr_ic_b);
-        const dim_t os_chunks
-                = div_up(div_up(jbgp.nb_os, jbgp.nb_os_blocking), jbgp.nthr_mb);
+        const dim_t num_ic_chunks_per_thread
+                = jbgp.ip_bwd_w_local_buffers_for_input_tensors
+                ? 1
+                : div_up(div_up(jbgp.nb_ic, jbgp.nb_ic_blocking),
+                        jbgp.nthr_ic_b);
+        const dim_t num_os_chunks_per_thread
+                = jbgp.ip_bwd_w_local_buffers_for_input_tensors
+                ? 1
+                : div_up(div_up(jbgp.nb_os, jbgp.nb_os_blocking), jbgp.nthr_mb);
+        const dim_t num_elems_per_thread = num_ic_chunks_per_thread
+                * num_os_chunks_per_thread * jbgp.gemm_batch_size
+                * jbgp.os_block * jbgp.ic_block * jbgp.nb_ic_blocking;
         scratchpad.book(key_brgemm_primitive_buffer_a,
-                jbgp.nthr * ic_chunks * os_chunks * jbgp.gemm_batch_size
-                        * jbgp.os_block * jbgp.ic_block * jbgp.nb_ic_blocking,
+                jbgp.nthr * num_elems_per_thread,
                 types::data_type_size(jbgp.src_dt));
     } else if (jbgp.use_buffer_a && jbgp.prop_kind == dnnl_backward_data) {
         scratchpad.book(key_brgemm_primitive_buffer_a,
@@ -1021,11 +1129,14 @@ void init_scratchpad(memory_tracking::registrar_t &scratchpad,
     }
 
     if (jbgp.use_buffer_b && jbgp.prop_kind == dnnl_backward_weights) {
-        int os_chunks
-                = div_up(div_up(jbgp.nb_os, jbgp.nb_os_blocking), jbgp.nthr_mb);
+        int num_os_chunks_per_thread
+                = jbgp.ip_bwd_w_local_buffers_for_input_tensors
+                ? 1
+                : div_up(div_up(jbgp.nb_os, jbgp.nb_os_blocking), jbgp.nthr_mb);
+        const dim_t num_elems_per_thread = num_os_chunks_per_thread
+                * jbgp.gemm_batch_size * jbgp.os_block * jbgp.LDB;
         scratchpad.book(key_brgemm_primitive_buffer_b,
-                (size_t)jbgp.nthr * os_chunks * jbgp.gemm_batch_size
-                        * jbgp.os_block * jbgp.LDB,
+                (size_t)jbgp.nthr * num_elems_per_thread,
                 types::data_type_size(jbgp.dst_dt));
     }
 
@@ -1056,9 +1167,7 @@ void init_scratchpad(memory_tracking::registrar_t &scratchpad,
     if (jbgp.isa == avx512_core_bf16_amx_int8
             || jbgp.isa == avx512_core_bf16_amx_bf16)
         scratchpad.book(key_conv_amx_tile_buffer,
-                (size_t)jbgp.nthr
-                        * jit_brgemm_primitive_conf_t::tile_wsp_per_thread,
-                sizeof(char));
+                (size_t)jbgp.nthr * jbgp.amx_buf_size_per_thread, sizeof(char));
 }
 
 } // namespace brgemm_inner_product_utils
