@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2020-2022 Intel Corporation
+* Copyright 2020-2023 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -35,6 +35,12 @@ namespace cpu {
 namespace x64 {
 
 struct brgemm_kernel_diff_bias_t {
+    brgemm_kernel_diff_bias_t()
+        : ptr_diff_dst(nullptr)
+        , ptr_diff_bias_acc(nullptr)
+        , ptr_diff_bias(nullptr)
+        , flags(0) {};
+
     void *ptr_diff_dst;
     void *ptr_diff_bias_acc;
     void *ptr_diff_bias;
@@ -46,7 +52,8 @@ struct brgemm_kernel_diff_bias_t {
 struct jit_brgemm_kernel_diff_bias_t : public jit_generator {
     jit_brgemm_kernel_diff_bias_t(
             const jit_brgemm_primitive_conf_t &ajbgp, const brgemm_t &abrg)
-        : brg_(abrg)
+        : jit_generator(jit_name())
+        , brg_(abrg)
         , ddst_dt_(ajbgp.dst_dt)
         , bia_dt_(ajbgp.bia_dt)
         , acc_dt_(ajbgp.acc_dt)
@@ -75,6 +82,7 @@ private:
     const reg64_t aux_reg_ddst = r12;
     const reg64_t reg_k_iter = r11;
     const reg64_t reg_flag = r10;
+    const reg64_t reg_mask = rax;
 
     Xbyak::Opmask k_full_mask = Xbyak::Opmask(2);
     Xbyak::Opmask k_tail_mask = Xbyak::Opmask(3);
@@ -219,7 +227,6 @@ private:
 
         const auto full_mask = size_t {0xffffffffffffffff};
         const auto tail_mask = size_t((1 << nb_tail) - 1);
-        reg64_t reg_mask = rax;
 
         mov(reg_mask, full_mask);
         kmovq(k_full_mask, reg_mask);
@@ -261,7 +268,6 @@ struct brgemm_kernel_post_ops_t {
     void *ptr_bias;
     void *ptr_scales;
     const void *ptr_binary_post_ops_rhs;
-    size_t oc_l_offset;
     size_t apply_comp = 0;
     int32_t a_comp_val = 1;
     int32_t *a_zp_compensation;
@@ -274,32 +280,33 @@ struct jit_brgemm_kernel_post_ops : public jit_generator {
 
     jit_brgemm_kernel_post_ops(const jit_brgemm_conv_conf_t &ajcp,
             const brgemm_t &abrg, const primitive_attr_t &aattr)
-        : brg(abrg)
+        : jit_generator(jit_name())
+        , brg(abrg)
         , jcp(ajcp)
         , attr(aattr)
         , postops_injector_(nullptr)
-        , with_binary_per_oc_bcast_(brg.with_binary
-                  && binary_injector::any_binary_postop_rhs_per_oc_broadcast(
-                          brg.attr->post_ops_,
-                          memory_desc_wrapper(brg.dst_md))) {
+        , with_binary_non_scalar_bcast_(brg.with_binary
+                  && binary_injector::
+                          any_binary_postop_rhs_non_scalar_broadcast(
+                                  brg.attr->post_ops_,
+                                  memory_desc_wrapper(brg.dst_md))) {
 
-        if ((jcp.with_sum && brg.beta != 0)
-                || ((jcp.with_binary || jcp.with_eltwise) && brg.alpha != 0)) {
+        if (brg.beta != 0) {
             static constexpr bool preserve_gpr = true;
             static constexpr bool preserve_vmm = true;
             static constexpr bool use_exact_tail_scalar_bcast = false;
 
             const binary_injector::rhs_arg_static_params_t rhs_sp {
                     static_cast<size_t>(Xbyak::Zmm(28).getIdx()), this->r14,
-                    this->r15, preserve_gpr, preserve_vmm,
+                    this->r15, this->r13, preserve_gpr, preserve_vmm,
                     GET_OFF(ptr_binary_post_ops_rhs), GET_OFF(dst_orig),
                     memory_desc_wrapper(brg.dst_md),
                     static_cast<size_t>(brg.load_dim % brg.ld_block),
                     k_tail_mask, use_exact_tail_scalar_bcast};
             const binary_injector::static_params_t bsp {this->param1, rhs_sp};
 
-            const bool save_state = (brg.alpha != 0) && jcp.with_eltwise;
-            const auto &reserved_eltwise_gpr = rax;
+            const bool save_state = jcp.with_eltwise;
+            const auto &reserved_eltwise_gpr = reg_reserved_eltwise;
             const auto reserved_eltwise_maskr = Xbyak::Opmask(1);
 
             const eltwise_injector::static_params_t esp {
@@ -345,7 +352,7 @@ private:
             postops_injector_;
     std::unique_ptr<bf16_emulation_t> bf16_emu_;
 
-    const bool with_binary_per_oc_bcast_;
+    const bool with_binary_non_scalar_bcast_;
 
     int inp_typesize_;
     int out_typesize_;
@@ -356,6 +363,7 @@ private:
     using reg64_t = const Xbyak::Reg64;
 
     // Register decomposition
+    const reg64_t reg_reserved_eltwise = rax;
     const reg64_t param1 = abi_param1;
     const reg64_t reg_in = r15;
     const reg64_t reg_out = r14;
@@ -371,9 +379,6 @@ private:
     const reg64_t reg_ptr_sum_scale = rdx;
     const reg64_t reg_ptr_sum_zp = rsi;
 
-    const reg64_t reg_oc_l_offset_ = abi_not_param1;
-    const reg64_t aux_reg_oc_l_offset_ = rbx;
-
     const reg64_t reg_zp_c_values = rbx;
     const reg64_t aux_reg_zp_c_values = rbx;
     const reg64_t reg_zp_a_comp = rbx;
@@ -382,24 +387,24 @@ private:
     const reg64_t aux_reg_s8s8_comp = rbx;
     const reg64_t reg_zp_a_val = rbx;
     const reg64_t reg_apply_comp = rbx;
+    const reg64_t reg_tmp = abi_not_param1;
 
-    constexpr static int reg_aux_oc_l_offset_offs_ = 0;
-    constexpr static int reg_zp_c_values_offs_ = 8;
-    constexpr static int aux_reg_zp_c_values_offs_ = 16;
-    constexpr static int reg_zp_a_comp_offs_ = 24;
-    constexpr static int aux_reg_zp_a_comp_offs_ = 32;
-    constexpr static int reg_s8s8_comp_offs_ = 40;
-    constexpr static int aux_reg_s8s8_comp_offs_ = 48;
-    constexpr static int reg_zp_a_val_offs_ = 56;
-    constexpr static int reg_apply_comp_offs_ = 64;
-    constexpr static int stack_space_needed_ = 72;
+    constexpr static int reg_zp_c_values_offs_ = 0;
+    constexpr static int aux_reg_zp_c_values_offs_ = 8;
+    constexpr static int reg_zp_a_comp_offs_ = 16;
+    constexpr static int aux_reg_zp_a_comp_offs_ = 24;
+    constexpr static int reg_s8s8_comp_offs_ = 32;
+    constexpr static int aux_reg_s8s8_comp_offs_ = 40;
+    constexpr static int reg_zp_a_val_offs_ = 48;
+    constexpr static int reg_apply_comp_offs_ = 56;
+    constexpr static int stack_space_needed_ = 64;
 
     /* bf16 emulation */
     Xbyak::Zmm bf16_emu_reserv_1 = Xbyak::Zmm(27);
     Xbyak::Zmm bf16_emu_reserv_2 = Xbyak::Zmm(24);
     Xbyak::Zmm bf16_emu_reserv_3 = Xbyak::Zmm(25);
     Xbyak::Zmm bf16_emu_reserv_4 = Xbyak::Zmm(26);
-    reg64_t bf16_emu_scratch = rax;
+    reg64_t bf16_emu_scratch = reg_tmp;
 
     Xbyak::Opmask k_full_mask = Xbyak::Opmask(2);
     Xbyak::Opmask k_tail_mask = Xbyak::Opmask(3);
@@ -414,16 +419,16 @@ private:
 
         return 0;
     }
-    int zp_comp_a_vpad_offset(int n, int m, bool is_tail = false) const
-            noexcept {
+    int zp_comp_a_vpad_offset(
+            int n, int m, bool is_tail = false) const noexcept {
         return (is_tail) ? sizeof(int32_t) * (brg.ldb_tail + m * brg.LDB)
                          : sizeof(int32_t) * (n * brg.ld_block + m * brg.LDB);
     }
     int mb_zp_comp_a_offset(int m_block) const noexcept {
         return sizeof(int32_t) * m_block * brg.LDB;
     }
-    int compensation_vpad_offset(int n, int m, bool is_tail = false) const
-            noexcept {
+    int compensation_vpad_offset(
+            int n, int m, bool is_tail = false) const noexcept {
         return (is_tail) ? sizeof(int32_t) * (brg.ldb_tail + m * brg.LDB)
                          : sizeof(int32_t) * (n * brg.ld_block + m * brg.LDB);
     }
@@ -445,9 +450,11 @@ private:
                 : ymm_in;
     }
 
+    // maybe_req_comp == true -> convert integral values to s32, not f32
+    // (compensation should be applied in s32 to avoid accuracy issues)
     void cvt2ps(data_type_t type_in, const Xbyak::Zmm zmm_in,
             const Xbyak::Operand &op, bool mask_flag, bool store,
-            Xbyak::Opmask ktail_mask) {
+            Xbyak::Opmask ktail_mask, bool maybe_req_comp = false) {
         const Xbyak::Zmm zmm = zmm_mask(zmm_in, mask_flag, store, ktail_mask);
         switch (type_in) {
             case data_type::f32:
@@ -460,7 +467,7 @@ private:
                 break;
             default: assert(!"unsupported data type");
         }
-        if (!utils::one_of(type_in, data_type::f32, data_type::bf16))
+        if (!maybe_req_comp && types::is_integral_dt(type_in))
             vcvtdq2ps(zmm_in, zmm_in);
     }
 
@@ -468,21 +475,6 @@ private:
         return Xbyak::Zmm(m * n_block + n);
     };
 
-    void advance_mb_post_ops_regs(int m_block) {
-        if (brg.alpha != 0) {
-            if (brg.zp_type_a != brgemm_broadcast_t::none) {
-                mov(reg_zp_a_comp, ptr[rsp + reg_zp_a_comp_offs_]);
-                add(reg_zp_a_comp, mb_zp_comp_a_offset(m_block));
-                mov(ptr[rsp + reg_zp_a_comp_offs_], reg_zp_a_comp);
-            }
-            if (brg.req_s8s8_compensation) {
-                mov(reg_s8s8_comp, ptr[rsp + reg_s8s8_comp_offs_]);
-                add(reg_s8s8_comp, mb_compensation_offset(m_block));
-                mov(ptr[rsp + reg_s8s8_comp_offs_], reg_s8s8_comp);
-            }
-        }
-    }
-
     void inject_attr_postops(int m_block, int n_block, int tail = 0) {
         const auto &p = attr.post_ops_;
         const int sum_idx = p.find(primitive_kind::sum);
@@ -516,14 +508,14 @@ private:
             }
         };
 
-        if (jcp.with_sum && brg.beta != 0) {
+        if (jcp.with_sum) {
             postops_injector_->set_lambda_injector(
                     primitive_kind::sum, sum_injector);
         }
 
         binary_injector::rhs_arg_dynamic_params_t rhs_arg_params;
 
-        if (with_binary_per_oc_bcast_) {
+        if (with_binary_non_scalar_bcast_) {
             for_(int m = 0; m < m_block; m++)
             for (int n = 0; n < n_block; n++) {
                 const auto zmm_idx = vector(m, n, n_block).getIdx();
@@ -544,7 +536,7 @@ private:
     void apply_comp(int m_block, int n_block, int tail = 0) {
         auto k_mask = (tail == 0) ? k_full_mask : k_tail_mask;
 
-        if (brg.alpha != 0 && brg.zp_type_a != brgemm_broadcast_t::none) {
+        if (brg.zp_type_a != brgemm_broadcast_t::none) {
             auto zmm_zp_a_val = Xbyak::Zmm(30);
             mov(reg_zp_a_val, ptr[rsp + reg_zp_a_val_offs_]);
             vpbroadcastd(zmm_zp_a_val, reg_zp_a_val.cvt32());
@@ -559,20 +551,13 @@ private:
                 vpmulld(zmm_zp_comp_a, zmm_zp_a_val, zp_comp_a_addr);
 
                 for (int m = 0; m < m_block; m++) {
-                    if (brg.with_comp_pads) {
-                        auto zp_comp_a_vpad_offs = zp_comp_a_vpad_offset(n, m);
-                        auto zp_comp_a_vpad_addr = zword[aux_reg_zp_a_comp
-                                + zp_comp_a_vpad_offs];
-                        vmovups(zmm_zp_comp_a, zp_comp_a_vpad_addr);
-                        vpmulld(zmm_zp_comp_a, zmm_zp_comp_a, zmm_zp_a_val);
-                    }
                     auto zmm = vector(m, n, n_block);
                     vpaddd(zmm, zmm, zmm_zp_comp_a);
                 }
             }
         }
 
-        if (brg.alpha != 0 && brg.req_s8s8_compensation) {
+        if (brg.req_s8s8_compensation) {
             mov(aux_reg_s8s8_comp, ptr[rsp + aux_reg_s8s8_comp_offs_]);
             for (int n = 0; n < n_block; n++) {
                 auto zmm_comp = Xbyak::Zmm(31);
@@ -582,12 +567,6 @@ private:
                 vmovups(zmm_comp, comp_addr);
 
                 for (int m = 0; m < m_block; m++) {
-                    if (brg.with_comp_pads) {
-                        auto comp_vpad_offs = compensation_vpad_offset(n, m);
-                        auto comp_vpad_addr
-                                = zword[aux_reg_s8s8_comp + comp_vpad_offs];
-                        vmovups(zmm_comp, comp_vpad_addr);
-                    }
                     auto zmm = vector(m, n, n_block);
                     vpaddd(zmm, zmm, zmm_comp);
                 }
@@ -613,53 +592,42 @@ private:
         const auto vector
                 = [=](int m, int n) { return Xbyak::Zmm(m * n_block + n); };
         auto k_mask = (tail == 0) ? k_full_mask : k_tail_mask;
-        const auto &p = attr.post_ops_;
-        const int sum_idx = p.find(primitive_kind::sum);
-        const auto maybe_req_comp = brg.is_int8 && brg.alpha != 0
+        const auto maybe_req_comp = brg.is_int8 && brg.beta != 0
                 && (brg.req_s8s8_compensation
                         || brg.zp_type_a != brgemm_broadcast_t::none);
 
-        // brg.alpha == 0 means no read from input, no bias, no eltwise - just
-        // initialize registers by zero at the beginning of kernel
-        // brg.beta == 0 means no sum - just registers write to output
+        // brg.alpha == 0 means initialize registers, 1 means read from input
+        // brg.beta == 0 means skip postwork, 1 means do postwork
         // maybe_req_comp == true -> convert accumulated values to f32 after apply
         // compensation to avoid the lost of accuracy when converting s32 to f32
         for_(int m = 0; m < m_block; m++)
         for (int n = 0; n < n_block; n++) {
-            if (brg.alpha == 0) {
-                if (sum_idx != -1 && brg.beta != 0) {
-                    // if sum then have to init zmm each time
-                    vpxord(vector(m, n), vector(m, n), vector(m, n));
-                }
-            } else {
+            if (brg.alpha == 0 && brg.beta != 0) {
+                // if postwork then have to init vmm each time
+                uni_vpxor(vector(m, n), vector(m, n), vector(m, n));
+            } else if (brg.alpha != 0) {
                 auto inp_addr = ptr[aux_reg_in
                         + inp_typesize_ * (m * brg.LDC + n * brg.ld_block)];
-                if (maybe_req_comp) {
-                    const Xbyak::Zmm zmm
-                            = zmm_mask(vector(m, n), true, false, k_mask);
-                    vmovups(zmm, inp_addr);
-                } else {
-                    cvt2ps(inp_dt_, vector(m, n), inp_addr, true, false,
-                            k_mask);
-                }
+                cvt2ps(inp_dt_, vector(m, n), inp_addr, true, false, k_mask,
+                        maybe_req_comp);
             }
         }
 
         if (maybe_req_comp) maybe_apply_comp(m_block, n_block, tail);
 
-        if (brg.alpha != 0 && jcp.with_bias) {
-            for_(int m = 0; m < m_block; m++)
+        if (brg.beta != 0 && jcp.with_bias) {
             for (int n = 0; n < n_block; n++) {
                 auto zmm_bias = Xbyak::Zmm(31);
                 auto bias_addr = ptr[aux_reg_bias
                         + bia_typesize_ * (n * brg.ld_block)];
-
                 cvt2ps(bia_dt_, zmm_bias, bias_addr, true, false, k_mask);
-                vaddps(vector(m, n), zmm_bias);
+                for (int m = 0; m < m_block; m++) {
+                    vaddps(vector(m, n), zmm_bias);
+                }
             }
         }
 
-        if (brg.alpha != 0) {
+        if (brg.beta != 0) {
             for_(int m = 0; m < m_block; m++)
             for (int n = 0; n < n_block; n++) {
                 const Xbyak::Zmm zmm
@@ -673,7 +641,7 @@ private:
 
         if (postops_injector_) inject_attr_postops(m_block, n_block, tail);
 
-        if (brg.alpha != 0 && brg.zp_type_c != brgemm_broadcast_t::none) {
+        if (brg.beta != 0 && brg.zp_type_c != brgemm_broadcast_t::none) {
             mov(aux_reg_zp_c_values, ptr[rsp + aux_reg_zp_c_values_offs_]);
             auto zmm_zp_c = Xbyak::Zmm(31);
             if (brg.zp_type_c == brgemm_broadcast_t::per_tensor) {
@@ -696,7 +664,7 @@ private:
         const bool dt_requires_saturation = utils::one_of(
                 brg.dt_d, data_type::u8, data_type::s8, data_type::s32);
 
-        const reg64_t reg_tmp_gpr = rax;
+        const reg64_t reg_tmp_gpr = reg_tmp;
         auto zmm_lbound = Xbyak::Zmm(31);
         auto zmm_ubound = Xbyak::Zmm(30);
         if (dt_requires_saturation) {
@@ -714,7 +682,7 @@ private:
 
             if (out_dt_ == data_type::bf16) {
                 Xbyak::Ymm ymm = Xbyak::Ymm(zmm.getIdx());
-                if (brg.alpha != 0 || (sum_idx != -1 && brg.beta != 0)) {
+                if (brg.beta != 0) {
                     if (brg.is_bf16_emu)
                         bf16_emu_->vcvtneps2bf16(ymm, zmm);
                     else
@@ -723,7 +691,7 @@ private:
                 const Xbyak::Ymm r_ymm = ymm_mask(ymm, true, true, k_mask);
                 vmovdqu16(addr, r_ymm);
             } else {
-                if (brg.alpha != 0 || (sum_idx != -1 && brg.beta != 0)) {
+                if (brg.beta != 0) {
                     saturate_f32(zmm, zmm_lbound, zmm_ubound, brg.dt_d);
                     if (out_dt_ != data_type::f32) vcvtps2dq(zmm, zmm);
                 }
@@ -742,11 +710,9 @@ private:
 
     void loop_by_N(int m_block, int nb2, int nb2_tail, int nb_tail) {
 
-        if (brg.alpha) {
-            mov(aux_reg_in, reg_in);
+        if (brg.alpha) { mov(aux_reg_in, reg_in); }
+        if (brg.beta != 0) {
             if (jcp.with_bias) mov(aux_reg_bias, reg_bias);
-            if (with_binary_per_oc_bcast_)
-                mov(aux_reg_oc_l_offset_, reg_oc_l_offset_);
             if (brg.zp_type_c != brgemm_broadcast_t::none) {
                 mov(aux_reg_zp_c_values, ptr[rsp + reg_zp_c_values_offs_]);
                 mov(ptr[rsp + aux_reg_zp_c_values_offs_], aux_reg_zp_c_values);
@@ -771,7 +737,8 @@ private:
             add(aux_reg_out, out_typesize_ * oc_l_offset);
             if (brg.alpha != 0) {
                 add(aux_reg_in, inp_typesize_ * oc_l_offset);
-
+            }
+            if (brg.beta != 0) {
                 if (jcp.with_bias)
                     add(aux_reg_bias, bia_typesize_ * oc_l_offset);
                 if (brg.zp_type_c != brgemm_broadcast_t::none) {
@@ -791,13 +758,6 @@ private:
                     add(aux_reg_s8s8_comp, sizeof(int32_t) * oc_l_offset);
                     mov(ptr[rsp + aux_reg_s8s8_comp_offs_], aux_reg_s8s8_comp);
                 }
-                if (with_binary_per_oc_bcast_) {
-                    mov(aux_reg_oc_l_offset_,
-                            ptr[rsp + reg_aux_oc_l_offset_offs_]);
-                    add(aux_reg_oc_l_offset_, oc_l_offset);
-                    mov(ptr[rsp + reg_aux_oc_l_offset_offs_],
-                            aux_reg_oc_l_offset_);
-                }
 
                 add(aux_reg_scales, is_oc_scale_ * sizeof(float) * oc_l_offset);
             }
@@ -809,6 +769,8 @@ private:
             add(aux_reg_out, out_typesize_ * oc_l_offset);
             if (brg.alpha != 0) {
                 add(aux_reg_in, inp_typesize_ * oc_l_offset);
+            }
+            if (brg.beta != 0) {
                 if (jcp.with_bias)
                     add(aux_reg_bias, bia_typesize_ * oc_l_offset);
                 if (brg.zp_type_c != brgemm_broadcast_t::none) {
@@ -828,13 +790,6 @@ private:
                     add(aux_reg_s8s8_comp, sizeof(int32_t) * oc_l_offset);
                     mov(ptr[rsp + aux_reg_s8s8_comp_offs_], aux_reg_s8s8_comp);
                 }
-                if (with_binary_per_oc_bcast_) {
-                    mov(aux_reg_oc_l_offset_,
-                            ptr[rsp + reg_aux_oc_l_offset_offs_]);
-                    add(aux_reg_oc_l_offset_, oc_l_offset);
-                    mov(ptr[rsp + reg_aux_oc_l_offset_offs_],
-                            aux_reg_oc_l_offset_);
-                }
 
                 add(aux_reg_scales, is_oc_scale_ * sizeof(float) * oc_l_offset);
             }
@@ -842,8 +797,8 @@ private:
         if (nb_tail > 0) {
             apply_post_ops(m_block, 1, nb_tail);
 
-            if (brg.alpha != 0) {
-                add(aux_reg_in, inp_typesize_ * (nb_tail));
+            if (brg.alpha != 0) { add(aux_reg_in, inp_typesize_ * (nb_tail)); }
+            if (brg.beta != 0) {
                 if (jcp.with_bias) add(aux_reg_bias, bia_typesize_ * (nb_tail));
                 if (brg.zp_type_c != brgemm_broadcast_t::none) {
                     mov(aux_reg_zp_c_values,
@@ -862,13 +817,6 @@ private:
                     add(aux_reg_s8s8_comp, sizeof(int32_t) * nb_tail);
                     mov(ptr[rsp + aux_reg_s8s8_comp_offs_], aux_reg_s8s8_comp);
                 }
-                if (with_binary_per_oc_bcast_) {
-                    mov(aux_reg_oc_l_offset_,
-                            ptr[rsp + reg_aux_oc_l_offset_offs_]);
-                    add(aux_reg_oc_l_offset_, nb_tail);
-                    mov(ptr[rsp + reg_aux_oc_l_offset_offs_],
-                            aux_reg_oc_l_offset_);
-                }
                 add(aux_reg_scales, is_oc_scale_ * bia_typesize_ * (nb_tail));
             }
             add(aux_reg_out, out_typesize_ * (nb_tail));
@@ -896,15 +844,15 @@ private:
         const auto full_mask = size_t {0xffffffffffffffff};
         const auto tail_mask = size_t((1 << nb_tail) - 1);
 
-        reg64_t reg_mask = rax;
+        reg64_t reg_mask = reg_tmp;
 
         mov(reg_mask, full_mask);
         kmovq(k_full_mask, reg_mask);
         mov(reg_mask, tail_mask);
         kmovq(k_tail_mask, reg_mask);
 
-        if (brg.alpha != 0) {
-            mov(reg_in, ptr[param1 + GET_OFF(ptr_in)]);
+        if (brg.alpha != 0) { mov(reg_in, ptr[param1 + GET_OFF(ptr_in)]); }
+        if (brg.beta != 0) {
             mov(reg_scales, ptr[param1 + GET_OFF(ptr_scales)]);
             mov(reg_apply_comp, ptr[param1 + GET_OFF(apply_comp)]);
             mov(ptr[rsp + reg_apply_comp_offs_], reg_apply_comp);
@@ -925,15 +873,12 @@ private:
                 mov(reg_s8s8_comp, ptr[param1 + GET_OFF(s8s8_compensation)]);
                 mov(ptr[rsp + reg_s8s8_comp_offs_], reg_s8s8_comp);
             }
-            if (with_binary_per_oc_bcast_)
-                mov(reg_oc_l_offset_, ptr[param1 + GET_OFF(oc_l_offset)]);
         }
         mov(reg_out, ptr[param1 + GET_OFF(ptr_out)]);
 
-        // brg.alpha == 0 means no read from input, no bias, no eltwise - just
-        // initialize registers by zero
-        // brg.beta == 0 means no sum - just registers write to output
-        if (brg.alpha == 0) {
+        // brg.alpha == 0 means initialize registers, 1 means read from input
+        // brg.beta == 0 means skip postwork, 1 means do postwork
+        if (brg.alpha == 0 && brg.beta == 0) {
             for_(int m = 0; m < m_block; m++)
             for (int n = 0; n < n_block; n++) {
                 auto zmm = Xbyak::Zmm(m * n_block + n);
@@ -946,7 +891,6 @@ private:
 
             if (brg.alpha != 0)
                 add(reg_in, inp_typesize_ * (m_block * brg.LDC));
-            advance_mb_post_ops_regs(m_block);
             add(reg_out, out_typesize_ * (m_block * LDD_));
         }
         if (mb_tail > 0) loop_by_N(mb_tail, nb2, nb2_tail, nb_tail);
@@ -955,8 +899,7 @@ private:
 
         postamble();
 
-        if (brg.alpha != 0 && jcp.with_eltwise)
-            postops_injector_->prepare_table();
+        if (postops_injector_) postops_injector_->prepare_table();
     }
 };
 
