@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2016-2022 Intel Corporation
+* Copyright 2016-2023 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -662,6 +662,7 @@ struct jit_pool_conf_t {
 
     int dt_size;
     bool is_bf16;
+    bool is_f16;
     jit_memory_tag_kind_t tag_kind;
     bool is_plain() const {
         return (tag_kind == jit_memory_tag_kind_t::ncsp
@@ -824,8 +825,8 @@ struct jit_brgemm_conv_conf_t {
     int mb;
     int ngroups, ic, oc, oc_without_padding, ic_without_padding;
 
-    int od_block, oh_block, nb_od,
-            nb_oh; // blocking  - included in parallelization
+    int od_block, oh_block, nb_od, nb_oh,
+            ih_block; // blocking  - included in parallelization
     dim_t inp_buffer_size, inp_buffer_mask_size;
     conv_brgemm_exec_type_t exec_type;
 
@@ -886,6 +887,7 @@ struct jit_brgemm_conv_conf_t {
 
     int max_batch;
     int max_vpad;
+    int amx_buf_size_per_thread;
 
     bool wei_plain;
     bool is_ic_padded;
@@ -897,13 +899,33 @@ struct jit_brgemm_conv_conf_t {
     bool brgemm_bd_loop_innermost;
 
     bool use_uker;
+    bool var_bs {false};
     bool use_interleave_stores;
     brgemm_kernel_prefetching_t hint_prefetching;
     bool is_1x1;
     bool s8s8_avx512;
     bool src_zero_point;
     bool dst_zero_point;
+    bool req_brg_comp_pad;
+    bool req_cal_comp_pad;
+    bool is_bf32;
     bool comp_with_vpads;
+
+    int nthr_mb, nthr_g, nthr_oc_b, nthr_ic_b, nthr_oh;
+    bool transform_to_vnni;
+    bool has_vnni;
+    int ic_tail, oc_tail;
+    size_t tr_src_buf_size, tr_src_buf_count;
+    size_t tr_diff_dst_buf_size, tr_diff_dst_buf_count;
+    int tr_src_num_guard_elems;
+    bool global_transpose; // diff_dst & src tensors are transposed in one go
+    int nthr_mb_work;
+    int tr_iw, tr_ow;
+    int spatial_blk_size; // Height/depth block size inside the driver
+    int typesize_in;
+    int typesize_out;
+    bool tr_ocb_chunk = false;
+    bool tr_icb_chunk = false;
 };
 
 struct jit_shuffle_conf_t {
@@ -965,10 +987,12 @@ struct jit_binary_conf_t {
     bool postops_per_oc_broadcast_exists = false;
     bool is_i8 = false;
     bool is_bf16 = false;
+    bool is_f16 = false;
     bool is_src_different_layouts = false;
     dim_t outer_dims = 1;
     int src1_stride = 1;
     int not_bcasted_sp_dims = 0;
+    cpu_isa_t isa = isa_any;
 
     data_type_t src0_type = data_type::undef;
     data_type_t src1_type = data_type::undef;
