@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2020-2022 Intel Corporation
+* Copyright 2020-2023 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -367,7 +367,9 @@ void jit_uni_x8s8s32x_deconv_fwd_kernel<isa>::init_scratchpad(
         memory_tracking::registrar_t &scratchpad, const jit_conv_conf_t &jcp,
         const primitive_attr_t &attr) {
     if (jcp.signed_input && (!jcp.has_vnni)) {
-        dim_t count = nstl::max<dim_t>(attr.output_scales_.count_, 8);
+        const int mask = attr.output_scales_.mask_;
+        const dim_t scales_count = mask == 0 ? 1 : jcp.oc * jcp.ngroups;
+        dim_t count = nstl::max<dim_t>(scales_count, 8);
         scratchpad.book<float>(key_conv_adjusted_scales, count);
     }
 
@@ -395,9 +397,10 @@ template <cpu_isa_t isa, typename Vmm>
 _jit_uni_x8s8s32x_deconv_fwd_kernel<isa,
         Vmm>::_jit_uni_x8s8s32x_deconv_fwd_kernel(const jit_conv_conf_t &ajcp,
         const primitive_attr_t &attr, const memory_desc_wrapper &dst_d)
-    : jit_generator(nullptr, MAX_CODE_SIZE, true, isa)
+    : jit_generator(jit_name(), nullptr, MAX_CODE_SIZE, true, isa)
     , jcp_(ajcp)
-    , postops_injector_(nullptr) {
+    , postops_injector_(nullptr)
+    , ker_max_regs_(jcp_.has_vnni ? 14 : 12) {
 
     if (jcp_.with_eltwise || jcp_.with_binary || jcp_.with_sum) {
         const std::size_t tail_size = get_tail_size();
@@ -408,7 +411,7 @@ _jit_uni_x8s8s32x_deconv_fwd_kernel<isa,
         static constexpr size_t vmm_helper_idx = 15;
 
         const binary_injector::rhs_arg_static_params_t rhs_sp {vmm_helper_idx,
-                this->r14, this->r15, preserve_gpr, preserve_vmm,
+                this->r14, this->r15, this->r13, preserve_gpr, preserve_vmm,
                 GET_OFF(post_ops_binary_rhs_arg_vec), GET_OFF(dst_orig), dst_d,
                 tail_size, Xbyak::Opmask(2), use_exact_tail_scalar_bcast};
         const binary_injector::static_params_t bsp {this->param1_, rhs_sp};
@@ -428,7 +431,7 @@ template <cpu_isa_t isa, typename Vmm>
 Vmm _jit_uni_x8s8s32x_deconv_fwd_kernel<isa, Vmm>::vmm_out(
         int i_ur, int i_oc) const {
     const int idx = i_ur * jcp_.nb_oc_blocking + i_oc;
-    assert(idx < KER_MAX_REG_IDX);
+    assert(idx < ker_max_regs_);
     /* remap the reg indices to avoid using xmm0 in eltwise injector */
     return Vmm(15 - idx);
 }
@@ -437,7 +440,7 @@ template <cpu_isa_t isa, typename Vmm>
 Vmm _jit_uni_x8s8s32x_deconv_fwd_kernel<isa, Vmm>::vmm_inp(
         int i_ic, int nb_x_blocking) const {
     const int idx = i_ic + nb_x_blocking * jcp_.ur_w;
-    assert(idx < KER_MAX_REG_IDX);
+    assert(idx < ker_max_regs_);
     return Vmm(15 - idx);
 }
 
@@ -475,14 +478,14 @@ int _jit_uni_x8s8s32x_deconv_fwd_kernel<isa, Vmm>::get_ow_end(
 }
 
 template <cpu_isa_t isa, typename Vmm>
-int _jit_uni_x8s8s32x_deconv_fwd_kernel<isa, Vmm>::get_blocking_size() const
-        noexcept {
+int _jit_uni_x8s8s32x_deconv_fwd_kernel<isa, Vmm>::get_blocking_size()
+        const noexcept {
     return jcp_.is_depthwise ? jcp_.ch_block : jcp_.oc_block;
 }
 
 template <cpu_isa_t isa, typename Vmm>
-int _jit_uni_x8s8s32x_deconv_fwd_kernel<isa, Vmm>::get_tail_size() const
-        noexcept {
+int _jit_uni_x8s8s32x_deconv_fwd_kernel<isa, Vmm>::get_tail_size()
+        const noexcept {
     return jcp_.is_depthwise ? jcp_.ngroups % jcp_.ch_block
                              : jcp_.oc_without_padding % jcp_.oc_block;
 }
@@ -1117,7 +1120,12 @@ void _jit_uni_x8s8s32x_deconv_fwd_kernel<isa, Vmm>::store_output(
         }
 
         /* add to ymm_accum: compensation, bias */
-        uni_vmovups(vmm_scale_, ptr[reg_ptr_scales_ + scale_offset]);
+        if (mask_flag) {
+            load_data(data_type::f32, vmm_scale_, reg_ptr_scales_, scale_offset,
+                    get_tail_size());
+        } else {
+            uni_vmovups(vmm_scale_, ptr[reg_ptr_scales_ + scale_offset]);
+        }
         for (int ur = 0; ur < ur_w; ur++) {
             const Vmm vmm = vmm_out(ur, ocb);
             uni_vcvtdq2ps(vmm, vmm);
@@ -1382,7 +1390,7 @@ status_t jit_uni_x8s8s32x_deconvolution_fwd_t<isa>::pd_t::init(
                     utils::one_of(weights_md(1)->data_type, f32, s32, s8, u8))
             && utils::one_of(dst_md(0)->data_type, f32, s32, s8, u8)
             && desc()->accum_data_type == s32
-            && attr()->has_default_values(skip_mask_t::oscale
+            && attr()->has_default_values(skip_mask_t::oscale_runtime
                     | skip_mask_t::post_ops | skip_mask_t::zero_points_runtime);
     if (!ok) return status::unimplemented;
 
@@ -1413,6 +1421,22 @@ status_t jit_uni_x8s8s32x_deconvolution_fwd_t<isa>::init(engine_t *engine) {
     return kernel_->create_kernel();
 }
 
+template <cpu_isa_t isa>
+const float *jit_uni_x8s8s32x_deconvolution_fwd_t<isa>::adjust_oscales(
+        const memory_tracking::grantor_t &scratchpad,
+        const float *oscales) const {
+    auto loc_scales = scratchpad.template get<float>(key_conv_adjusted_scales);
+    const int mask = pd()->attr()->output_scales_.mask_;
+    const float factor = 1.f / pd()->jcp_.wei_adj_scale;
+    if (mask == 0) {
+        utils::array_set(loc_scales, oscales[0] * factor, 8);
+    } else {
+        for (dim_t c = 0; c < pd()->OC(); c++)
+            loc_scales[c] = oscales[c] * factor;
+    }
+    return loc_scales;
+}
+
 template <cpu_isa_t isa>
 status_t jit_uni_x8s8s32x_deconvolution_fwd_t<isa>::execute(
         const exec_ctx_t &ctx) const {
@@ -1461,20 +1485,11 @@ status_t jit_uni_x8s8s32x_deconvolution_fwd_t<isa>::execute_forward_1d(
     const int oc_chunks = jcp.nb_oc / jcp.nb_oc_blocking;
     const int nb_groups = jcp.nb_ch;
 
-    const float *oscales = pd()->attr()->output_scales_.scales_;
-    if (jcp.signed_input && (!jcp.has_vnni)) {
-        auto local_scales = ctx.get_scratchpad_grantor().template get<float>(
-                key_conv_adjusted_scales);
-        const size_t count = pd()->attr()->output_scales_.count_;
-        const float factor = 1.f / pd()->jcp_.wei_adj_scale;
-        if (count == 1) {
-            utils::array_set(local_scales, oscales[0] * factor, 8);
-        } else {
-            for (size_t c = 0; c < count; c++)
-                local_scales[c] = oscales[c] * factor;
-        }
-        oscales = local_scales;
-    }
+    DEFINE_SCALES_BUFFER(oscales);
+
+    if (jcp.signed_input && (!jcp.has_vnni))
+        oscales = adjust_oscales(ctx.get_scratchpad_grantor(), oscales);
+
     const size_t offset = weights_d.size() - weights_d.additional_buffer_size();
     auto w = const_cast<int8_t *>(weights);
     int32_t *compensation = (jcp.signed_input)
@@ -1579,20 +1594,11 @@ status_t jit_uni_x8s8s32x_deconvolution_fwd_t<isa>::execute_forward_2d(
     const size_t dst_h_stride = dst_d.blk_off(0, 0, 1);
     const size_t wht_kh_stride = wht_blk_off(weights_d, 0, 0, 0, 1);
 
-    const float *oscales = pd()->attr()->output_scales_.scales_;
-    if (jcp.signed_input && (!jcp.has_vnni)) {
-        auto local_scales = ctx.get_scratchpad_grantor().template get<float>(
-                key_conv_adjusted_scales);
-        const size_t count = pd()->attr()->output_scales_.count_;
-        const float factor = 1.f / pd()->jcp_.wei_adj_scale;
-        if (count == 1) {
-            utils::array_set(local_scales, oscales[0] * factor, 8);
-        } else {
-            for (size_t c = 0; c < count; c++)
-                local_scales[c] = oscales[c] * factor;
-        }
-        oscales = local_scales;
-    }
+    DEFINE_SCALES_BUFFER(oscales);
+
+    if (jcp.signed_input && (!jcp.has_vnni))
+        oscales = adjust_oscales(ctx.get_scratchpad_grantor(), oscales);
+
     const size_t offset = weights_d.size() - weights_d.additional_buffer_size();
     auto w = const_cast<int8_t *>(weights);
     int32_t *compensation = (jcp.signed_input)
@@ -1761,21 +1767,11 @@ status_t jit_uni_x8s8s32x_deconvolution_fwd_t<isa>::execute_forward_3d(
     const size_t wht_kd_stride = wht_blk_off(weights_d, 0, 0, 0, 1);
     const size_t wht_kh_stride = wht_blk_off(weights_d, 0, 0, 0, 0, 1);
 
-    const float *oscales = pd()->attr()->output_scales_.scales_;
-    if (jcp.signed_input && (!jcp.has_vnni)) {
-        const auto local_scales
-                = ctx.get_scratchpad_grantor().template get<float>(
-                        key_conv_adjusted_scales);
-        const size_t count = pd()->attr()->output_scales_.count_;
-        const float factor = 1.f / pd()->jcp_.wei_adj_scale;
-        if (count == 1) {
-            utils::array_set(local_scales, oscales[0] * factor, 8);
-        } else {
-            for (size_t c = 0; c < count; c++)
-                local_scales[c] = oscales[c] * factor;
-        }
-        oscales = local_scales;
-    }
+    DEFINE_SCALES_BUFFER(oscales);
+
+    if (jcp.signed_input && (!jcp.has_vnni))
+        oscales = adjust_oscales(ctx.get_scratchpad_grantor(), oscales);
+
     const size_t offset = weights_d.size() - weights_d.additional_buffer_size();
     auto w = const_cast<int8_t *>(weights);
     int32_t *compensation = (jcp.signed_input)
