@@ -48,8 +48,8 @@ using namespace Xbyak;
 template <template <cpu_isa_t isa, data_type_t d_type> class Derived,
         cpu_isa_t isa, data_type_t d_type>
 jit_uni_lrn_kernel_t<Derived<isa, d_type>>::jit_uni_lrn_kernel_t(
-        void *code_ptr, size_t code_size)
-    : jit_generator(code_ptr, code_size, true, isa)
+        void *code_ptr, size_t code_size, const char *name)
+    : jit_generator(name, code_ptr, code_size, true, isa)
     , emulate_bfloat_(isa == avx512_core
               && d_type == dnnl::impl::data_type::bf16
               && !mayiuse(avx512_core_bf16))
@@ -57,16 +57,14 @@ jit_uni_lrn_kernel_t<Derived<isa, d_type>>::jit_uni_lrn_kernel_t(
               emulate_bfloat_ ? utils::make_unique<bf16_emulation_t>(this,
                       bf16_emu_reserv_1_, bf16_emu_reserv_2_,
                       bf16_emu_reserv_3_, bf16_emu_scratch_, bf16_emu_reserv_4_)
-                              : nullptr) {
-
-    if (bf16_emu_) bf16_emu_->init_vcvtneps2bf16();
-}
+                              : nullptr) {}
 
 template <template <cpu_isa_t isa, data_type_t d_type> class Derived,
         cpu_isa_t isa, data_type_t d_type>
 jit_uni_lrn_kernel_t<Derived<isa, d_type>>::jit_uni_lrn_kernel_t(
-        const within_config_t &config, void *code_ptr, size_t code_size)
-    : jit_uni_lrn_kernel_t(code_ptr, code_size) {
+        const within_config_t &config, void *code_ptr, size_t code_size,
+        const char *name)
+    : jit_uni_lrn_kernel_t(code_ptr, code_size, name) {
     if (config.dat_tag == nhwc)
         single_pixel_offset_
                 = config.C * sizeof(typename prec_traits<d_type>::type);
@@ -411,7 +409,7 @@ template <cpu_isa_t isa, data_type_t d_type>
 jit_uni_lrn_fwd_kernel_t<isa, d_type>::jit_uni_lrn_fwd_kernel_t(
         const within_config_t &config, float A, float K, prop_kind_t pk,
         void *code_ptr, size_t code_size)
-    : Base(config, code_ptr, code_size)
+    : Base(config, code_ptr, code_size, jit_name())
     , config_(lrn_config_t::within_config)
     , within_config_(config)
     , alpha_(A)
@@ -422,6 +420,7 @@ template <cpu_isa_t isa, data_type_t d_type>
 void jit_uni_lrn_fwd_kernel_t<isa, d_type>::generate(
         const within_config_t &config) {
     this->preamble();
+    if (this->bf16_emu_) this->bf16_emu_->init_vcvtneps2bf16();
 
 #define GET_OFF(field) offsetof(jit_args_fwd_t, field)
     this->mov(src_, this->ptr[this->param1 + GET_OFF(src)]);
@@ -446,7 +445,7 @@ template <cpu_isa_t isa, data_type_t d_type>
 jit_uni_lrn_fwd_kernel_t<isa, d_type>::jit_uni_lrn_fwd_kernel_t(
         const struct nchw8c_across_t &J, float A, float K, prop_kind_t pk,
         void *code_ptr, size_t code_size)
-    : Base(code_ptr, code_size)
+    : Base(code_ptr, code_size, jit_name())
     , config_(lrn_config_t::nchw8c_across)
     , nchw8c_across_(J)
     , alpha_(A)
@@ -471,6 +470,7 @@ void jit_uni_lrn_fwd_kernel_t<isa, d_type>::generate(const nchw8c_across_t &J) {
     const Xbyak::Ymm &ybase = this->ymm12;
 
     this->preamble();
+    if (this->bf16_emu_) this->bf16_emu_->init_vcvtneps2bf16();
 
     this->mov(src_, this->ptr[this->param1 + 0]);
     this->mov(dst_, this->ptr[this->param1 + 8]);
@@ -545,7 +545,7 @@ template <>
 jit_uni_lrn_fwd_kernel_t<sse41, dnnl::impl::data_type::f32>::
         jit_uni_lrn_fwd_kernel_t(const struct nchw8c_across_t &J, float A,
                 float K, prop_kind_t pk, void *code_ptr, size_t code_size)
-    : Base(code_ptr, code_size)
+    : Base(code_ptr, code_size, jit_name())
     , config_(lrn_config_t::nchw8c_across)
     , nchw8c_across_(J)
     , alpha_(A)
@@ -578,6 +578,7 @@ void jit_uni_lrn_fwd_kernel_t<sse41, dnnl::impl::data_type::f32>::generate(
     const Xbyak::Xmm &xbase_hi = this->xmm15;
 
     this->preamble();
+    if (this->bf16_emu_) this->bf16_emu_->init_vcvtneps2bf16();
 
     this->mov(src_, this->ptr[this->param1 + 0]);
     this->mov(dst_, this->ptr[this->param1 + 8]);
@@ -685,7 +686,7 @@ template <cpu_isa_t isa, data_type_t d_type>
 jit_uni_lrn_fwd_kernel_t<isa, d_type>::jit_uni_lrn_fwd_kernel_t(
         const struct nhwc_across_t &J, float A, float K, prop_kind_t pk,
         void *code_ptr, size_t code_size)
-    : Base(code_ptr, code_size)
+    : Base(code_ptr, code_size, jit_name())
     , config_(lrn_config_t::nhwc_across)
     , nhwc_across_(J)
     , alpha_(A)
@@ -709,17 +710,18 @@ void jit_uni_lrn_fwd_kernel_t<isa, d_type>::generate(const nhwc_across_t &J) {
     const Xbyak::Ymm &ymask = this->ymm10;
 
     this->preamble();
+    if (this->bf16_emu_) this->bf16_emu_->init_vcvtneps2bf16();
 
     this->mov(src_, this->ptr[this->param1 + 0]);
     this->mov(dst_, this->ptr[this->param1 + 8]);
     if (pk_ != prop_kind::forward_inference)
         this->mov(scratch_, this->ptr[this->param1 + 16]);
     this->mov(this->imm_addr64_, float2int(this->alpha_));
-    this->movq(xalpha_, this->imm_addr64_);
+    this->vmovq(xalpha_, this->imm_addr64_);
     this->vbroadcastss(valpha_, xalpha_);
 
     this->mov(this->imm_addr64_, float2int(this->k_));
-    this->movq(xk_, this->imm_addr64_);
+    this->vmovq(xk_, this->imm_addr64_);
     this->vbroadcastss(yk_, xk_);
 
     this->vxorps(ysum, ysum, ysum);
@@ -808,7 +810,7 @@ template <>
 jit_uni_lrn_fwd_kernel_t<sse41, dnnl::impl::data_type::f32>::
         jit_uni_lrn_fwd_kernel_t(const struct nhwc_across_t &J, float A,
                 float K, prop_kind_t pk, void *code_ptr, size_t code_size)
-    : Base(code_ptr, code_size)
+    : Base(code_ptr, code_size, jit_name())
     , config_(lrn_config_t::nhwc_across)
     , nhwc_across_(J)
     , alpha_(A)
@@ -840,6 +842,7 @@ void jit_uni_lrn_fwd_kernel_t<sse41, dnnl::impl::data_type::f32>::generate(
     const Xbyak::Xmm &xbase_hi = this->xmm15;
 
     this->preamble();
+    if (this->bf16_emu_) this->bf16_emu_->init_vcvtneps2bf16();
 
     this->mov(src_, this->ptr[this->param1 + 0]);
     this->mov(dst_, this->ptr[this->param1 + 8]);
@@ -1214,7 +1217,7 @@ template <cpu_isa_t isa, data_type_t d_type>
 jit_uni_lrn_fwd_kernel_t<isa, d_type>::jit_uni_lrn_fwd_kernel_t(
         const nchw_across_t &J, float A, float K, prop_kind_t pk,
         void *code_ptr, size_t code_size)
-    : Base(code_ptr, code_size)
+    : Base(code_ptr, code_size, jit_name())
     , config_(lrn_config_t::nchw_across)
     , nchw_across_(J)
     , alpha_(A)
@@ -1236,6 +1239,7 @@ void jit_uni_lrn_fwd_kernel_t<isa, d_type>::generate(const nchw_across_t &J) {
     const Xbyak::Ymm &ysum = this->ymm8;
 
     this->preamble();
+    if (this->bf16_emu_) this->bf16_emu_->init_vcvtneps2bf16();
 
     if (J.tail != 0) {
         this->mov(
@@ -1307,7 +1311,7 @@ template <>
 jit_uni_lrn_fwd_kernel_t<sse41, dnnl::impl::data_type::f32>::
         jit_uni_lrn_fwd_kernel_t(const nchw_across_t &J, float A, float K,
                 prop_kind_t pk, void *code_ptr, size_t code_size)
-    : Base(code_ptr, code_size)
+    : Base(code_ptr, code_size, jit_name())
     , config_(lrn_config_t::nchw_across)
     , nchw_across_(J)
     , alpha_(A)
@@ -1371,6 +1375,7 @@ void jit_uni_lrn_fwd_kernel_t<sse41, dnnl::impl::data_type::f32>::generate(
     size_t l_shift = 0;
 
     this->preamble();
+    if (this->bf16_emu_) this->bf16_emu_->init_vcvtneps2bf16();
 
     this->mov(src_, this->ptr[this->param1 + 0]);
     this->mov(dst_, this->ptr[this->param1 + 8]);
@@ -1486,7 +1491,7 @@ template <cpu_isa_t isa, data_type_t d_type>
 jit_uni_lrn_bwd_kernel_t<isa, d_type>::jit_uni_lrn_bwd_kernel_t(
         const nchw8c_across_t &J, float A, float B, int use_h_parallel,
         void *code_ptr, size_t code_size)
-    : Base(code_ptr, code_size)
+    : Base(code_ptr, code_size, jit_name())
     , config_(lrn_config_t::nchw8c_across)
     , nchw8c_across_(J)
     , nalphabeta_(-2 * A * B)
@@ -1515,6 +1520,7 @@ void jit_uni_lrn_bwd_kernel_t<isa, d_type>::generate(const nchw8c_across_t &J) {
     const Xbyak::Ymm &ydiffsrc = this->ymm15;
 
     this->preamble();
+    if (this->bf16_emu_) this->bf16_emu_->init_vcvtneps2bf16();
 
 #define GET_OFF(field) offsetof(jit_args_bwd_t, field)
     this->mov(src_, this->ptr[this->param1 + GET_OFF(src)]);
@@ -1623,7 +1629,7 @@ template <cpu_isa_t isa, data_type_t d_type>
 jit_uni_lrn_bwd_kernel_t<isa, d_type>::jit_uni_lrn_bwd_kernel_t(
         const within_config_t &config, float A, float B, void *code_ptr,
         size_t code_size)
-    : Base(config, code_ptr, code_size)
+    : Base(config, code_ptr, code_size, jit_name())
     , config_(lrn_config_t::within_config)
     , within_config_(config)
     , nalphabeta_(-2.0f * A * B) {}
@@ -1633,6 +1639,7 @@ void jit_uni_lrn_bwd_kernel_t<isa, d_type>::generate(
         const within_config_t &config) {
 
     this->preamble();
+    if (this->bf16_emu_) this->bf16_emu_->init_vcvtneps2bf16();
 
 #define GET_OFF(field) offsetof(jit_args_bwd_t, field)
     this->mov(src_, this->ptr[this->param1 + GET_OFF(src)]);
