@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2021-2022 Intel Corporation
+* Copyright 2021-2023 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -40,6 +40,22 @@ using namespace nstl;
 
 using namespace data_type;
 
+namespace {
+void maybe_tile_configure(bool is_amx,
+        const char brg_kernel_palettes[][AMX_PALETTE_SIZE], int brg_ker_idx,
+        int &prev_ker_idx) {
+    if (!is_amx) return;
+    if (brg_ker_idx == prev_ker_idx) return;
+    // TODO: more accurately estimate the costs of memcmp and tile configuration
+    if (prev_ker_idx == -1
+            || std::memcmp(&brg_kernel_palettes[brg_ker_idx][0],
+                       &brg_kernel_palettes[prev_ker_idx][0], AMX_PALETTE_SIZE)
+                    != 0)
+        amx_tile_configure(&brg_kernel_palettes[brg_ker_idx][0]);
+    prev_ker_idx = brg_ker_idx;
+}
+} // namespace
+
 template <cpu_isa_t isa>
 status_t brgemm_matmul_t<isa>::pd_t::init(engine_t *engine) {
     const auto src_dt = src_md_.data_type;
@@ -72,9 +88,10 @@ status_t brgemm_matmul_t<isa>::pd_t::init(engine_t *engine) {
             = [&]() -> bool { return attr()->zero_points_.common(); };
 
     const bool problem_dt_correct = is_int8 || is_bf16 || is_f32;
-    bool ok = mayiuse(isa) && problem_dt_correct
+    bool ok = mayiuse(isa) && problem_dt_correct && !has_zero_dim_memory()
             && !has_runtime_dims_or_strides()
-            && attr()->has_default_values(primitive_attr_t::skip_mask_t::oscale
+            && attr()->has_default_values(
+                    primitive_attr_t::skip_mask_t::oscale_runtime
                             | primitive_attr_t::skip_mask_t::zero_points_runtime
                             | primitive_attr_t::skip_mask_t::post_ops
                             | primitive_attr_t::skip_mask_t::sum_dt,
@@ -139,6 +156,8 @@ status_t brgemm_matmul_t<isa>::pd_t::init(engine_t *engine) {
         }
 
         CHECK(brgemm_desc_set_attr(&brg, brgattr));
+        bgmmc_.wsp_tile_per_thr_bytes = nstl::max(
+                brg.get_wsp_buffer_size(), bgmmc_.wsp_tile_per_thr_bytes);
     }
 
     auto scratchpad = scratchpad_registry().registrar();
@@ -190,9 +209,10 @@ status_t brgemm_matmul_t<isa>::execute_body(const exec_ctx_t &ctx) const {
     DEFINE_ZERO_POINT_VALUE(src_zero_point, DNNL_ARG_SRC);
     DEFINE_ZERO_POINT_VALUE(wei_zero_point, DNNL_ARG_WEIGHTS);
     DEFINE_ZERO_POINT_VALUE(dst_zero_point, DNNL_ARG_DST);
+    DEFINE_SCALES_BUFFER(oscales);
 
     brg_matmul_exec_ctx_t brgmm_ctx(
-            ctx, pd(), src_zero_point, wei_zero_point, dst_zero_point);
+            ctx, pd(), oscales, src_zero_point, wei_zero_point, dst_zero_point);
 
     const auto &bgmmc = pd()->get_brgemm_matmul_conf();
     const bool use_buffer_a
@@ -213,10 +233,9 @@ status_t brgemm_matmul_t<isa>::execute_body(const exec_ctx_t &ctx) const {
             balance211((int)bgmmc.K_chunks, brgmm_ctx.get_num_threads_for_k(),
                     ithr_k, kc_start, kc_end);
 
-        if (is_amx) {
-            const auto base_ker_idx = brgmm_ctx.get_base_brgemm_kernel_idx();
-            amx_tile_configure(&brg_kernel_palettes_[base_ker_idx][0]);
-        }
+        int prev_ker_idx = -1;
+        maybe_tile_configure(is_amx, brg_kernel_palettes_,
+                brgmm_ctx.get_base_brgemm_kernel_idx(), prev_ker_idx);
 
         int b {0}, mc {0}, nc {0};
         nd_iterator_init(
@@ -235,8 +254,8 @@ status_t brgemm_matmul_t<isa>::execute_body(const exec_ctx_t &ctx) const {
                 for (int mb = m_start; mb < m_end; mb++) {
                     if (use_buffer_a && nb == n_start)
                         copy_a_chunk_in_buffer(brgmm_ctx, ithr, b, mb, kc);
-                    compute_kernel(
-                            brgmm_ctx, ithr, b, mb, nb, kc, kc == kc_start);
+                    compute_kernel(brgmm_ctx, ithr, b, mb, nb, kc,
+                            kc == kc_start, prev_ker_idx);
                 }
             }
             ++start;
@@ -254,12 +273,12 @@ status_t brgemm_matmul_t<isa>::execute_body(const exec_ctx_t &ctx) const {
 template <cpu_isa_t isa>
 void brgemm_matmul_t<isa>::compute_kernel(
         const brg_matmul_exec_ctx_t &brgmm_ctx, int ithr, int b_idx,
-        int m_blk_idx, int n_blk_idx, int k_chunk_idx, bool do_init) const {
+        int m_blk_idx, int n_blk_idx, int k_chunk_idx, bool do_init,
+        int &prev_ker_idx) const {
     constexpr bool is_amx
             = one_of(isa, avx512_core_bf16_amx_int8, avx512_core_bf16_amx_bf16);
     const auto &bgmmc = pd()->get_brgemm_matmul_conf();
     const auto addr_batch = brgmm_ctx.get_batch_elem_ptr(ithr);
-    const int base_brg_ker_idx = brgmm_ctx.get_base_brgemm_kernel_idx();
 
     const auto wsp_tile = brgmm_ctx.get_tile_workspace(ithr);
     const int m = m_blk_idx * bgmmc.M_blk;
@@ -298,10 +317,8 @@ void brgemm_matmul_t<isa>::compute_kernel(
     if (gemm_batch > 0 && brg_ker_idx >= 0) {
         const auto brg_kernel = brg_kernels_[brg_ker_idx].get();
         assert(brg_kernel != nullptr);
-
-        const bool is_tile_reconf_required = is_amx && (is_M_tail || is_N_tail);
-        if (is_tile_reconf_required)
-            amx_tile_configure(&brg_kernel_palettes_[brg_ker_idx][0]);
+        maybe_tile_configure(
+                is_amx, brg_kernel_palettes_, brg_ker_idx, prev_ker_idx);
 
         brgmm_ctx.init_brgemm_batch_elements_values(
                 ithr, 0, gemm_batch, b_idx, m_blk_idx, k_blk_idx, n_blk_idx);
@@ -335,9 +352,6 @@ void brgemm_matmul_t<isa>::compute_kernel(
             brgemm_kernel_execute(brg_kernel, gemm_batch, addr_batch,
                     (void *)ptr_C, is_amx ? (void *)wsp_tile : nullptr);
         }
-
-        if (is_tile_reconf_required)
-            amx_tile_configure(&brg_kernel_palettes_[base_brg_ker_idx][0]);
     }
     if (is_K_tail) {
         brgmm_ctx.init_brgemm_batch_elements_values(
@@ -346,11 +360,10 @@ void brgemm_matmul_t<isa>::compute_kernel(
         const bool use_init_ker = (do_init && gemm_batch == 0);
         const int brg_ker_idx = pd()->get_brg_kernel_idx(
                 false, use_init_ker, is_M_tail, is_N_tail, true);
+        maybe_tile_configure(
+                is_amx, brg_kernel_palettes_, brg_ker_idx, prev_ker_idx);
         const auto brg_kernel_k_tail = brg_kernels_[brg_ker_idx].get();
-        const bool is_tile_reconf_required
-                = is_amx && bgmmc.K_tail != bgmmc.K_blk;
-        if (is_tile_reconf_required)
-            amx_tile_configure(&brg_kernel_palettes_[brg_ker_idx][0]);
+
         if (post_ops_applicable) {
             void *scratch = is_amx
                     ? static_cast<void *>(wsp_tile)
@@ -380,8 +393,6 @@ void brgemm_matmul_t<isa>::compute_kernel(
             brgemm_kernel_execute(brg_kernel_k_tail, 1, addr_batch,
                     (void *)ptr_C, is_amx ? (void *)wsp_tile : nullptr);
         }
-        if (is_tile_reconf_required)
-            amx_tile_configure(&brg_kernel_palettes_[base_brg_ker_idx][0]);
     }
 }
 
@@ -390,6 +401,9 @@ void brgemm_matmul_t<isa>::maybe_reduce_partial_results_and_apply_postops(
         const brg_matmul_exec_ctx_t &brgmm_ctx) const {
     if (!brgmm_ctx.parallel_reduction_is_used()) return;
 
+    constexpr bool is_amx
+            = one_of(isa, avx512_core_bf16_amx_int8, avx512_core_bf16_amx_bf16);
+
     const auto &bgmmc = pd()->get_brgemm_matmul_conf();
     const int num_threads = brgmm_ctx.get_num_threads_for_parallelization();
 
@@ -408,6 +422,8 @@ void brgemm_matmul_t<isa>::maybe_reduce_partial_results_and_apply_postops(
                 bmn_end);
         balance211(bmn_end - bmn_start, nthr_k, ithr_k, start, end);
 
+        int prev_ker_idx = -1;
+
         int b {0}, mc {0}, nc {0};
 
         assert(bgmmc.batch == 1);
@@ -446,6 +462,8 @@ void brgemm_matmul_t<isa>::maybe_reduce_partial_results_and_apply_postops(
                                 = (bgmmc.N - nb * bgmmc.N_blk < bgmmc.N_blk);
                         const int brg_ker_idx = pd()->get_brg_kernel_idx(
                                 false, false, is_M_tail, is_N_tail, false);
+                        maybe_tile_configure(is_amx, brg_kernel_palettes_,
+                                brg_ker_idx, prev_ker_idx);
                         const auto brg_kernel = brg_kernels_[brg_ker_idx].get();
                         const int m = mb * bgmmc.M_blk;
                         const int n = nb * bgmmc.N_blk;
@@ -606,8 +624,9 @@ void brgemm_matmul_t<isa>::accumulate(
 
 template <cpu_isa_t isa>
 struct brgemm_matmul_t<isa>::brg_matmul_exec_ctx_t {
-    brg_matmul_exec_ctx_t(const exec_ctx_t &ctx, const pd_t *pd, int32_t src_zp,
-            int32_t wei_zp, int32_t dst_zp)
+    brg_matmul_exec_ctx_t(const exec_ctx_t &ctx, const pd_t *pd,
+            const float *oscales, int32_t src_zp, int32_t wei_zp,
+            int32_t dst_zp)
         : bgmmc_(pd->get_brgemm_matmul_conf()) {
 
         data_A_ptr_ = CTX_IN_MEM(const char *, DNNL_ARG_SRC);
@@ -615,7 +634,7 @@ struct brgemm_matmul_t<isa>::brg_matmul_exec_ctx_t {
         data_C_ptr_ = CTX_OUT_MEM(char *, DNNL_ARG_DST);
 
         bias_ptr_ = CTX_IN_MEM(const char *, DNNL_ARG_BIAS);
-        oscales_ptr_ = pd->attr()->output_scales_.scales_;
+        oscales_ptr_ = oscales;
         memory_tracking::grantor_t scratchpad = ctx.get_scratchpad_grantor();
         const auto &bgmmc = pd->get_brgemm_matmul_conf();
 
@@ -654,6 +673,8 @@ struct brgemm_matmul_t<isa>::brg_matmul_exec_ctx_t {
                                         reinterpret_cast<const int32_t *>(
                                                 &data_B_ptr_[comp_offset])))
                 : nullptr;
+        assert(IMPLICATION(bgmmc.s8s8_compensation_required,
+                bgmmc_.b_dt_sz == bgmmc_.tr_b_dt_sz));
 
         zero_point_a_compensations_ptr_ = bgmmc.has_zero_point_a
                 ? scratchpad.template get<int32_t>(
@@ -675,9 +696,9 @@ struct brgemm_matmul_t<isa>::brg_matmul_exec_ctx_t {
                 pd->attr()->post_ops_, ctx);
         base_brg_ker_idx_
                 = pd->get_brg_kernel_idx(false, true, false, false, false);
-        vnni_factor = isa == avx512_core_bf16_amx_int8
-                ? 4
-                : isa == avx512_core_bf16_amx_bf16 ? 2 : 1;
+        vnni_factor = isa == avx512_core_bf16_amx_int8 ? 4
+                : isa == avx512_core_bf16_amx_bf16     ? 2
+                                                       : 1;
 
         reorder_zp_a_comp_ptr_ = nullptr;
         if (bgmmc_.has_zero_point_a && bgmmc_.blocked_B) {
@@ -894,7 +915,10 @@ struct brgemm_matmul_t<isa>::brg_matmul_exec_ctx_t {
             }
             return b_off + bgmmc_.B_strides[1] * k + bgmmc_.B_strides[0] * n;
         } else {
-            int k_idx = bgmmc_.blocked_B ? k / bgmmc_.wei_k_blk : k;
+            int dt_b_k_blk = bgmmc_.is_bf32
+                    ? data_type_vnni_simd_elems<avx512_core>(f32)
+                    : bgmmc_.wei_k_blk;
+            int k_idx = bgmmc_.blocked_B ? k / dt_b_k_blk : k;
             int n_idx = bgmmc_.blocked_B ? n / bgmmc_.wei_n_blk : n;
             return bgmmc_.B_strides[2] * b + bgmmc_.B_strides[1] * k_idx
                     + bgmmc_.B_strides[0] * n_idx
