@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2021 Intel Corporation
+* Copyright 2021-2023 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -40,8 +40,10 @@ struct jit_brgemm_matmul_copy_a_impl_t : public jit_brgemm_matmul_copy_a_t,
 
     jit_brgemm_matmul_copy_a_impl_t(const brgemm_matmul_conf_t *conf)
         : jit_brgemm_matmul_copy_a_t(conf)
+        , jit_generator(jit_name())
         , typesize(conf_->a_dt_sz)
-        , vnni_granularity(granularity_max / typesize)
+        , tr_typesize(conf_->tr_a_dt_sz)
+        , vnni_granularity(data_type_vnni_granularity(conf_->src_dt))
         , k_step(bytes_in_zmm / typesize) {}
 
     void operator()(ctx_t *ctx) override { jit_generator::operator()(ctx); }
@@ -56,12 +58,12 @@ private:
     using xmm = const Xbyak::Xmm;
 
     enum {
-        granularity_max = 4,
         num_comp_acc = 8,
         k_loop_unroll = 16,
         bytes_in_zmm = 64,
     };
     const int typesize;
+    const int tr_typesize;
     const int vnni_granularity;
     const int k_step;
 
@@ -170,41 +172,83 @@ void jit_brgemm_matmul_copy_a_impl_t::copy_K_loop(
             for (int k = k_start; k < k_end; k++)
                 vpaddb(get_zmm_copy(k), get_zmm_copy(k), zmm_comp_add);
         }
+        if (conf_->is_bf32) {
+            assert(typesize != tr_typesize);
+            int k = k_start;
+            const int k_end_2 = rnd_dn(k_end, 2);
+            for (; k < k_end_2; k += 2) {
+                const size_t offset = ((size_t)kb * k_loop_unroll + k) * k_step
+                        * tr_typesize;
+                auto tr_src_addr = EVEX_compress_addr(reg_tr_src, offset);
 
-        for (int k = k_start; k < k_end; k++) {
-            const size_t offset
-                    = ((size_t)kb * k_loop_unroll + k) * k_step * typesize;
-            vmovdqu8(EVEX_compress_addr(reg_tr_src, offset), get_zmm_copy(k));
+                auto zmm_src = get_zmm_copy(k);
+                auto zmm_src_next = get_zmm_copy(k + 1);
+
+                vcvtne2ps2bf16(zmm_src, zmm_src_next, zmm_src);
+                vmovups(tr_src_addr, zmm_src);
+            }
+            if (k < k_end) {
+                const size_t offset = ((size_t)kb * k_loop_unroll + k) * k_step
+                        * tr_typesize;
+                auto tr_src_addr = EVEX_compress_addr(reg_tr_src, offset);
+                ymm ymm_downcvt_bf16 = ymm(get_zmm_copy(k).getIdx());
+                vcvtneps2bf16(ymm_downcvt_bf16, get_zmm_copy(k));
+                vmovdqu16(tr_src_addr, ymm_downcvt_bf16);
+            }
+        } else {
+            for (int k = k_start; k < k_end; k++) {
+                const size_t offset = ((size_t)kb * k_loop_unroll + k) * k_step
+                        * tr_typesize;
+                auto tr_src_addr = EVEX_compress_addr(reg_tr_src, offset);
+                vmovdqu8(tr_src_addr, get_zmm_copy(k));
+            }
         }
     }
 
     if (k_tail > 0) {
-        const auto kmovq = [=](Opmask k, size_t q) {
-            mov(regq_tmp, q);
-            jit_generator::kmovq(k, regq_tmp);
+        const auto kmovx = [=](Opmask k, size_t q) {
+            if (conf_->is_bf32) {
+                mov(regq_tmp.cvt32(), q);
+                jit_generator::kmovw(k, regq_tmp.cvt32());
+            } else {
+                mov(regq_tmp, q);
+                jit_generator::kmovq(k, regq_tmp);
+            }
         };
 
+        const size_t dt_step = conf_->is_bf32 ? 1 : typesize;
         const size_t tail_mask_load
-                = size_t(((size_t)1 << (typesize * k_tail)) - 1);
-        kmovq(kTail_load, tail_mask_load);
+                = size_t(((size_t)1 << (dt_step * k_tail)) - 1);
+        kmovx(kTail_load, tail_mask_load);
         const int k_tail_st = rnd_up(k_tail, vnni_granularity);
+        const size_t full_mask
+                = conf_->is_bf32 ? ((size_t)1 << 16) - 1 : 0xffffffffffffffff;
         const size_t tail_mask_store = k_tail_st == k_step
-                ? 0xffffffffffffffff
-                : size_t(((size_t)1 << (typesize * k_tail_st)) - 1);
-        kmovq(kTail_store, tail_mask_store);
+                ? full_mask
+                : size_t(((size_t)1 << (dt_step * k_tail_st)) - 1);
+        kmovx(kTail_store, tail_mask_store);
 
         auto zmm_tail = get_zmm_copy(0) | kTail_load | T_z;
-        vmovdqu8(zmm_tail,
-                EVEX_compress_addr(reg_src, num_k_iters * k_step * typesize));
+        auto load_addr
+                = EVEX_compress_addr(reg_src, num_k_iters * k_step * typesize);
+        if (conf_->is_bf32)
+            vmovups(zmm_tail, load_addr);
+        else
+            vmovdqu8(zmm_tail, load_addr);
 
         maybe_compute_compensation(0, get_zmm_copy(0));
 
         if (allow_input_shift_for_s8s8 && conf_->s8s8_compensation_required)
             vpaddb(get_zmm_copy(0), get_zmm_copy(0), zmm_comp_add);
 
-        vmovdqu8(
-                EVEX_compress_addr(reg_tr_src, num_k_iters * k_step * typesize),
-                get_zmm_copy(0) | kTail_store);
+        auto tr_src_addr = EVEX_compress_addr(
+                reg_tr_src, num_k_iters * k_step * tr_typesize);
+        if (conf_->is_bf32) {
+            ymm ymm_downcvt_bf16 = ymm(get_zmm_copy(0).getIdx());
+            vcvtneps2bf16(ymm_downcvt_bf16, get_zmm_copy(0));
+            vmovdqu16(tr_src_addr, ymm_downcvt_bf16 | kTail_store);
+        } else
+            vmovdqu8(tr_src_addr, get_zmm_copy(0) | kTail_store);
     }
 
     if (do_compute_compensation) {
@@ -297,11 +341,12 @@ void jit_brgemm_matmul_copy_a_impl_t::copy_M_loop(
 
 void jit_brgemm_matmul_copy_a_impl_t::generate() {
     preamble();
-    src_stride = (conf_->src_tag == format_tag::acbd ? conf_->copy_A_src_stride
-                                                     : conf_->K * typesize);
+
+    src_stride = conf_->src_tag == format_tag::acbd ? conf_->copy_A_src_stride
+                                                    : conf_->K * typesize;
     const dim_t LDA = conf_->use_buffer_a_tail_only ? (dim_t)conf_->wei_k_blk
                                                     : conf_->LDA;
-    tr_src_stride = LDA * typesize;
+    tr_src_stride = LDA * tr_typesize;
     do_compute_compensation = conf_->has_zero_point_b;
 
     mov(reg_src, ptr[param1 + GET_OFF(src)]);
@@ -377,16 +422,19 @@ struct jit_brgemm_matmul_copy_a_transposed_impl_t
 
     jit_brgemm_matmul_copy_a_transposed_impl_t(const brgemm_matmul_conf_t *conf)
         : jit_brgemm_matmul_copy_a_t(conf)
+        , jit_generator(jit_name())
         , typesize(conf_->a_dt_sz)
+        , tr_typesize(conf_->tr_a_dt_sz)
         , src_stride(conf_->src_tag == format_tag::adbc
                           ? conf_->copy_A_src_stride
                           : conf_->M * typesize)
-        , dst_stride(conf_->LDA * typesize)
+        , dst_stride(conf_->LDA * tr_typesize)
         , m_loop_src_shift(columns_step * typesize)
         , m_loop_dst_shift(columns_step * dst_stride)
         , k_loop_src_shift(rows_step * src_stride)
-        , k_loop_dst_shift(rows_step * typesize)
-        , is_f32(everyone_is(data_type::f32, conf_->src_dt, conf_->wei_dt)) {}
+        , k_loop_dst_shift(rows_step * tr_typesize)
+        , is_f32(everyone_is(data_type::f32, conf_->src_dt, conf_->wei_dt))
+        , is_bf32(conf_->is_bf32) {}
 
     void operator()(ctx_t *ctx) override { jit_generator::operator()(ctx); }
     status_t create_kernel() override { return jit_generator::create_kernel(); }
@@ -397,6 +445,7 @@ private:
     using opmask_t = const Xbyak::Opmask;
 
     const size_t typesize;
+    const size_t tr_typesize;
     const int rows_step = 16;
     const int columns_step = rows_step;
     const dim_t src_stride, dst_stride;
@@ -405,6 +454,7 @@ private:
     const dim_t k_loop_src_shift;
     const dim_t k_loop_dst_shift;
     const bool is_f32;
+    const bool is_bf32;
 
     opmask_t kFFFF = k1;
     opmask_t k3333 = k1;
@@ -454,9 +504,12 @@ void jit_brgemm_matmul_copy_a_transposed_impl_t::transpose_bf16(
         return Ymm(i);
     };
 
-    auto kmovd = [=](Opmask k, unsigned w) {
+    auto kmovx = [=](Opmask k, unsigned w, bool use_word_sz = false) {
         mov(regw_tmp, w);
-        jit_generator::kmovd(k, regw_tmp);
+        if (use_word_sz)
+            jit_generator::kmovw(k, regw_tmp);
+        else
+            jit_generator::kmovd(k, regw_tmp);
     };
 
     auto store = [=](Zmm r, int i) {
@@ -466,17 +519,25 @@ void jit_brgemm_matmul_copy_a_transposed_impl_t::transpose_bf16(
 
     const int load_mask
             = ncolumns < columns_step ? (1 << ncolumns) - 1 : 0xffff;
-    kmovd(kFFFF, load_mask);
+    kmovx(kFFFF, load_mask, is_bf32);
 
     for (int i = 0; i < nrows / 2; i++) {
-        auto zmm_src0 = src_zmm(2 * i);
-        auto zmm_src1 = src_zmm(2 * i + 1);
-        auto src1 = src_ymm(2 * i + 1);
-        vmovdqu16(zmm_src0 | kFFFF | T_z,
-                EVEX_compress_addr(src, 2 * i * src_stride));
-        vmovdqu16(zmm_src1 | kFFFF | T_z,
-                EVEX_compress_addr(src, (2 * i + 1) * src_stride));
-        vinsertf64x4(zmm_src0, zmm_src0, src1, 1);
+        auto idx0 = 2 * i;
+        auto idx1 = 2 * i + 1;
+        auto zmm_src0 = src_zmm(idx0);
+        auto zmm_src1 = src_zmm(idx1);
+        auto src_addr_0 = EVEX_compress_addr(src, idx0 * src_stride);
+        auto src_addr_1 = EVEX_compress_addr(src, idx1 * src_stride);
+        if (is_bf32) {
+            vmovups(zmm_src0 | kFFFF | T_z, src_addr_0);
+            vmovups(zmm_src1 | kFFFF | T_z, src_addr_1);
+            vcvtne2ps2bf16(zmm_src0, zmm_src1, zmm_src0);
+        } else {
+            auto src1 = src_ymm(idx1);
+            vmovdqu16(zmm_src0 | kFFFF | T_z, src_addr_0);
+            vmovdqu16(zmm_src1 | kFFFF | T_z, src_addr_1);
+            vinsertf64x4(zmm_src0, zmm_src0, src1, 1);
+        }
         vpermw(zmm_src0, vidx5, zmm_src0);
     }
 
@@ -484,8 +545,12 @@ void jit_brgemm_matmul_copy_a_transposed_impl_t::transpose_bf16(
     if (nrows % 2) {
         int i = nrows / 2;
         auto zmm_src0 = src_zmm(2 * i);
-        vmovdqu16(zmm_src0 | kFFFF | T_z,
-                EVEX_compress_addr(src, 2 * i * src_stride));
+        auto src_addr = EVEX_compress_addr(src, 2 * i * src_stride);
+        if (is_bf32) {
+            vmovups(zmm_src0 | kFFFF | T_z, src_addr);
+            vcvtneps2bf16(Ymm(zmm_src0.getIdx()), zmm_src0);
+        } else
+            vmovdqu16(zmm_src0 | kFFFF | T_z, src_addr);
         vpermw(zmm_src0, vidx5, zmm_src0);
     }
 
@@ -573,7 +638,7 @@ void jit_brgemm_matmul_copy_a_transposed_impl_t::transpose_bf16(
     const int store_mask = columns_to_store < rows_step
             ? (1 << columns_to_store) - 1
             : 0xffff;
-    kmovd(kTail, store_mask);
+    kmovx(kTail, store_mask);
 
     for (int col_idx = 0; col_idx < ncolumns; col_idx++)
         store(src_zmm(get_vec_idx(col_idx)), col_idx);
@@ -887,7 +952,7 @@ struct jit_brgemm_matmul_copy_b_int8_t : public jit_brgemm_matmul_copy_b_t,
     DECLARE_CPU_JIT_AUX_FUNCTIONS(jit_brgemm_matmul_copy_b_int8_t)
 
     jit_brgemm_matmul_copy_b_int8_t(const brgemm_matmul_conf_t *conf)
-        : jit_brgemm_matmul_copy_b_t(conf) {}
+        : jit_brgemm_matmul_copy_b_t(conf), jit_generator(jit_name()) {}
 
     void operator()(ctx_t *ctx) override { jit_generator::operator()(ctx); }
     status_t create_kernel() override { return jit_generator::create_kernel(); }
@@ -1367,7 +1432,16 @@ struct jit_brgemm_matmul_copy_b_bf16_t : public jit_brgemm_matmul_copy_b_t,
     DECLARE_CPU_JIT_AUX_FUNCTIONS(jit_brgemm_matmul_copy_b_bf16_t)
 
     jit_brgemm_matmul_copy_b_bf16_t(const brgemm_matmul_conf_t *conf)
-        : jit_brgemm_matmul_copy_b_t(conf) {}
+        : jit_brgemm_matmul_copy_b_t(conf)
+        , jit_generator(jit_name())
+        , typesize(conf->b_dt_sz)
+        , tr_typesize(conf->tr_b_dt_sz)
+        , src_stride(conf_->wei_tag == format_tag::acbd
+                          ? conf->copy_B_wei_stride
+                          : conf->req_wei_vnni_downconvert
+                          ? conf_->LDB * typesize
+                          : conf_->N * typesize)
+        , tr_src_stride(conf_->LDB * k_blk_step * tr_typesize) {}
 
     void operator()(ctx_t *ctx) override { jit_generator::operator()(ctx); }
     status_t create_kernel() override { return jit_generator::create_kernel(); }
@@ -1379,8 +1453,9 @@ private:
     using zmm = const Xbyak::Zmm;
     using ymm = const Xbyak::Ymm;
 
-    enum { typesize = sizeof(int16_t), k_blk_step = 2, n_blk_step = 16 };
-    dim_t src_stride = 0, tr_src_stride = 0;
+    enum { k_blk_step = 2, n_blk_step = 16 };
+    const int typesize, tr_typesize;
+    const dim_t src_stride, tr_src_stride;
 
     opmask_t kTail = k7;
     opmask_t kFFFF = k6;
@@ -1403,14 +1478,17 @@ private:
 
 void jit_brgemm_matmul_copy_b_bf16_t::copy_2x32_vnni(int nrows, int ncolumns) {
 
-    auto kmovd = [=](Opmask k, unsigned w) {
+    auto kmovx = [=](Opmask k, unsigned w) {
         mov(regw_tmp, w);
-        jit_generator::kmovd(k, regw_tmp);
+        if (conf_->is_bf32)
+            jit_generator::kmovw(k, regw_tmp);
+        else
+            jit_generator::kmovd(k, regw_tmp);
     };
 
     const int columns_tail = ncolumns % n_blk_step;
     const auto tail_mask = (1 << columns_tail) - 1;
-    if (columns_tail < n_blk_step) kmovd(kTail, tail_mask);
+    if (columns_tail < n_blk_step) kmovx(kTail, tail_mask);
 
     const int blk_sz = k_blk_step;
     const int max_regs_available = 30;
@@ -1425,8 +1503,13 @@ void jit_brgemm_matmul_copy_b_bf16_t::copy_2x32_vnni(int nrows, int ncolumns) {
     auto load = [=](int blk, int k, int n, opmask_t current_mask) {
         auto src_reg = get_zmm(blk, k % k_blk_step);
         auto src_load = src_reg | current_mask | T_z;
-        vmovdqu16(src_load,
-                EVEX_compress_addr(reg_src, k * src_stride + n * typesize));
+        auto load_addr
+                = EVEX_compress_addr(reg_src, k * src_stride + n * typesize);
+        if (conf_->is_bf32) {
+            vmovups(src_load, load_addr);
+        } else {
+            vmovdqu16(src_load, load_addr);
+        }
     };
 
     int iter = 0;
@@ -1434,7 +1517,7 @@ void jit_brgemm_matmul_copy_b_bf16_t::copy_2x32_vnni(int nrows, int ncolumns) {
     for (int n = 0; n < conf_->wei_n_blk; n += n_blk_step) {
         const int k_blk = k / k_blk_step;
         const dim_t tr_src_off
-                = k_blk * tr_src_stride + n * k_blk_step * typesize;
+                = k_blk * tr_src_stride + n * k_blk_step * tr_typesize;
         const auto store_addr = EVEX_compress_addr(reg_tr_src, tr_src_off);
         if (ncolumns - n <= 0) {
             vmovups(store_addr, zmm_zero);
@@ -1444,12 +1527,19 @@ void jit_brgemm_matmul_copy_b_bf16_t::copy_2x32_vnni(int nrows, int ncolumns) {
         const opmask_t curr_msk = ncolumns - n < n_blk_step ? kTail : kFFFF;
         const int blk_idx = iter % max_unroll;
         load(blk_idx, k, n, curr_msk);
+
         const auto src_zmm0 = get_zmm(blk_idx, 0);
         if (nrows - k >= k_blk_step) {
             load(blk_idx, k + 1, n, curr_msk);
             const auto src_zmm1 = get_zmm(blk_idx, 1);
-            const auto src_ymm1 = ymm(src_zmm1.getIdx());
-            vinsertf64x4(src_zmm0, src_zmm0, src_ymm1, 1);
+            if (conf_->is_bf32) {
+                vcvtne2ps2bf16(src_zmm0, src_zmm1, src_zmm0);
+            } else {
+                const auto src_ymm1 = ymm(src_zmm1.getIdx());
+                vinsertf64x4(src_zmm0, src_zmm0, src_ymm1, 1);
+            }
+        } else if (conf_->is_bf32) {
+            vcvtneps2bf16(ymm(src_zmm0.getIdx()), src_zmm0);
         }
 
         vpermw(src_zmm0, zmm_permw, src_zmm0);
@@ -1460,11 +1550,9 @@ void jit_brgemm_matmul_copy_b_bf16_t::copy_2x32_vnni(int nrows, int ncolumns) {
 }
 
 void jit_brgemm_matmul_copy_b_bf16_t::generate() {
+    assert(tr_typesize == sizeof(bfloat16_t));
     preamble();
     vpxord(zmm_zero, zmm_zero, zmm_zero);
-    src_stride = (conf_->wei_tag == format_tag::acbd ? conf_->copy_B_wei_stride
-                                                     : conf_->N * typesize);
-    tr_src_stride = conf_->LDB * k_blk_step * typesize;
 
     alignas(64) static constexpr const int16_t bf16_vnni_permute[32]
             = {0, 16, 1, 17, 2, 18, 3, 19, 4, 20, 5, 21, 6, 22, 7, 23, 8, 24, 9,
@@ -1475,12 +1563,7 @@ void jit_brgemm_matmul_copy_b_bf16_t::generate() {
     mov(reg_K_iters, ptr[param1 + GET_OFF(current_K_iters)]);
     mov(reg_N_blk, ptr[param1 + GET_OFF(current_N_blk)]);
 
-    auto kmovw = [=](Opmask k, unsigned w) {
-        mov(regw_tmp, w);
-        jit_generator::kmovw(k, regw_tmp);
-    };
-
-    kmovw(kFFFF, 0xffff); // 1111111111111111
+    kxnorw(kFFFF, kFFFF, kFFFF); // 1111 1111 1111 1111
     auto vmovdqa64 = [=](Zmm z, const int64_t *addr) {
         mov(imm_addr64, reinterpret_cast<size_t>(addr));
         jit_generator::vmovdqa64(z, ptr[imm_addr64]);
@@ -1553,6 +1636,7 @@ struct jit_brgemm_matmul_copy_b_f32_t : public jit_brgemm_matmul_copy_b_t,
 
     jit_brgemm_matmul_copy_b_f32_t(const brgemm_matmul_conf_t *conf)
         : jit_brgemm_matmul_copy_b_t(conf)
+        , jit_generator(jit_name())
         , src_stride_(conf_->wei_tag == acbd ? conf_->copy_B_wei_stride
                                              : conf_->N * typesize)
         , tr_src_stride_(conf_->LDB * typesize) {}
@@ -1692,13 +1776,20 @@ struct jit_brgemm_matmul_copy_b_transposed_t
 
     jit_brgemm_matmul_copy_b_transposed_t(const brgemm_matmul_conf_t *conf)
         : jit_brgemm_matmul_copy_b_t(conf)
+        , jit_generator(jit_name())
         , typesize(conf_->b_dt_sz)
-        , vnni_granularity(granularity_max / typesize)
-        , k_blk_step(bytes_in_zmm / typesize)
+        , tr_typesize(conf_->tr_b_dt_sz)
+        , vnni_granularity(data_type_vnni_granularity(conf_->wei_dt))
+        , k_blk_step(bytes_in_zmm / tr_typesize)
         , do_compute_compensation(
                   conf_->has_zero_point_a || conf_->s8s8_compensation_required)
+        , is_bf32(conf->is_bf32)
         , req_zp_comp(conf_->has_zero_point_a)
-        , req_s8s8_comp(conf_->s8s8_compensation_required) {}
+        , req_s8s8_comp(conf_->s8s8_compensation_required)
+        , src_stride(conf_->wei_tag == format_tag::adbc
+                          ? conf_->copy_B_wei_stride
+                          : conf_->K * typesize)
+        , tr_src_stride(conf_->LDB * vnni_granularity * tr_typesize) {}
 
     void operator()(ctx_t *ctx) override { jit_generator::operator()(ctx); }
     status_t create_kernel() override { return jit_generator::create_kernel(); }
@@ -1710,19 +1801,21 @@ private:
     using zmm = const Xbyak::Zmm;
 
     enum {
-        granularity_max = 4,
         n_blk_step = 16,
         bytes_in_zmm = 64,
+        bf32_k_blk_step = 16,
     };
 
     const int typesize;
+    const int tr_typesize;
     const int vnni_granularity;
     const int k_blk_step;
     const bool do_compute_compensation;
+    const bool is_bf32;
     const bool req_zp_comp;
     const bool req_s8s8_comp;
 
-    dim_t src_stride = 0, tr_src_stride = 0;
+    const dim_t src_stride, tr_src_stride;
 
     opmask_t k3333 = k1;
     opmask_t k5555 = k2;
@@ -1755,6 +1848,16 @@ private:
     zmm zmm_all_bits_1 = zmm27;
     zmm zmm_one_s32 = zmm26;
 
+    void kmovw(Opmask k, unsigned w) {
+        mov(regw_tmp, w);
+        jit_generator::kmovw(k, regw_tmp);
+    };
+
+    void kmovq(Opmask k, size_t q) {
+        mov(regq_tmp, q);
+        jit_generator::kmovq(k, regq_tmp);
+    };
+
     void copy_16x64_vnni(int nrows, int ncolumns);
     void compute_K_loop(bool is_N_tail, int curr_K_tail, bool is_first_K_iter,
             bool is_last_K_iter);
@@ -1781,14 +1884,44 @@ void jit_brgemm_matmul_copy_b_transposed_t::copy_16x64_vnni(
         return Zmm(16 + i);
     };
 
-    auto kmovq = [=](Opmask k, size_t q) {
-        mov(regq_tmp, q);
-        jit_generator::kmovq(k, regq_tmp);
-    };
+    const int columns_tail
+            = ncolumns % (is_bf32 ? bf32_k_blk_step : k_blk_step);
+    if (columns_tail > 0) {
+        const auto tail_mask = size_t(
+                ((size_t)1 << (is_bf32 ? 1 : typesize) * columns_tail) - 1);
+        if (is_bf32)
+            kmovw(kTail, tail_mask);
+        else
+            kmovq(kTail, tail_mask);
+    }
 
-    const int columns_tail = ncolumns % k_blk_step;
-    const auto tail_mask = size_t(((size_t)1 << (typesize * columns_tail)) - 1);
-    if (columns_tail > 0) kmovq(kTail, tail_mask);
+    auto load_bf32 = [=](int i) {
+        auto src_reg = src_zmm(i);
+        auto src_reg_next = tmp_zmm(i);
+
+        if (i >= nrows) {
+            vpxord(src_reg, src_reg, src_reg);
+            return;
+        }
+
+        // check if k_tail exists and it's in the first zmm
+        auto zmm_src = columns_tail > 0 && ncolumns < bf32_k_blk_step
+                ? src_reg | kTail | T_z
+                : src_reg;
+        vmovups(zmm_src, EVEX_compress_addr(reg_src, i * src_stride));
+
+        if (ncolumns <= bf32_k_blk_step) {
+            vpxord(src_reg_next, src_reg_next, src_reg_next);
+        } else {
+            auto zmm_src_next = columns_tail > 0 ? src_reg_next | kTail | T_z
+                                                 : src_reg_next;
+            vmovups(zmm_src_next,
+                    EVEX_compress_addr(reg_src,
+                            i * src_stride + bf32_k_blk_step * typesize));
+        }
+
+        vcvtne2ps2bf16(src_reg, src_reg_next, src_reg);
+    };
 
     auto load = [=](int i) {
         auto src_reg = src_zmm(i);
@@ -1813,32 +1946,65 @@ void jit_brgemm_matmul_copy_b_transposed_t::copy_16x64_vnni(
         const int tmp_corr_idx = do_compute_compensation * base_idx;
 
         // swap 1
-        for (int i = 0; i < 4; i++) {
-            const int src_idx0 = base_idx + i * 2;
-            const int src_idx1 = src_idx0 + 1;
+        if (is_bf32) {
+            for (int i = 0; i < 4; i++) {
+                const int src_idx0 = base_idx + i * 2;
+                const int src_idx1 = src_idx0 + 1;
 
-            int next_src_idx0 = src_idx0 + 2;
-            int next_src_idx1 = src_idx1 + 2;
-            bool load_next = base_idx == 0 || i < 3;
+                if (base_idx == 0 && i == 0) {
+                    load_bf32(src_idx0);
+                    load_bf32(src_idx1);
+                }
 
-            if (base_idx == 0 && i == 0) {
-                load(src_idx0);
-                load(src_idx1);
+                int next_src_idx0 = src_idx0 + 2;
+                int next_src_idx1 = src_idx1 + 2;
+
+                bool load_next = base_idx == 0 || i < 3;
+
+                const auto tmp0 = tmp_zmm(src_idx0 - tmp_corr_idx);
+                const auto tmp1 = tmp_zmm(src_idx1 - tmp_corr_idx);
+                const auto src0 = src_zmm(src_idx0);
+                const auto src1 = src_zmm(src_idx1);
+
+                if (next_src_idx0 < nrows && load_next)
+                    load_bf32(next_src_idx0);
+                valignd(tmp0, src0, src0, 0x1);
+
+                if (next_src_idx1 < nrows && load_next)
+                    load_bf32(next_src_idx1);
+                valignd(tmp1, src1, src1, 0xf);
+
+                vmovaps(src0 | kAAAA, tmp1);
+                vmovaps(src1 | k5555, tmp0);
             }
+        } else {
+            for (int i = 0; i < 4; i++) {
+                const int src_idx0 = base_idx + i * 2;
+                const int src_idx1 = src_idx0 + 1;
 
-            const auto tmp0 = tmp_zmm(src_idx0 - tmp_corr_idx);
-            const auto tmp1 = tmp_zmm(src_idx1 - tmp_corr_idx);
-            const auto src0 = src_zmm(src_idx0);
-            const auto src1 = src_zmm(src_idx1);
+                int next_src_idx0 = src_idx0 + 2;
+                int next_src_idx1 = src_idx1 + 2;
+                bool load_next = base_idx == 0 || i < 3;
 
-            if (next_src_idx0 < nrows && load_next) load(next_src_idx0);
-            valignd(tmp0, src0, src0, 0x1);
+                if (base_idx == 0 && i == 0) {
+                    load(src_idx0);
+                    load(src_idx1);
+                }
 
-            if (next_src_idx1 < nrows && load_next) load(next_src_idx1);
-            valignd(tmp1, src1, src1, 0xf);
+                const auto tmp0 = tmp_zmm(src_idx0 - tmp_corr_idx);
+                const auto tmp1 = tmp_zmm(src_idx1 - tmp_corr_idx);
+                const auto src0 = src_zmm(src_idx0);
+                const auto src1 = src_zmm(src_idx1);
 
-            vmovaps(src0 | kAAAA, tmp1);
-            vmovaps(src1 | k5555, tmp0);
+                if (next_src_idx0 < nrows && load_next) load(next_src_idx0);
+                valignd(tmp0, src0, src0, 0x1);
+
+                if (next_src_idx1 < nrows && load_next) load(next_src_idx1);
+                valignd(tmp1, src1, src1, 0xf);
+
+                vmovaps(src0 | kAAAA, tmp1);
+                vmovaps(src1 | k5555, tmp0);
+            }
         }
         // swap 2
         for (int i = 0; i < 4; i++) {
@@ -1974,7 +2140,8 @@ void jit_brgemm_matmul_copy_b_transposed_t::compute_N_loop(
     L(N_loop);
     compute_K_loop(false, curr_K_tail, is_first_K_iter, is_last_K_iter);
     add(reg_src_base, n_blk_step * src_stride);
-    add(reg_tr_src_base, n_blk_step * vnni_granularity * typesize);
+    add(reg_tr_src_base, n_blk_step * vnni_granularity * tr_typesize);
+
     if (req_zp_comp) add(reg_zp_comp_ptr, comp_shift);
     if (req_s8s8_comp) add(reg_comp_ptr, comp_shift);
 
@@ -1997,20 +2164,11 @@ void jit_brgemm_matmul_copy_b_transposed_t::generate() {
 
     preamble();
 
-    src_stride = (conf_->wei_tag == format_tag::adbc ? conf_->copy_B_wei_stride
-                                                     : conf_->K * typesize);
-    tr_src_stride = conf_->LDB * vnni_granularity * typesize;
-
     mov(reg_src_base, ptr[param1 + GET_OFF(src)]);
     mov(reg_tr_src_base, ptr[param1 + GET_OFF(tr_src)]);
     mov(reg_K_iters, ptr[param1 + GET_OFF(current_K_iters)]);
     mov(reg_N_iters, ptr[param1 + GET_OFF(current_N_blk)]);
 
-    auto kmovw = [=](Opmask k, unsigned w) {
-        mov(regw_tmp, w);
-        jit_generator::kmovw(k, regw_tmp);
-    };
-
     kmovw(k3333, 0x3333); // 0011001100110011
     kmovw(k5555, 0x5555); // 0101010101010101
     kmovw(kAAAA, 0xaaaa); // 1010101010101010
@@ -2115,7 +2273,7 @@ status_t create_brgemm_matmul_copy_b(
         CHECK(safe_ptr_assign(
                 copy_ker, new jit_brgemm_matmul_copy_b_transposed_t(conf)));
     } else {
-        if (is_bf16) {
+        if (is_bf16 || conf->is_bf32) {
             CHECK(safe_ptr_assign(
                     copy_ker, new jit_brgemm_matmul_copy_b_bf16_t(conf)));
         } else if (is_f32) {
