@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2021 Intel Corporation
+* Copyright 2021-2022 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -275,12 +275,42 @@ void rnn_brgemm_base_t::init_scratchpad(const cpu::rnn_utils::rnn_conf_t &rnn,
 
     using namespace memory_tracking::names;
 
-    if (rnn.is_int8_amx() || rnn.is_bf16_amx()) {
-        size_t n_elements = rnn.m_block * rnn.n_block;
+    if (rnn.is_cell_int8_amx() || rnn.is_cell_bf16_amx()) {
+        const auto m_block = rnn.merge_gemm_layer
+                ? nstl::max(rnn.m_block, rnn.mlayermerged_block)
+                : rnn.m_block;
+        size_t n_elements = m_block * rnn.n_block;
         scratchpad.book(key_brgemm_primitive_buffer, rnn.nthr * n_elements,
                 gemm_acc_type_size, gemm_acc_align);
     }
 
+    if (rnn.is_bf32()) {
+
+        const dims_t wei_layer_dims
+                = {rnn.n_layer, rnn.n_dir, rnn.n_gates, rnn.slc, rnn.dlc};
+        const dims_t wei_iter_dims
+                = {rnn.n_layer, rnn.n_dir, rnn.n_gates, rnn.sic, rnn.dic};
+
+        memory_desc_t wei_layer_desc;
+        const auto tag = rnn.n_block == 64 ? format_tag::ldgOI64o2i
+                                           : format_tag::ldgOI32o2i;
+        dnnl_memory_desc_init_by_tag(
+                &wei_layer_desc, 5, wei_layer_dims, data_type::bf16, tag);
+
+        memory_desc_t wei_iter_desc;
+        dnnl_memory_desc_init_by_tag(
+                &wei_iter_desc, 5, wei_iter_dims, data_type::bf16, tag);
+
+        scratchpad.book(key_rnn_bf32_wei_layer_trans,
+                memory_desc_wrapper(wei_layer_desc).size(), 64);
+
+        scratchpad.book(key_rnn_bf32_wei_iter_trans,
+                memory_desc_wrapper(wei_iter_desc).size(), 64);
+
+        scratchpad.book(key_rnn_bf32_attention_trans,
+                rnn.n_iter * rnn.mb * sizeof(bfloat16_t), 64);
+    }
+
     const int max_K_Block
             = nstl::max(rnn.KB1_blocks + 1,
                       nstl::max(rnn.KBproj_blocks + 1, rnn.KB2_blocks + 1))
@@ -292,14 +322,39 @@ void rnn_brgemm_base_t::init_scratchpad(const cpu::rnn_utils::rnn_conf_t &rnn,
 status_t rnn_brgemm_t<prop_kind::forward>::configure_brgemm(
         cpu::rnn_utils::rnn_conf_t &rnn, alg_kind_t cell_kind,
         dim_t src_layer_type_size, dim_t scratch_type_size) {
+    using namespace cpu::rnn_utils;
 
     rnn.M = rnn.mb;
     rnn.N = rnn.dhc;
     rnn.K1 = rnn.slc;
     rnn.K2 = rnn.sic;
+    const auto is_int8 = rnn.is_cell_dt_int8();
+    const auto is_bf16 = rnn.is_cell_dt_bf16();
+
+    const dim_t padding = (is_int8 ? 4 : (is_bf16 ? 2 : 1));
+    rnn.K1padded = utils::rnd_up(rnn.K1, padding);
+    rnn.K2padded = utils::rnd_up(rnn.K2, padding);
+
+    rnn.brgemm_isa = brgemm_calc_isa(rnn.K1, rnn.K2, is_int8, is_bf16);
+    const int bf32_reduction_dim_threshold = 128;
+    const bool is_shape_ok_for_bf32 = rnn.K1 >= bf32_reduction_dim_threshold
+            && rnn.K2 >= bf32_reduction_dim_threshold;
+    const bool is_bf32 = is_bf16 && rnn.brgemm_isa == avx512_core_bf16_amx_bf16
+            && rnn.dt_conf == all_f32
+            // workspace data type and layouts can differ between fwd and bwd
+            // implementations during training.
+            && !rnn.is_training
+            // bf16 lstm_projection is not supported, so neither is bf32.
+            && !rnn.is_lstm_projection && is_shape_ok_for_bf32;
+    if (!IMPLICATION(rnn.is_cell_dt_bf16(), rnn.is_bf16_conf() || is_bf32))
+        return status::unimplemented;
 
     rnn.nthr = dnnl_get_max_threads();
-    rnn.n_block = 32;
+    const bool is_amx_isa_selected
+            = rnn.is_cell_int8_amx() || rnn.is_cell_bf16_amx();
+    const bool can_use_block64
+            = is_amx_isa_selected && rnn.N % 64 == 0 && !rnn.is_lstm_projection;
+    rnn.n_block = can_use_block64 ? 64 : 32;
     rnn.N_blocks = utils::div_up(rnn.N, rnn.n_block);
     rnn.n_tail = rnn.N % rnn.n_block;
 
@@ -313,32 +368,26 @@ status_t rnn_brgemm_t<prop_kind::forward>::configure_brgemm(
     const dim_t Cs
             = scratch_type_size * (rnn.n_gates + 1) * (rnn.M * rnn.n_block);
 
-    const auto is_int8 = rnn.is_int8();
-    const auto is_bf16 = rnn.is_bf16();
-
-    const dim_t padding = (is_int8 ? 4 : (is_bf16 ? 2 : 1));
-    rnn.K1padded = utils::rnd_up(rnn.K1, padding);
-    rnn.K2padded = utils::rnd_up(rnn.K2, padding);
-
-    rnn.brgemm_isa = brgemm_calc_isa(rnn.K1, rnn.K2, is_int8, is_bf16);
-
     std::tie(rnn.k1_block, rnn.k2_block) = brgemm_calc_k_block(rnn.K1, rnn.K2,
             rnn.M, rnn.n_block, cell_kind, src_layer_type_size, As, Bs, Cs,
-            l2_cache_size, rnn.brgemm_isa, rnn.is_int8(), rnn.is_bf16());
+            l2_cache_size, rnn.brgemm_isa, is_int8, is_bf16);
     rnn.KB1_blocks = rnn.K1 / rnn.k1_block;
     rnn.k1_tail = rnn.K1 % rnn.k1_block;
     rnn.KB2_blocks = rnn.K2 / rnn.k2_block;
     rnn.k2_tail = rnn.K2 % rnn.k2_block;
     rnn.m_block = brgemm_calc_m_block(cell_kind, prop_kind::forward, rnn.nthr,
-            rnn.M, rnn.N_blocks, rnn.is_f32(), rnn.is_int8_amx(),
-            rnn.is_bf16_amx(), work_by_N, As, Bs, Cs, l2_cache_size);
+            rnn.M, rnn.N_blocks, rnn.is_cell_dt_f32(), rnn.is_cell_int8_amx(),
+            rnn.is_cell_bf16_amx(), work_by_N, As, Bs, Cs, l2_cache_size);
 
     rnn.M_blocks = rnn.M / rnn.m_block;
 
     rnn.brgemm_isa = adjust_isa_by_m_block(
-            rnn.brgemm_isa, rnn.m_block, rnn.is_int8_amx());
-    rnn.unfused_post_gemm
-            = cell_kind == alg_kind::vanilla_lstm ? (rnn.M_blocks == 1) : false;
+            rnn.brgemm_isa, rnn.m_block, rnn.is_cell_int8_amx());
+    // Unfused post-gemm for lstm cell allows to parallelize across gates loop
+    // and reduces brgemm problem size for the single iteration of parallel loop
+    rnn.unfused_post_gemm = cell_kind == alg_kind::vanilla_lstm
+            ? IMPLICATION(rnn.M_blocks > 1, rnn.is_cell_bf16_amx())
+            : false;
 
     rnn.LDA1[0] = rnn.src_layer_ld_;
     rnn.LDA1[1] = rnn.dst_iter_ld_;
@@ -348,7 +397,10 @@ status_t rnn_brgemm_t<prop_kind::forward>::configure_brgemm(
     rnn.LDA2[1] = rnn.dst_layer_ld_;
     rnn.LDA2[2] = rnn.ws_states_iter_ld;
 
-    rnn.brgemm_fwd_iter_layer_fuse_possible = rnn.slc == rnn.sic;
+    rnn.LDA2_2[0] = rnn.dst_layer_ld_;
+    rnn.LDA2_2[1] = rnn.dst_iter_ld_;
+    rnn.LDA2_2[2] = rnn.ws_states_layer_ld;
+    rnn.LDA2_2[3] = rnn.ws_states_iter_ld;
 
     rnn.LDB1 = rnn.n_block;
     rnn.LDB2 = rnn.n_block;
@@ -382,8 +434,8 @@ status_t rnn_brgemm_t<prop_kind::forward>::configure_brgemm(
 
         rnn.Kproj = rnn.dhc;
         rnn.Kprojpadded = utils::rnd_up(rnn.Kproj, padding);
-        if (rnn.is_int8_amx() || rnn.is_bf16_amx()) {
-            const dim_t max_row_width = rnn.is_int8_amx() ? 64 : 32;
+        if (rnn.is_cell_int8_amx() || rnn.is_cell_bf16_amx()) {
+            const dim_t max_row_width = rnn.is_cell_int8_amx() ? 64 : 32;
             rnn.kproj_block = nstl::min(rnn.Kproj, (dim_t)max_row_width);
 
             rnn.KBproj_blocks = rnn.Kproj / rnn.kproj_block;
@@ -392,11 +444,12 @@ status_t rnn_brgemm_t<prop_kind::forward>::configure_brgemm(
             if ((rnn.kproj_tail % padding) || (rnn.kproj_block % padding)) {
                 rnn.kproj_block = rnn.Kproj;
                 rnn.kproj_tail = 0;
-                rnn.brgemm_isa = rnn.is_int8() ? x64::avx512_core_vnni
-                                               : x64::avx512_core_bf16;
+                rnn.brgemm_isa = rnn.is_cell_dt_int8() ? x64::avx512_core_vnni
+                                                       : x64::avx512_core_bf16;
             } else {
-                rnn.brgemm_isa = rnn.is_int8() ? x64::avx512_core_bf16_amx_int8
-                                               : x64::avx512_core_bf16_amx_bf16;
+                rnn.brgemm_isa = rnn.is_cell_dt_int8()
+                        ? x64::avx512_core_bf16_amx_int8
+                        : x64::avx512_core_bf16_amx_bf16;
             }
         } else {
             rnn.kproj_block = rnn.Kproj;
@@ -428,6 +481,51 @@ status_t rnn_brgemm_t<prop_kind::forward>::configure_brgemm(
                 || rnn.LDBproj < get_dim(n_block, n_tail) || check_LDC)
             return status::unimplemented;
     }
+
+    // enable merged across n_iter dimension layer part of the cell computation
+    // TODO: extend coverage for other problem types
+    const bool mlc_cell_type_ok = cell_kind == alg_kind::vanilla_lstm
+            && !rnn.is_lstm_projection && !rnn.is_lstm_peephole;
+    const int mlc_mb_max_threshold = 1;
+    const int mlc_n_iter_min_threshold = 2;
+    const int mlc_n_layer_max_threshold = 1;
+    const bool mlc_problem_shape_ok = rnn.mb <= mlc_mb_max_threshold
+            && rnn.n_iter >= mlc_n_iter_min_threshold
+            && rnn.n_layer <= mlc_n_layer_max_threshold;
+    // if rnn.skip_dst_iter_copy() == false we might need to reduce number of
+    // merged cells by 1 (rnn.n_iter - 1)
+    // but if in addition rnn.skip_src_layer_copy() == true then on cell
+    // position 'first_layer' we still should merge for all the n_iter cells
+    // so current support is limited by the case when layer computation is
+    // merged for all rnn.n_iter cells
+    const bool mlc_m_dim_adjustment_not_required
+            = IMPLICATION(rnn.skip_dst_iter_copy(),
+                    rnn.skip_src_layer_copy() && rnn.n_layer == 1);
+    const bool merged_layer_compute_applicable = rnn.src_layer_is_trivial_stride
+            && mlc_cell_type_ok && mlc_problem_shape_ok
+            && mlc_m_dim_adjustment_not_required;
+    if (merged_layer_compute_applicable) {
+        rnn.merge_gemm_layer = true;
+
+        // required adjustment if mlc_m_dim_adjustment_not_required = false
+        const int n_iters_to_merge = rnn.n_iter;
+        rnn.Mlayermerged = rnn.mb * n_iters_to_merge;
+        rnn.mlayermerged_block = brgemm_calc_m_block(cell_kind,
+                prop_kind::forward, rnn.nthr, rnn.Mlayermerged, rnn.N_blocks,
+                rnn.is_cell_dt_f32(), rnn.is_cell_int8_amx(),
+                rnn.is_cell_bf16_amx(), work_by_N, As, Bs, Cs, l2_cache_size);
+
+        rnn.Mlayermerged_blocks = rnn.Mlayermerged / rnn.mlayermerged_block;
+    }
+
+    rnn.brgemm_fwd_iter_layer_fuse_possible
+            = rnn.slc == rnn.sic && !rnn.merge_gemm_layer;
+
+    if (!rnn.is_orig_gru) {
+        rnn.loop_order = rnn.is_cell_int8_amx() || rnn.is_cell_bf16_amx()
+                ? brgemm_rnn_execute_loop_order_t::mblk_nblk
+                : brgemm_rnn_execute_loop_order_t::nblk_mblk;
+    }
     return status::success;
 }
 
@@ -501,9 +599,17 @@ status_t rnn_brgemm_t<prop_kind::forward>::init_kernels(
     const int max_bs_factor = rnn.brgemm_fwd_iter_layer_fuse_possible ? 2 : 1;
 
     for (int i = 0; i < num_base_kernels_; i++) {
-        init_brgemm(&desc_layer_b0_[i], rnn.brgemm_isa, kernel_layer_b0_[i],
-                rnn.m_block, brgemm_n, rnn.k1_block, rnn.LDA1[i], rnn.LDB1,
-                rnn.LDC, 0.0, max_bs_factor * rnn.KB1_blocks);
+        if (rnn.merge_gemm_layer) {
+            init_brgemm(&desc_layermerged_b0_[i], rnn.brgemm_isa,
+                    kernel_layermerged_b0_[i], rnn.mlayermerged_block, brgemm_n,
+                    rnn.k1_block, rnn.LDA1[i], rnn.LDB1, rnn.LDC, 0.0,
+                    rnn.KB1_blocks);
+        } else {
+            init_brgemm(&desc_layer_b0_[i], rnn.brgemm_isa, kernel_layer_b0_[i],
+                    rnn.m_block, brgemm_n, rnn.k1_block, rnn.LDA1[i], rnn.LDB1,
+                    rnn.LDC, 0.0, max_bs_factor * rnn.KB1_blocks);
+        }
+
         init_brgemm(&desc_iter_b0_[i], rnn.brgemm_isa, kernel_iter_b0_[i],
                 rnn.m_block, brgemm_n, rnn.k2_block, rnn.LDA2[i], rnn.LDB2,
                 rnn.LDC, 0.0, rnn.KB2_blocks);
@@ -511,10 +617,18 @@ status_t rnn_brgemm_t<prop_kind::forward>::init_kernels(
                 rnn.m_block, brgemm_n, rnn.k2_block, rnn.LDA2[i], rnn.LDB2,
                 rnn.LDC, 1.0, rnn.KB2_blocks);
         if (rnn.n_tail) {
-            init_brgemm(&desc_layer_N_tail_b0_[i], rnn.brgemm_isa,
-                    kernel_layer_N_tail_b0_[i], rnn.m_block, brgemm_n_tail,
-                    rnn.k1_block, rnn.LDA1[i], rnn.LDB1, rnn.LDC, 0.0,
-                    max_bs_factor * rnn.KB1_blocks);
+            if (rnn.merge_gemm_layer) {
+                init_brgemm(&desc_layermerged_N_tail_b0_[i], rnn.brgemm_isa,
+                        kernel_layermerged_N_tail_b0_[i],
+                        rnn.mlayermerged_block, brgemm_n_tail, rnn.k1_block,
+                        rnn.LDA1[i], rnn.LDB1, rnn.LDC, 0.0, rnn.KB1_blocks);
+            } else {
+                init_brgemm(&desc_layer_N_tail_b0_[i], rnn.brgemm_isa,
+                        kernel_layer_N_tail_b0_[i], rnn.m_block, brgemm_n_tail,
+                        rnn.k1_block, rnn.LDA1[i], rnn.LDB1, rnn.LDC, 0.0,
+                        max_bs_factor * rnn.KB1_blocks);
+            }
+
             init_brgemm(&desc_iter_N_tail_b0_[i], rnn.brgemm_isa,
                     kernel_iter_N_tail_b0_[i], rnn.m_block, brgemm_n_tail,
                     rnn.k2_block, rnn.LDA2[i], rnn.LDB2, rnn.LDC, 0.0,
@@ -524,25 +638,62 @@ status_t rnn_brgemm_t<prop_kind::forward>::init_kernels(
                     rnn.k2_block, rnn.LDA2[i], rnn.LDB2, rnn.LDC, 1.0,
                     rnn.KB2_blocks);
         }
-        if (rnn.k1_tail)
-            init_brgemm(&desc_layer_K1_tail_b1_[i], rnn.brgemm_isa,
-                    kernel_layer_K1_tail_b1_[i], rnn.m_block, brgemm_n,
-                    rnn.k1_tail, rnn.LDA1[i], rnn.LDB1, rnn.LDC, 1.0,
-                    max_bs_factor * 1);
+        if (rnn.k1_tail) {
+            if (rnn.merge_gemm_layer) {
+                init_brgemm(&desc_layermerged_K1_tail_b1_[i], rnn.brgemm_isa,
+                        kernel_layermerged_K1_tail_b1_[i],
+                        rnn.mlayermerged_block, brgemm_n, rnn.k1_tail,
+                        rnn.LDA1[i], rnn.LDB1, rnn.LDC, 1.0, 1);
+            } else {
+                init_brgemm(&desc_layer_K1_tail_b1_[i], rnn.brgemm_isa,
+                        kernel_layer_K1_tail_b1_[i], rnn.m_block, brgemm_n,
+                        rnn.k1_tail, rnn.LDA1[i], rnn.LDB1, rnn.LDC, 1.0,
+                        max_bs_factor * 1);
+            }
+        }
         if (rnn.k2_tail)
             init_brgemm(&desc_iter_K2_tail_b1_[i], rnn.brgemm_isa,
                     kernel_iter_K2_tail_b1_[i], rnn.m_block, brgemm_n,
                     rnn.k2_tail, rnn.LDA2[i], rnn.LDB2, rnn.LDC, 1.0, 1);
-        if (rnn.k1_tail && rnn.n_tail)
-            init_brgemm(&desc_layer_NK1_tail_b1_[i], rnn.brgemm_isa,
-                    kernel_layer_NK1_tail_b1_[i], rnn.m_block, brgemm_n_tail,
-                    rnn.k1_tail, rnn.LDA1[i], rnn.LDB1, rnn.LDC, 1.0,
-                    max_bs_factor * 1);
+        if (rnn.k1_tail && rnn.n_tail) {
+            if (rnn.merge_gemm_layer) {
+                init_brgemm(&desc_layermerged_NK1_tail_b1_[i], rnn.brgemm_isa,
+                        kernel_layermerged_NK1_tail_b1_[i],
+                        rnn.mlayermerged_block, brgemm_n_tail, rnn.k1_tail,
+                        rnn.LDA1[i], rnn.LDB1, rnn.LDC, 1.0, 1);
+            } else {
+                init_brgemm(&desc_layer_NK1_tail_b1_[i], rnn.brgemm_isa,
+                        kernel_layer_NK1_tail_b1_[i], rnn.m_block,
+                        brgemm_n_tail, rnn.k1_tail, rnn.LDA1[i], rnn.LDB1,
+                        rnn.LDC, 1.0, max_bs_factor * 1);
+            }
+        }
         if (rnn.k2_tail && rnn.n_tail)
             init_brgemm(&desc_iter_NK2_tail_b1_[i], rnn.brgemm_isa,
                     kernel_iter_NK2_tail_b1_[i], rnn.m_block, brgemm_n_tail,
                     rnn.k2_tail, rnn.LDA2[i], rnn.LDB2, rnn.LDC, 1.0, 1);
     }
+    if (rnn.is_orig_gru) {
+        for (int i = 0; i < num_vanilla_gru_iter_part2_kernels_; i++) {
+            init_brgemm(&desc_iter_p2_b1_[i], rnn.brgemm_isa,
+                    kernel_iter_p2_b1_[i], rnn.m_block, brgemm_n, rnn.k2_block,
+                    rnn.LDA2_2[i], rnn.LDB2, rnn.LDC, 1.0, rnn.KB2_blocks);
+            if (rnn.n_tail)
+                init_brgemm(&desc_iter_p2_N_tail_b1_[i], rnn.brgemm_isa,
+                        kernel_iter_p2_N_tail_b1_[i], rnn.m_block,
+                        brgemm_n_tail, rnn.k2_block, rnn.LDA2_2[i], rnn.LDB2,
+                        rnn.LDC, 1.0, rnn.KB2_blocks);
+            if (rnn.k2_tail)
+                init_brgemm(&desc_iter_p2_K2_tail_b1_[i], rnn.brgemm_isa,
+                        kernel_iter_p2_K2_tail_b1_[i], rnn.m_block, brgemm_n,
+                        rnn.k2_tail, rnn.LDA2_2[i], rnn.LDB2, rnn.LDC, 1.0, 1);
+            if (rnn.k2_tail && rnn.n_tail)
+                init_brgemm(&desc_iter_p2_NK2_tail_b1_[i], rnn.brgemm_isa,
+                        kernel_iter_p2_NK2_tail_b1_[i], rnn.m_block,
+                        brgemm_n_tail, rnn.k2_tail, rnn.LDA2_2[i], rnn.LDB2,
+                        rnn.LDC, 1.0, 1);
+        }
+    }
     if (rnn.is_lstm_projection) {
         const dim_t brgemm_np = nstl::min(rnn.Nproj, rnn.n_block);
         const dim_t brgemm_np_tail = nstl::min(rnn.Nproj, rnn.nproj_tail);
@@ -563,7 +714,7 @@ status_t rnn_brgemm_t<prop_kind::forward>::init_kernels(
                         rnn.kproj_block, rnn.LDAproj, rnn.LDBproj,
                         rnn.LDCproj[i], 1.0, rnn.KBproj_blocks);
             }
-            if (rnn.is_int8_amx() || rnn.is_bf16_amx()) {
+            if (rnn.is_cell_int8_amx() || rnn.is_cell_bf16_amx()) {
                 if (rnn.kproj_tail)
                     init_brgemm(&desc_proj_K_tail_b1_[i], rnn.brgemm_isa,
                             kernel_proj_K_tail_b1_[i], rnn.m_block, brgemm_np,
@@ -577,25 +728,44 @@ status_t rnn_brgemm_t<prop_kind::forward>::init_kernels(
             }
         }
     }
-    if (rnn.is_int8_amx() || rnn.is_bf16_amx()) {
-        CHECK(brgemm_rnn_init_tiles(desc_layer_b0_, pallete_buff_layer_));
+
+    if (rnn.is_cell_int8_amx() || rnn.is_cell_bf16_amx()) {
+        if (rnn.merge_gemm_layer)
+            CHECK(brgemm_rnn_init_tiles(
+                    desc_layermerged_b0_, pallete_buff_layermerged_));
+        else
+            CHECK(brgemm_rnn_init_tiles(desc_layer_b0_, pallete_buff_layer_));
         CHECK(brgemm_rnn_init_tiles(desc_iter_b0_, pallete_buff_iter_));
 
         if (rnn.n_tail) {
-            CHECK(brgemm_rnn_init_tiles(
-                    desc_layer_N_tail_b0_, pallete_buff_layer_n_tail_));
+            if (rnn.merge_gemm_layer)
+                CHECK(brgemm_rnn_init_tiles(desc_layermerged_N_tail_b0_,
+                        pallete_buff_layermerged_n_tail_));
+            else
+                CHECK(brgemm_rnn_init_tiles(
+                        desc_layer_N_tail_b0_, pallete_buff_layer_n_tail_));
             CHECK(brgemm_rnn_init_tiles(
                     desc_iter_N_tail_b0_, pallete_buff_iter_n_tail_));
         }
-        if (rnn.k1_tail)
-            CHECK(brgemm_rnn_init_tiles(
-                    desc_layer_K1_tail_b1_, pallete_buff_k1_tail_));
+        if (rnn.k1_tail) {
+            if (rnn.merge_gemm_layer)
+                CHECK(brgemm_rnn_init_tiles(desc_layermerged_K1_tail_b1_,
+                        pallete_buff_layermerged_k1_tail_));
+            else
+                CHECK(brgemm_rnn_init_tiles(
+                        desc_layer_K1_tail_b1_, pallete_buff_k1_tail_));
+        }
         if (rnn.k2_tail)
             CHECK(brgemm_rnn_init_tiles(
                     desc_iter_K2_tail_b1_, pallete_buff_k2_tail_));
-        if (rnn.k1_tail && rnn.n_tail)
-            CHECK(brgemm_rnn_init_tiles(
-                    desc_layer_NK1_tail_b1_, pallete_buff_nk1_tail_));
+        if (rnn.k1_tail && rnn.n_tail) {
+            if (rnn.merge_gemm_layer)
+                CHECK(brgemm_rnn_init_tiles(desc_layermerged_NK1_tail_b1_,
+                        pallete_buff_layermerged_nk1_tail_));
+            else
+                CHECK(brgemm_rnn_init_tiles(
+                        desc_layer_NK1_tail_b1_, pallete_buff_nk1_tail_));
+        }
         if (rnn.k2_tail && rnn.n_tail)
             CHECK(brgemm_rnn_init_tiles(
                     desc_iter_NK2_tail_b1_, pallete_buff_nk2_tail_));
@@ -628,7 +798,8 @@ void rnn_brgemm_t<prop_kind::backward>::init_scratchpad(
     using namespace memory_tracking::names;
 
     // init scratchpad for internal reorders:
-    const auto data_size = rnn.is_bf16() ? sizeof(bfloat16_t) : sizeof(float);
+    const auto data_size
+            = rnn.is_bf16_conf() ? sizeof(bfloat16_t) : sizeof(float);
     const auto &d_wei = rnn.diff_wei_brgemm;
     const auto scratch_gates_blocked_per_thr = d_wei.Kpadded * d_wei.n_block;
     const auto scratch_gates_blocked_size
@@ -652,6 +823,10 @@ void rnn_brgemm_t<prop_kind::backward>::init_scratchpad(
 status_t rnn_brgemm_t<prop_kind::backward>::configure_brgemm(
         cpu::rnn_utils::rnn_conf_t &rnn, alg_kind_t cell_kind,
         dim_t src_layer_type_size, dim_t scratch_type_size) {
+    using namespace cpu::rnn_utils;
+
+    if (rnn.is_int8_conf() || rnn.is_cell_dt_int8())
+        return status::unimplemented;
 
     auto &diff_src_conf = rnn.diff_src_brgemm;
 
@@ -683,31 +858,32 @@ status_t rnn_brgemm_t<prop_kind::backward>::configure_brgemm(
     const dim_t Cs = scratch_type_size * (rnn.n_gates + 1)
             * (diff_src_conf.M * diff_src_conf.n_block);
 
-    const auto is_int8 = rnn.is_int8();
-    const auto is_bf16 = rnn.is_bf16();
+    const auto is_bf16 = rnn.is_cell_dt_bf16();
 
-    const dim_t padding = (is_int8 ? 4 : (is_bf16 ? 2 : 1));
+    const dim_t padding = is_bf16 ? 2 : 1;
     diff_src_conf.Kpadded = utils::rnd_up(diff_src_conf.K, padding);
 
-    diff_src_conf.isa = brgemm_calc_isa(
-            diff_src_conf.K, diff_src_conf.K, is_int8, is_bf16);
+    diff_src_conf.isa
+            = brgemm_calc_isa(diff_src_conf.K, diff_src_conf.K, false, is_bf16);
+    const bool is_bf16_amx
+            = is_bf16 && diff_src_conf.isa == x64::avx512_core_bf16_amx_bf16;
+    const bool split_gates_computation = is_bf16_amx && diff_src_conf.K >= 1024
+            && diff_src_conf.n_tail == 0;
+    diff_src_conf.gates_block = split_gates_computation ? 1 : rnn.n_gates;
 
     std::tie(diff_src_conf.k_block, std::ignore) = brgemm_calc_k_block(
             diff_src_conf.K, diff_src_conf.K, diff_src_conf.M,
             diff_src_conf.n_block, cell_kind, src_layer_type_size, As, Bs, Cs,
-            l2_cache_size, diff_src_conf.isa, rnn.is_int8(), rnn.is_bf16());
+            l2_cache_size, diff_src_conf.isa, false, rnn.is_cell_dt_bf16());
 
     diff_src_conf.K_blocks = diff_src_conf.K / diff_src_conf.k_block;
     diff_src_conf.K_blocks *= rnn.n_gates;
     diff_src_conf.k_tail = diff_src_conf.K % diff_src_conf.k_block;
 
-    const bool is_int8_amx = rnn.is_int8()
-            && diff_src_conf.isa == x64::avx512_core_bf16_amx_int8;
-    const bool is_bf16_amx = rnn.is_bf16()
-            && diff_src_conf.isa == x64::avx512_core_bf16_amx_bf16;
     diff_src_conf.m_block = brgemm_calc_m_block(cell_kind, prop_kind::backward,
-            rnn.nthr, diff_src_conf.M, diff_src_conf.N_blocks, rnn.is_f32(),
-            is_int8_amx, is_bf16_amx, work_by_N, As, Bs, Cs, l2_cache_size);
+            rnn.nthr, diff_src_conf.M, diff_src_conf.N_blocks,
+            rnn.is_cell_dt_f32(), false, is_bf16_amx, work_by_N, As, Bs, Cs,
+            l2_cache_size);
 
     diff_src_conf.M_blocks = diff_src_conf.M / diff_src_conf.m_block;
     diff_src_conf.LDA = rnn.scratch_gates_ld;
@@ -735,7 +911,15 @@ status_t rnn_brgemm_t<prop_kind::backward>::configure_brgemm(
             ? utils::rnd_up(rnn.mb, 2)
             : rnn.mb;
     diff_wei_conf.Kpadded = utils::rnd_up(diff_wei_conf.K, padding);
-    diff_wei_conf.n_block = 32;
+
+    diff_wei_conf.isa
+            = brgemm_calc_isa(diff_wei_conf.K, diff_wei_conf.K, false, is_bf16);
+
+    const bool is_wei_bf16_amx = rnn.is_cell_dt_bf16()
+            && diff_wei_conf.isa == x64::avx512_core_bf16_amx_bf16;
+    const bool diff_wei_can_use_nblock64 = is_wei_bf16_amx
+            && diff_wei_conf.N % 64 == 0 && !rnn.is_lstm_peephole;
+    diff_wei_conf.n_block = diff_wei_can_use_nblock64 ? 64 : 32;
     diff_wei_conf.N_blocks
             = utils::div_up(diff_wei_conf.N, diff_wei_conf.n_block);
     diff_wei_conf.n_tail = diff_wei_conf.N % diff_wei_conf.n_block;
@@ -747,22 +931,15 @@ status_t rnn_brgemm_t<prop_kind::backward>::configure_brgemm(
     const dim_t Cs_wei = scratch_type_size * (rnn.n_gates + 1)
             * (diff_wei_conf.M * diff_wei_conf.n_block);
 
-    diff_wei_conf.isa = brgemm_calc_isa(
-            diff_wei_conf.K, diff_wei_conf.K, is_int8, is_bf16);
-
     std::tie(diff_wei_conf.k_block, std::ignore)
             = brgemm_calc_k_block(diff_wei_conf.K, diff_wei_conf.K,
                     diff_wei_conf.M, diff_wei_conf.n_block, cell_kind,
                     src_layer_type_size, As_wei, Bs_wei, Cs_wei, l2_cache_size,
-                    diff_wei_conf.isa, rnn.is_int8(), rnn.is_bf16());
+                    diff_wei_conf.isa, false, rnn.is_cell_dt_bf16());
 
     diff_wei_conf.K_blocks = diff_wei_conf.K / diff_wei_conf.k_block;
     diff_wei_conf.k_tail = diff_wei_conf.K % diff_wei_conf.k_block;
 
-    const bool is_wei_int8_amx = rnn.is_int8()
-            && diff_wei_conf.isa == x64::avx512_core_bf16_amx_int8;
-    const bool is_wei_bf16_amx = rnn.is_bf16()
-            && diff_wei_conf.isa == x64::avx512_core_bf16_amx_bf16;
     if (diff_wei_conf.M_iter != diff_wei_conf.M_layer) {
         diff_wei_conf.m_block = diff_wei_conf.M;
         diff_wei_conf.M_blocks = 1;
@@ -772,9 +949,9 @@ status_t rnn_brgemm_t<prop_kind::backward>::configure_brgemm(
 
         diff_wei_conf.m_block
                 = brgemm_calc_m_block(cell_kind, prop_kind::backward, rnn.nthr,
-                        diff_wei_conf.M, diff_wei_conf.N_blocks, rnn.is_f32(),
-                        is_wei_int8_amx, is_wei_bf16_amx, work_by_N_wei, As_wei,
-                        Bs_wei, Cs_wei, l2_cache_size);
+                        diff_wei_conf.M, diff_wei_conf.N_blocks,
+                        rnn.is_cell_dt_f32(), false, is_wei_bf16_amx,
+                        work_by_N_wei, As_wei, Bs_wei, Cs_wei, l2_cache_size);
         diff_wei_conf.M_blocks = diff_wei_conf.M / diff_wei_conf.m_block;
     }
 
@@ -816,6 +993,17 @@ status_t rnn_brgemm_t<prop_kind::backward>::configure_brgemm(
         rnn.brgemm_isa = diff_wei_conf.isa;
     }
 
+    if (!rnn.is_orig_gru) {
+        rnn.diff_src_brgemm.loop_order
+                = is_bf16 && diff_src_conf.isa == x64::avx512_core_bf16_amx_bf16
+                ? brgemm_rnn_execute_loop_order_t::mblk_nblk
+                : brgemm_rnn_execute_loop_order_t::nblk_mblk;
+        rnn.diff_wei_brgemm.loop_order
+                = is_bf16 && diff_wei_conf.isa == x64::avx512_core_bf16_amx_bf16
+                ? brgemm_rnn_execute_loop_order_t::mblk_nblk
+                : brgemm_rnn_execute_loop_order_t::nblk_mblk;
+    }
+
     return status::success;
 }
 
@@ -913,25 +1101,46 @@ static status_t init_kernels_diff_src(rnn_diff_src_brgemm_t &diff_src,
     const int n_diff_src_layer_tail
             = nstl::min(diff_src_conf.N_layer, diff_src_conf.n_layer_tail);
     const auto K_batch_size = rnn.n_gates * diff_src_conf.K_blocks;
-
+    const auto split_gates_computation
+            = diff_src_conf.gates_block != rnn.n_gates;
     init_brgemm_diff_src(&diff_src.desc_iter_layer_beta0_, diff_src_conf.isa,
             diff_src.kernel_iter_layer_beta0_, diff_src_conf.m_block,
             n_diff_src, diff_src_conf.k_block, diff_src_conf.LDA,
             diff_src_conf.LDB, diff_src_conf.LDC, 0.0, K_batch_size);
+    if (split_gates_computation)
+        init_brgemm_diff_src(&diff_src.desc_iter_layer_beta1_,
+                diff_src_conf.isa, diff_src.kernel_iter_layer_beta1_,
+                diff_src_conf.m_block, n_diff_src, diff_src_conf.k_block,
+                diff_src_conf.LDA, diff_src_conf.LDB, diff_src_conf.LDC, 1.0,
+                K_batch_size);
 
-    if (n_diff_src_layer_tail)
+    if (n_diff_src_layer_tail) {
         init_brgemm_diff_src(&diff_src.desc_layer_N_tail_beta0_,
                 diff_src_conf.isa, diff_src.kernel_layer_N_tail_beta0_,
                 diff_src_conf.m_block, n_diff_src_layer_tail,
                 diff_src_conf.k_block, diff_src_conf.LDA, diff_src_conf.LDB,
                 diff_src_conf.LDC, 0.0, K_batch_size);
+        if (split_gates_computation)
+            init_brgemm_diff_src(&diff_src.desc_layer_N_tail_beta1_,
+                    diff_src_conf.isa, diff_src.kernel_layer_N_tail_beta1_,
+                    diff_src_conf.m_block, n_diff_src_layer_tail,
+                    diff_src_conf.k_block, diff_src_conf.LDA, diff_src_conf.LDB,
+                    diff_src_conf.LDC, 1.0, K_batch_size);
+    }
 
-    if (n_diff_src_iter_tail)
+    if (n_diff_src_iter_tail) {
         init_brgemm_diff_src(&diff_src.desc_iter_N_tail_beta0_,
                 diff_src_conf.isa, diff_src.kernel_iter_N_tail_beta0_,
                 diff_src_conf.m_block, n_diff_src_iter_tail,
                 diff_src_conf.k_block, diff_src_conf.LDA, diff_src_conf.LDB,
                 diff_src_conf.LDC, 0.0, K_batch_size);
+        if (split_gates_computation)
+            init_brgemm_diff_src(&diff_src.desc_iter_N_tail_beta1_,
+                    diff_src_conf.isa, diff_src.kernel_iter_N_tail_beta1_,
+                    diff_src_conf.m_block, n_diff_src_iter_tail,
+                    diff_src_conf.k_block, diff_src_conf.LDA, diff_src_conf.LDB,
+                    diff_src_conf.LDC, 1.0, K_batch_size);
+    }
 
     if (diff_src_conf.k_tail) {
         init_brgemm_diff_src(&diff_src.desc_iter_layer_K_tail_beta1_,
@@ -957,7 +1166,7 @@ static status_t init_kernels_diff_src(rnn_diff_src_brgemm_t &diff_src,
         }
     }
 
-    const bool is_bf16_amx = rnn.is_bf16()
+    const bool is_bf16_amx = rnn.is_cell_dt_bf16()
             && diff_src_conf.isa == x64::avx512_core_bf16_amx_bf16;
 
     if (is_bf16_amx) {
@@ -1067,7 +1276,7 @@ static status_t init_kernels_diff_wei(rnn_diff_wei_brgemm_t &diff_wei,
                 diff_wei_conf.LDC_layer, 1.0, 1);
     }
 
-    const bool is_bf16_amx_wei = rnn.is_bf16()
+    const bool is_bf16_amx_wei = rnn.is_cell_dt_bf16()
             && diff_wei_conf.isa == x64::avx512_core_bf16_amx_bf16;
 
     if (is_bf16_amx_wei) {
@@ -1097,6 +1306,29 @@ static status_t init_kernels_diff_wei(rnn_diff_wei_brgemm_t &diff_wei,
         }
     }
 
+    // Creating temporary matmul configuration descriptor to use copy_B jit
+    // kernels from brgemm matmul copy routines for reordering scratch gates in
+    // diff_wei rnn brgemm implementation.
+    // TODO: provide unification of jit-based copy routines with implementation
+    // independent interface
+    matmul::brgemm_matmul_conf_t tmp_matmul_conf_for_reorder;
+    tmp_matmul_conf_for_reorder.wei_tag = format_tag::ab;
+    tmp_matmul_conf_for_reorder.N = rnn.scratch_gates_ld;
+    tmp_matmul_conf_for_reorder.K = rnn.mb;
+    tmp_matmul_conf_for_reorder.wei_n_blk = tmp_matmul_conf_for_reorder.N_blk
+            = diff_wei_conf.n_block;
+    tmp_matmul_conf_for_reorder.N_tail = diff_wei_conf.n_tail;
+    tmp_matmul_conf_for_reorder.LDB = diff_wei_conf.LDB;
+    tmp_matmul_conf_for_reorder.src_dt = tmp_matmul_conf_for_reorder.wei_dt
+            = rnn.is_cell_dt_bf16() ? data_type::bf16 : data_type::f32;
+    tmp_matmul_conf_for_reorder.a_dt_sz = tmp_matmul_conf_for_reorder.tr_a_dt_sz
+            = types::data_type_size(tmp_matmul_conf_for_reorder.src_dt);
+    tmp_matmul_conf_for_reorder.b_dt_sz = tmp_matmul_conf_for_reorder.tr_b_dt_sz
+            = types::data_type_size(tmp_matmul_conf_for_reorder.wei_dt);
+    CHECK(matmul::create_brgemm_matmul_copy_b(
+            diff_wei.srcatch_gates_reorder_kernel_,
+            &tmp_matmul_conf_for_reorder));
+
     return status::success;
 }
 
