@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2022 Intel Corporation
+* Copyright 2022-2023 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -52,7 +52,8 @@ public:
                 = min(rem_bmnk_dim_, prb_dims_[prb_dim_idx_]->max_dim(level_));
         int dim = compute_next_block(level_, prb_level_dim(), target_dim,
                 bmnk_dim_.base_iter_block(), rem_prb_dim_,
-                prb_dims_[prb_dim_idx_]->base_iter_block(), is_last_prb_dim());
+                prb_dims_[prb_dim_idx_]->base_iter_block(), is_last_prb_dim(),
+                prb_dims_[prb_dim_idx_]->pad_block());
         if (level_ == tile_level_t::iter) {
             ir_assert(dim % prb_dims_[prb_dim_idx_]->base_iter_block() == 0);
         }
@@ -70,8 +71,14 @@ public:
         // Move to the next BMNK tile or next problem dimension (depending on
         // split/fuse settings and remaining sizes).
         if (rem_bmnk_dim_ != 1 && rem_prb_dim_ != 1) {
-            move_to_next_prb_dim();
-            return;
+            if (allow_fuse()) {
+                rem_prb_dim_ = 1;
+            } else if (prb_dims_[prb_dim_idx_]->allow_split()) {
+                rem_bmnk_dim_ = 1;
+            } else {
+                rem_prb_dim_ = 1;
+                rem_bmnk_dim_ = 1;
+            }
         }
 
         if (rem_bmnk_dim_ == 1 && rem_prb_dim_ == 1) {
@@ -150,14 +157,16 @@ private:
 
     int compute_next_block(tile_level_t level, int level_dim,
             dim_value_t target_dim, int target_base_blk, int dim,
-            int base_iter_block, bool is_last_dim,
+            int base_iter_block, bool is_last_dim, int pad_block,
             double target_eff = 0.75) const {
         if (target_dim.is_unlimited()) return dim;
 
         bool require_pow_2 = false;
         if (level == tile_level_t::tg) require_pow_2 = true;
         if (level == tile_level_t::iter
-                && utils::one_of(bmnk_dim_.bmnk(), 'B', 'M', 'N'))
+                && (bmnk_dim_.bmnk() == 'N'
+                        || (bmnk_dim_.bmnk() == 'M'
+                                && bmnk_dim_.inner_dims() == 1)))
             require_pow_2 = true;
 
         int step = 1;
@@ -189,6 +198,8 @@ private:
         if (ret == 0) ret = step;
         if (require_pow_2) ir_assert(math::is_pow2(ret));
         if (level == tile_level_t::iter) ir_assert(ret % base_iter_block == 0);
+        if (pad_block > ret && pad_block % ret != 0)
+            ret = math::gcd(pad_block, ret);
         return ret;
     }
 
@@ -358,8 +369,14 @@ void block_helper_t::init_bmnk_blocks() {
     int k_inst_blk = 0;
     int bn_inst_blk = 0;
     bool is_ge_hpc = (hw_cfg_.hw() >= ngen::HW::XeHPC);
-    bool reduce_m_block = (m_dim().base_iter_block() == 1
-            || k_dim().base_iter_block() == 1);
+    bool reduce_m_block = false;
+    if (reduce_m_block_hint_set_) {
+        reduce_m_block = reduce_m_block_hint_;
+    } else {
+        if (m_dim().base_iter_block() == 1) reduce_m_block = true;
+        if (k_dim().base_iter_block() == 1) reduce_m_block = true;
+    }
+    if (is_tf32() && fma_kind_ != fma_kind_t::mad) reduce_m_block = true;
     int eu_thr_mul = (!is_ge_hpc && reduce_m_block) ? 2 : 4;
 #ifdef GEN_CONV_DEBUG
     eu_thr_mul = getenv_int("eu_thr_mul", eu_thr_mul);
@@ -371,7 +388,14 @@ void block_helper_t::init_bmnk_blocks() {
             m_inst_blk = std::min(8, utils::rnd_down_pow2(max_m_iter_dim));
             bn_inst_blk = hw_cfg_.vec_size();
             k_inst_blk = 1;
-            m_blk = (is_x8x8s32() ? 8 : 16);
+            bool use_small_m_block = hw_cfg_.hw() <= ngen::HW::XeHP
+                    && m_dim().base_iter_block() == 1;
+            m_blk = (is_x8x8s32() || use_small_m_block ? 8 : 16);
+            bool small_m_tg = m_dim().base_iter_block() == 1
+                    && hw_cfg_.hw() == ngen::HW::XeHPG
+                    && !m_dim().pref_tg_block();
+            if (!m_dim().pref_tg_block())
+                m_dim().set_max_dim(tile_level_t::tg, small_m_tg ? 1 : 4);
             bn_blk = hw_cfg_.vec_size();
             k_blk = compute_mad_k_block();
             if (!allow_k_grid_slicing_ && !allow_k_tg_slicing_) {
@@ -389,16 +413,46 @@ void block_helper_t::init_bmnk_blocks() {
         }
         case fma_kind_t::dp4a:
         case fma_kind_t::dpas:
-        case fma_kind_t::dpasw:
+        case fma_kind_t::dpasw: {
             ir_assert(vectorize_by_n())
                     << "dpas can support N vectorization only.";
-            m_inst_blk = 8;
+            int max_iter_dim = prb_max_dim('M', tile_level_t::iter);
+            int target_m_blk = reduce_m_block ? 16 : 32;
+            if (max_iter_dim % target_m_blk != 0 && max_iter_dim > 32) {
+                float max_utilization_rate = 0.;
+                for (int i
+                        = min(32, utils::rnd_dn((int)(1.5 * target_m_blk), 4));
+                        i > target_m_blk; i -= 4) {
+                    float utilization_rate = (float)max_iter_dim
+                            / utils::rnd_up(max_iter_dim, i);
+                    // Heuristic constant preferring larger blocks, experimentally determined.
+                    const float threshhold = 1.05;
+                    if (utilization_rate > threshhold * max_utilization_rate) {
+                        max_utilization_rate = utilization_rate;
+                        target_m_blk = i;
+                    }
+                }
+            }
+            m_blk = target_m_blk;
+            m_inst_blk = m_blk % 8 == 0 ? 8 : m_blk;
             bn_inst_blk = 8;
             k_inst_blk = is_x8x8s32() ? 32 : 16;
-            m_blk = (reduce_m_block ? 16 : 32);
             bn_blk = is_ge_hpc ? 64 : 32;
+            int est_bmn_threads = 1;
+            est_bmn_threads *= utils::div_up(m_dim().size(), m_blk);
+            est_bmn_threads *= utils::div_up(bn_dim.size(), bn_blk);
+            auto thread_factor
+                    = utils::div_up(hw_cfg_.eu_count(), est_bmn_threads);
+            if (thread_factor > (is_ge_hpc ? 6 : 2) && !allow_k_grid_slicing_) {
+                if (m_inst_blk % 2 == 0 && bn_blk != bn_dim.size()) {
+                    m_blk /= 2;
+                    if (m_blk % m_inst_blk != 0) m_inst_blk /= 2;
+                }
+            }
+
             k_blk = k_inst_blk;
             break;
+        }
         default: ir_error_not_expected();
     }
 
@@ -428,12 +482,15 @@ void block_helper_t::init_bmnk_blocks() {
         if (is_bn && !vectorize_by_bmnk(bmnk)) continue;
 
         int blk = d.base_iter_block();
-        int inst_blk
-                = is_bn ? bn_inst_blk : (bmnk == 'M') ? m_inst_blk : k_inst_blk;
+        int inst_blk = is_bn    ? bn_inst_blk
+                : (bmnk == 'M') ? m_inst_blk
+                                : k_inst_blk;
         d.set_base_iter_block(math::lcm(blk, inst_blk));
     }
 
     m_blk = compute_block(m_dim().size(), m_blk, m_dim().base_iter_block());
+    // Require pow2 when only one m dim is non-trivial
+    if (m_dim().inner_dims() == 1) m_blk = utils::rnd_down_pow2(m_blk);
     k_blk = compute_block(k_dim().size(), k_blk, k_dim().base_iter_block());
     bn_blk = compute_block(bn_dim.size(), bn_blk, bn_dim.base_iter_block());
 
@@ -447,7 +504,6 @@ void block_helper_t::init_bmnk_blocks() {
         bn_blk = getenv_int("n_iter_blk", bn_blk);
     }
 #endif
-
     m_dim().set_iter_dim(m_blk);
     bn_dim.set_iter_dim(bn_blk);
     k_dim().set_iter_dim(k_blk);
@@ -483,7 +539,11 @@ void block_helper_t::init_bmnk_blocks() {
             int target_blk = i_max_tg_dim * d.iter_dim();
             int tg_dim = compute_block(d.size(), target_blk, d.iter_dim())
                     / d.iter_dim();
-            tg_dim = utils::rnd_down_pow2(tg_dim);
+            //restrict maximum single tg dim as max_tg size is reduced
+            tg_dim = std::min(utils::rnd_down_pow2(tg_dim),
+                    hw_cfg_.max_tg_overriden() ? target_tg_size
+                                    / (hw_cfg_.hw() >= ngen::HW::XeHPC ? 2 : 4)
+                                               : hw_cfg_.simd_size());
             tg_dims[i] = tg_dim;
             if (d.pref_tg_block()) {
                 //only one preferred dim allowed
@@ -504,7 +564,7 @@ void block_helper_t::init_bmnk_blocks() {
         };
 
         // Reduce thread group size until it fits the target size.
-        while (total_tg_dim() > target_tg_size || max_tg_dim() >= 16) {
+        while (total_tg_dim() > target_tg_size) {
             max_tg_dim() /= 2;
         }
 
@@ -528,7 +588,7 @@ void block_helper_t::init_k_blocking() {
         for (char bmnk : {'B', 'M', 'N', 'K'})
             est_threads *= bmnk_dim(bmnk).grid_dim();
         int def_k_thr_dim = utils::div_up(est_threads, 2 * hw_cfg_.eu_count());
-        def_k_thr_dim = std::min(50, def_k_thr_dim);
+        def_k_thr_dim = std::min(100, def_k_thr_dim);
         def_k_thr_dim = std::max(1, def_k_thr_dim);
         int k_thr_dim = def_k_thr_dim;
 #ifdef GEN_CONV_DEBUG
