@@ -84,6 +84,14 @@ public:
         return true;
     }
 
+    bool is_divisible(const tensor_t &other) const {
+        if (ndims() != other.ndims()) return false;
+        for (int i = 0; i < ndims(); i++) {
+            if (dims_[i] % other.dims_[i] != 0) return false;
+        }
+        return true;
+    }
+
     std::string str() const {
         using ir_utils::operator<<;
 
@@ -201,6 +209,12 @@ public:
         return ret;
     }
 
+    grid_info_t resize(const std::vector<int> &new_dims) const {
+        grid_info_t ret = *this;
+        ret.dims_ = new_dims;
+        return ret;
+    }
+
     grid_info_t slice(int dim_idx, int new_off, int new_dim,
             const expr_t &new_idx, expr_t &new_idx_value) const {
         ir_assert(dim_idx >= 0 && dim_idx < ndims());
@@ -549,12 +563,13 @@ public:
 
     const std::vector<block_t> &blocks() const { return blocks_; }
 
-    dim_t inner_block(int dim_idx) const {
+    dim_t inner_block(int dim_idx, bool skip_outer = true) const {
         dim_t block0 = -1;
         int nblocks = 0;
         for (auto &b : blocks_) {
             if (b.dim_idx == dim_idx) {
                 nblocks++;
+                if (!skip_outer) return b.block;
                 if (block0 == -1) block0 = b.block;
             }
         }
@@ -581,6 +596,22 @@ public:
     bool operator==(const layout_t &other) const { return is_equal(other); }
 
     bool operator!=(const layout_t &other) const { return !operator==(other); }
+    bool operator<=(const layout_t &other) const {
+        if (!type_.is_equal(other.type_)) return false;
+        const auto other_blocks = other.normalize().blocks();
+        const auto self_blocks = normalize().blocks();
+        if (self_blocks.size() > other_blocks.size()) return false;
+        if (self_blocks.size() == 0) return true;
+
+        int i = 0;
+        for (; i < (int)self_blocks.size() - 1; i++) {
+            if (!self_blocks[i].is_equal(other_blocks[i])) return false;
+        }
+        return (self_blocks[i].dim_idx == other_blocks[i].dim_idx
+                && self_blocks[i].stride == other_blocks[i].stride
+                && other_blocks[i].block % self_blocks[i].block == 0);
+    }
+    bool operator>=(const layout_t &other) const { return other <= *this; }
 
     bool is_equal(const layout_t &other, bool compare_offset = true) const {
         return normalize().is_strictly_equal(other.normalize(), compare_offset);
@@ -803,9 +834,28 @@ public:
     tensor_t split_into_max_tile(
             dim_t max_tile_elems, bool is_dense_tile) const;
 
-    tensor_t split(const grid_info_t &grid) const {
+    tensor_t split(const grid_info_t &grid_info) const {
+        tensor_t min_tile;
+        std::vector<int> cur_dims(grid_info.ndims(), 1);
+
+        for (int iter = 0; iter < grid_info.elems(); iter++) {
+            for (int i = 0; i < grid_info.ndims(); i++) {
+                if (++cur_dims[i] <= grid_info.dim(i)) break;
+                cur_dims[i] = 1;
+            }
+            auto sub_grid = grid_info.resize(cur_dims);
+            auto tile = split_exact(sub_grid);
+            if (tile.is_empty()) continue;
+            if (min_tile.is_empty() || tile.elems() < min_tile.elems()) {
+                min_tile = tile;
+            }
+        }
+        return min_tile;
+    }
+
+    tensor_t split_exact(const grid_info_t &grid) const {
         std::vector<dim_t> tile_dims(ndims(), 1);
-        ir_assert(elems() % grid.elems() == 0) << "Can't split across grid.";
+        if (elems() % grid.elems() != 0) return tensor_t();
 
         dim_t cur_elems_per_tile = 1;
         dim_t elems_per_tile = elems() / grid.elems();
@@ -815,8 +865,7 @@ public:
             tile_dims[b.dim_idx] *= block;
             cur_elems_per_tile *= block;
         }
-        ir_assert(cur_elems_per_tile == elems_per_tile)
-                << "Can't split across grid.";
+        if (cur_elems_per_tile != elems_per_tile) return tensor_t();
 
         return split(tensor_t(tile_dims), grid);
     }
@@ -855,7 +904,7 @@ public:
                     auto tmp_layout = split_block(eb, e, b.block / e);
                     return tmp_layout.split(tile, grid, outer_blocks);
                 } else {
-                    ir_error_not_expected() << "Can't split across grid.";
+                    return tensor_t();
                 }
             } else {
                 dim_t next_chunk
@@ -864,12 +913,12 @@ public:
                     auto idx = grid_splitter.pop_block(next_chunk);
                     start[b.dim_idx] += idx * dims[b.dim_idx];
                     if (outer_blocks) outer_blocks->push_back(b);
-                } else if (b.block % next_chunk == 0) {
+                } else if (b.block % next_chunk == 0 && next_chunk != 1) {
                     auto tmp_layout
                             = split_block(eb, next_chunk, b.block / next_chunk);
                     return tmp_layout.split(tile, grid, outer_blocks);
                 } else {
-                    ir_error_not_expected() << "Can't split across grid.";
+                    return tensor_t();
                 }
             }
             dims[b.dim_idx] *= b.block;
@@ -988,6 +1037,60 @@ public:
         return new_blocks;
     }
 
+    // Reinterprets layouts to wider data type (up to 4 bytes).
+    // Example: 16a16b (s8 type) -> 16a4b (s32 type)
+    static bool try_reinterpret_to_wider_type(layout_t &src, layout_t &dst,
+            const tensor_t &tile = {}, bool do_update = true,
+            int *new_size_out = nullptr) {
+        if (src.blocks().empty() || dst.blocks().empty()) return false;
+        if (src.type() != dst.type()) return false;
+
+        auto &s0 = src.blocks()[0];
+        auto &d0 = dst.blocks()[0];
+        if (s0.dim_idx != d0.dim_idx) return false;
+        if (int(s0.stride) != 1) return false;
+        if (int(d0.stride) != 1) return false;
+
+        int old_size = src.type().size();
+        int s0_old_size = int(s0.block) * old_size;
+        int d0_old_size = int(d0.block) * old_size;
+
+        int new_size = math::gcd(s0_old_size, d0_old_size);
+        new_size = math::gcd(new_size, 4); // Try types up to 4 bytes.
+        if (new_size <= old_size) return false;
+
+        auto tile_ok = [&](const layout_t &l) {
+            if (tile.is_empty()) return true;
+            int factor = new_size / old_size;
+            if (tile(l.blocks()[0].dim_idx) % factor != 0) return false;
+            return true;
+        };
+
+        auto strides_ok = [&](const layout_t &l) {
+            for (int i = 1; i < int(l.blocks().size()); i++) {
+                auto &b = l.blocks()[i];
+                if (int(b.stride) * old_size % new_size != 0) return false;
+            }
+            return true;
+        };
+
+        while (new_size > old_size) {
+            bool ok = true;
+            ok &= (tile_ok(src) && tile_ok(dst));
+            ok &= (strides_ok(src) && strides_ok(dst));
+            if (ok) {
+                if (do_update) {
+                    src = src.reinterpret(type_t::s(new_size * 8));
+                    dst = dst.reinterpret(type_t::s(new_size * 8));
+                }
+                if (new_size_out) *new_size_out = new_size;
+                return true;
+            }
+            new_size /= 2;
+        }
+        return false;
+    }
+
 private:
     // Returns vector of <dimension index, block size> pairs.
     static std::vector<std::pair<int, dim_t>> parse_format(
@@ -1012,6 +1115,77 @@ private:
     std::vector<block_t> blocks_;
 };
 
+// Helper class to incrementally increase a sub-layout of the given layout.
+// One step - adding the minimal factor of the next remaining block. Used
+// to find the minimal tile between two layouts that is innermost for both
+// layouts.
+class layout_iterator_t {
+public:
+    layout_iterator_t(const layout_t &l) : l_(l), block_idx_(-1), block_(1) {}
+
+    bool has_next() const {
+        dim_t b = block_;
+        int b_idx = block_idx_;
+        while (b == 1) {
+            b_idx++;
+            if (b_idx >= int(l_.blocks().size())) return false;
+            b = int(l_.blocks()[b_idx].block);
+        }
+        return true;
+    }
+
+    layout_iterator_t &operator++() {
+        ir_assert(has_next());
+        while (block_ == 1) {
+            block_idx_++;
+            block_ = int(l_.blocks()[block_idx_].block);
+        }
+        // Find smallest factor.
+        for (int factor = 2; factor <= int(block_); factor++) {
+            if (block_ % factor == 0) {
+                block_ /= factor;
+                return *this;
+            }
+        }
+
+        ir_error_not_expected();
+        return *this;
+    }
+
+    tensor_t tile() const {
+        std::vector<dim_t> dims(l_.ndims(), 1);
+        for (int i = 0; i <= block_idx_; i++) {
+            auto &b = l_.blocks()[i];
+            int b_block = b.block;
+            if (i == block_idx_) b_block /= block_;
+            dims[b.dim_idx] *= b_block;
+        }
+        return tensor_t(dims);
+    }
+
+    int nblocks() const { return block_idx_ + 1; }
+
+    layout_t outer_layout() const {
+        auto &blocks = l_.blocks();
+        std::vector<block_t> outer_blocks;
+        if (block_ > 1) {
+            auto &b = blocks[block_idx_];
+            outer_blocks.push_back(b);
+            outer_blocks[0].block = block_;
+            outer_blocks[0].stride = b.stride * (b.block / block_);
+        }
+        outer_blocks.insert(outer_blocks.end(),
+                blocks.begin() + (block_idx_ + 1), blocks.end());
+        return layout_t(l_.type(), l_.ndims(), l_.offset(), outer_blocks);
+    }
+
+private:
+    const layout_t &l_;
+
+    int block_idx_;
+    dim_t block_;
+};
+
 inline std::ostream &operator<<(std::ostream &out, const layout_t &layout) {
     out << layout.str();
     return out;
@@ -1222,7 +1396,7 @@ public:
     }
 
     static const expr_t &placeholder_var() {
-        static expr_t ph_var = var_t::make(type_t::s32(), "_ph");
+        static thread_local expr_t ph_var = var_t::make(type_t::s32(), "_ph");
         return ph_var;
     }
 
@@ -1275,6 +1449,8 @@ public:
 
     const std::vector<dim_t> &vdims() const { return vdims_; }
 
+    std::vector<expr_t> vstart() const { return vstart_; }
+
     expr_t vstart(int vidx) const { return vstart_[vidx]; }
 
     const layout_t &tlayout() const { return tlayout_; }
@@ -1482,7 +1658,15 @@ public:
 
     // FIXME: Offset of the returned layout is always 0.
     layout_t create_pseudo_vlayout() const {
-        return create_pseudo_vlayout(tlayout_);
+        return create_pseudo_vlayout(normalized_tlayout());
+    }
+
+    layout_t normalized_tlayout() const {
+        auto blocks = move_size_1_blocks_outer();
+        blocks = layout_t::normalize_blocks(tlayout_.ndims(), blocks, false);
+        auto layout
+                = layout_t(type(), tlayout_.ndims(), offset(), blocks, false);
+        return layout;
     }
 
     layout_t create_dense_vlayout() const {
@@ -1532,16 +1716,17 @@ public:
 
     view_t substitute(const expr_t &from, const expr_t &to) const;
 
-    mask_tensor_t create_mask_tensor(const constraint_set_t &cset) const {
+    mask_tensor_t create_mask_tensor(
+            const constraint_set_t &cset, uint32_t tmask = 0xFFFFFFFF) const {
         auto _vlayout = create_dense_vlayout();
         mask_tensor_t mask_tensor(_vlayout);
         std::vector<dim_t> vargs(nvdims());
-        create_mask_tensor(mask_tensor, _vlayout, 0, vargs);
+        create_mask_tensor(mask_tensor, _vlayout, 0, vargs, tmask);
         mask_tensor.simplify(cset);
         return mask_tensor;
     }
 
-    bool try_create_buffer_view(view_t &buf_view, view_t &inv_view) const {
+    void try_create_buffer_view(view_t &buf_view, view_t &inv_view) const {
         buf_view = view_t(create_vvars(ntdims()), ntdims());
         inv_view = view_t(vvars(), ntdims());
         for (int i = 0; i < nvdims(); i++) {
@@ -1575,7 +1760,11 @@ public:
             }
             buf_vdim++;
 
-            if (!ok) return false;
+            if (!ok) {
+                buf_view = view_t();
+                inv_view = view_t();
+                return;
+            }
 
             auto buf_vstart = tdim.expr();
             auto inv_vstart = tdim.expr();
@@ -1589,7 +1778,11 @@ public:
             buf_vstart = simplify(buf_vstart);
             inv_vstart = simplify(inv_vstart);
 
-            if (!is_const(inv_vstart)) return false;
+            if (!is_const(inv_vstart)) {
+                buf_view = view_t();
+                inv_view = view_t();
+                return;
+            }
 
             buf_view.set_vdim(buf_vvar, buf_vdim, buf_vstart);
 
@@ -1597,14 +1790,17 @@ public:
             // in the buffered view.
             auto &tmask = tdim.mask();
             for (auto &vvar : vvars()) {
-                if (contains_object(tmask, vvar)) { return false; }
+                if (contains_object(tmask, vvar)) {
+                    buf_view = view_t();
+                    inv_view = view_t();
+                    return;
+                }
             }
 
             buf_view.set_tdim(i, buf_vvar, tmask);
             inv_view.set_tdim(i, tdim.expr() - inv_vstart);
         }
         buf_view.set_tlayout(tlayout_);
-        return true;
     }
 
     static const expr_t &placeholder_var() {
@@ -1613,15 +1809,16 @@ public:
 
     static std::vector<expr_t> create_vvars(int nvdims);
 
-private:
     template <typename SrcT = expr_t, typename DstT = SrcT>
-    std::vector<DstT> cvt_vargs_to_targs(
-            const std::vector<SrcT> &_vargs = {}) const {
+    std::vector<DstT> cvt_vargs_to_targs(const std::vector<SrcT> &_vargs = {},
+            bool ignore_vstart = false) const {
         std::vector<expr_t> vargs = expr_cast<expr_t>(_vargs);
         if (vargs.empty()) vargs.resize(nvdims(), 0);
 
-        for (int i = 0; i < nvdims(); i++) {
-            if (!is_zero(vstart_[i])) vargs[i] += vstart_[i];
+        if (!ignore_vstart) {
+            for (int i = 0; i < nvdims(); i++) {
+                if (!is_zero(vstart_[i])) vargs[i] += vstart_[i];
+            }
         }
 
         std::vector<expr_t> targs(ntdims());
@@ -1637,11 +1834,12 @@ private:
         return expr_cast<DstT>(targs);
     }
 
+private:
     layout_t create_pseudo_vlayout(const layout_t &tlayout) const;
 
     void create_mask_tensor(mask_tensor_t &mask_tensor,
-            const layout_t &_vlayout, int vidx,
-            std::vector<dim_t> &vargs) const {
+            const layout_t &_vlayout, int vidx, std::vector<dim_t> &vargs,
+            uint32_t tmask) const {
         if (vidx == _vlayout.ndims()) {
             bool is_init = false;
             std::vector<expr_t> vvalues;
@@ -1649,6 +1847,7 @@ private:
             expr_t mask = bool_imm_t::make(true);
             for (int i = 0; i < ntdims(); i++) {
                 auto &tdim = tdims_[i];
+                if ((tmask & (1 << i)) == 0) continue;
                 if (tdim.mask().is_empty()) continue;
                 if (!is_init) {
                     // Lazily initialize values
@@ -1666,10 +1865,30 @@ private:
 
         for (int i = 0; i < vdims()[vidx]; i++) {
             vargs[vidx] = i;
-            create_mask_tensor(mask_tensor, _vlayout, vidx + 1, vargs);
+            create_mask_tensor(mask_tensor, _vlayout, vidx + 1, vargs, tmask);
         }
     }
 
+    std::vector<block_t> move_size_1_blocks_outer() const {
+        std::vector<block_t> new_blocks;
+        std::vector<block_t> size_1_blocks;
+        for (auto &b : tlayout_.blocks()) {
+            if (b.block == 1 && vdims_[b.dim_idx] == 1) {
+                size_1_blocks.emplace_back(b);
+            } else {
+                new_blocks.emplace_back(b);
+            }
+        }
+        stride_t stride = new_blocks.empty()
+                ? stride_t(1)
+                : new_blocks.back().block * new_blocks.back().stride;
+        for (auto &b : size_1_blocks) {
+            b.stride = stride;
+            new_blocks.emplace_back(b);
+        }
+        return new_blocks;
+    }
+
     std::vector<expr_t> vvars_;
     std::vector<dim_t> vdims_;
     std::vector<expr_t> vstart_;
@@ -1734,16 +1953,17 @@ layout_t normalize_conv_layout(const layout_t &_layout, bool with_groups,
         bool add_groups, bool is_wei);
 
 void normalize_conv_layouts(layout_t &src_layout, layout_t &wei_layout,
-        layout_t &dst_layout, layout_t &bia_layout, bool with_groups,
-        int groups, bool is_dw, int reduced_dim, bool fuse_spatial,
+        layout_t &dst_layout, layout_t &bia_layout, bool with_groups, int g,
+        int ic, int oc, bool is_dw, int reduced_dim, bool fuse_spatial,
         bool add_groups);
 
 inline void normalize_conv_layouts(layout_t &src_layout, layout_t &wei_layout,
-        layout_t &dst_layout, bool with_groups, int groups, bool is_dw,
-        int reduced_dim, bool fuse_spatial, bool add_groups) {
+        layout_t &dst_layout, bool with_groups, int g, int ic, int oc,
+        bool is_dw, int reduced_dim, bool fuse_spatial, bool add_groups) {
     layout_t bia_layout;
     normalize_conv_layouts(src_layout, wei_layout, dst_layout, bia_layout,
-            with_groups, groups, is_dw, reduced_dim, fuse_spatial, add_groups);
+            with_groups, g, ic, oc, is_dw, reduced_dim, fuse_spatial,
+            add_groups);
 }
 
 } // namespace jit
