@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2019-2022 Intel Corporation
+* Copyright 2019-2023 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -163,12 +163,12 @@ const nocopy_table_t xe_hp_x8x8s32_nocopy_table[] = {
         {{{384, 384}, {384, 384}}, {{384, 512}, {384, 256}}}};
 
 const nocopy_table_t xe_hpc_f16_nocopy_table[] = {
-        // NN    NT   TN   TT
-        {{{0, 1024}, {2048, 0}}, {{0, 0}, {0, 0}}}};
+        // NN    NT      TN    TT
+        {{{8192, 8192}, {8192, 8192}}, {{0, 0}, {0, 0}}}};
 
 const nocopy_table_t xe_hpc_bf16_nocopy_table[] = {
-        // NN    NT   TN   TT
-        {{{0, 1024}, {2048, 0}}, {{0, 0}, {0, 0}}}};
+        // NN    NT      TN    TT
+        {{{8192, 8192}, {8192, 8192}}, {{0, 0}, {0, 0}}}};
 
 const nocopy_table_t xe_hpc_x8x8s32_nocopy_table[] = {
         // NN    NT      TN    TT
@@ -219,11 +219,11 @@ bool xe_hp_systolic_gemm_t::pd_t::use_nocopy() {
 
         if (!packed_a_) {
             auto lda_bytes = d->lda() * types::data_type_size(d->a_type());
-            bad_ld |= ((lda_bytes & 0xF) != 0);
+            bad_ld |= ((lda_bytes & 0x3) != 0);
         }
         if (!packed_b_) {
             auto ldb_bytes = d->ldb() * types::data_type_size(d->b_type());
-            bad_ld |= ((ldb_bytes & 0xF) != 0);
+            bad_ld |= ((ldb_bytes & 0x3) != 0);
         }
 
         auto table = all_tables[arch_idx][int(bad_ld)][type_idx];
@@ -240,15 +240,19 @@ bool xe_hp_systolic_gemm_t::pd_t::use_nocopy() {
 
 bool xe_hp_systolic_gemm_t::pd_t::set_default_formats(data_type_t dt) {
     using namespace format_tag;
-    using new_kernel_t = gen_gemm_xe_systolic_kernel_t;
+    using new_kd_t = gen_gemm_xe_systolic_kernel_desc_t;
 
     auto sz = types::data_type_size(dt);
     const auto &d = desc();
     auto arch = dev_info_->gpu_arch();
 
-    memory_desc_wrapper a_mdw(&desc_.b_desc);
-    memory_desc_wrapper b_mdw(&desc_.a_desc);
-    memory_desc_wrapper c_mdw(&desc_.c_desc);
+    auto &a_desc = desc_.b_desc;
+    auto &b_desc = desc_.a_desc;
+    auto &c_desc = desc_.c_desc;
+
+    memory_desc_wrapper a_mdw(&a_desc);
+    memory_desc_wrapper b_mdw(&b_desc);
+    memory_desc_wrapper c_mdw(&c_desc);
 
     bool a_any = a_mdw.format_any();
     bool b_any = b_mdw.format_any();
@@ -296,7 +300,7 @@ bool xe_hp_systolic_gemm_t::pd_t::set_default_formats(data_type_t dt) {
 
     unroll_m_ = 0;
     unroll_n_ = 0;
-    kernel_tag_ = 0;
+    alt_ = false;
     if (a_prepacked_16) unroll_m_ = 16;
     if (a_prepacked_32) unroll_m_ = 32;
     if (a_prepacked_64) unroll_m_ = 64;
@@ -307,78 +311,76 @@ bool xe_hp_systolic_gemm_t::pd_t::set_default_formats(data_type_t dt) {
     use_new_kernels_ = !with_ab_zero_points() && (d->k() >= 64);
     use_new_kernels_ |= (arch >= compute::gpu_arch_t::xe_hpc);
 
-    new_kernel_t::choose_unrolls(arch, dev_info_->eu_count(), d->a_type(),
+    new_kd_t::choose_unrolls(arch, dev_info_->eu_count(), d->a_type(),
             d->b_type(), d->c_type(), d->m(), d->n(), d->k(), d->batch(),
-            unroll_m_, unroll_n_, kernel_tag_);
+            unroll_m_, unroll_n_, alt_);
 
-    format_tag_t a_packed_tag = (unroll_m_ == 64)
-            ? a_packed_tag_64
-            : (unroll_m_ == 32) ? a_packed_tag_32 : a_packed_tag_16;
-    format_tag_t b_packed_tag = (unroll_n_ == 48)
-            ? b_packed_tag_48
-            : (unroll_n_ == 32) ? b_packed_tag_32 : b_packed_tag_16;
+    format_tag_t a_packed_tag = (unroll_m_ == 64) ? a_packed_tag_64
+            : (unroll_m_ == 32)                   ? a_packed_tag_32
+                                                  : a_packed_tag_16;
+    format_tag_t b_packed_tag = (unroll_n_ == 48) ? b_packed_tag_48
+            : (unroll_n_ == 32)                   ? b_packed_tag_32
+                                                  : b_packed_tag_16;
     format_tag_t c_packed_tag = b_packed_tag;
 
     packed_a_ = packed_b_ = packed_c_ = false;
 
     if (a_any) {
         if (b_zp_) {
-            CHECK(memory_desc_init_by_tag(desc_.b_desc, unpacked_tag));
+            CHECK(memory_desc_init_by_tag(a_desc, unpacked_tag));
         } else {
-            CHECK(memory_desc_init_by_tag(desc_.b_desc, a_packed_tag));
-            auto ld = desc_.b_desc.padded_dims[batch ? 1 : 0];
+            CHECK(memory_desc_init_by_tag(a_desc, a_packed_tag));
+            auto ld = a_desc.padded_dims[batch ? 1 : 0];
             ld = nice_ld(ld, int(sz));
-            auto &ostride
-                    = desc_.b_desc.format_desc.blocking.strides[batch ? 2 : 1];
+            auto &ostride = a_desc.format_desc.blocking.strides[batch ? 2 : 1];
             if (batch) {
-                auto &bstride = desc_.b_desc.format_desc.blocking.strides[0];
+                auto &bstride = a_desc.format_desc.blocking.strides[0];
                 bstride = (bstride / ostride) * unroll_m_ * ld;
             }
             ostride = unroll_m_ * ld;
             packed_a_ = true;
         }
-    } else if (a_mdw.matches_one_of_tag(a_packed_tag, ab, ba, abc, acb)
-            == undef)
+    } else if (!a_mdw.matches_tag(a_packed_tag)
+            && !is_md_gemm_compatible_plain_format(&a_desc))
         return false;
 
     if (b_any) {
         if (a_zp_) {
-            CHECK(memory_desc_init_by_tag(desc_.a_desc, unpacked_tag));
+            CHECK(memory_desc_init_by_tag(b_desc, unpacked_tag));
         } else {
-            CHECK(memory_desc_init_by_tag(desc_.a_desc, b_packed_tag));
+            CHECK(memory_desc_init_by_tag(b_desc, b_packed_tag));
             if (unroll_n_ > 16) { // Bug in zero-padding when unroll_n_ == 16
-                auto ld = desc_.a_desc.padded_dims[batch ? 2 : 1];
+                auto ld = b_desc.padded_dims[batch ? 2 : 1];
                 ld = nice_ld(ld, int(sz));
-                auto &ostride = desc_.a_desc.format_desc.blocking
-                                        .strides[batch ? 1 : 0];
+                auto &ostride
+                        = b_desc.format_desc.blocking.strides[batch ? 1 : 0];
                 if (batch) {
-                    auto &bstride
-                            = desc_.a_desc.format_desc.blocking.strides[0];
+                    auto &bstride = b_desc.format_desc.blocking.strides[0];
                     bstride = (bstride / ostride) * unroll_n_ * ld;
                 }
                 ostride = unroll_n_ * ld;
             }
             packed_b_ = true;
         }
-    } else if (b_mdw.matches_one_of_tag(b_packed_tag, ab, ba, abc, acb)
-            == undef)
+    } else if (!b_mdw.matches_tag(b_packed_tag)
+            && !is_md_gemm_compatible_plain_format(&b_desc))
         return false;
 
     if (c_any) {
-        CHECK(memory_desc_init_by_tag(desc_.c_desc, c_packed_tag));
+        CHECK(memory_desc_init_by_tag(c_desc, c_packed_tag));
         if (unroll_n_ > 16) { // Bug in zero-padding when unroll_n_ == 16
-            auto ld = desc_.c_desc.padded_dims[batch ? 2 : 1];
+            auto ld = c_desc.padded_dims[batch ? 2 : 1];
             ld = nice_ld(ld, int(sz));
-            auto &ostride
-                    = desc_.c_desc.format_desc.blocking.strides[batch ? 1 : 0];
+            auto &ostride = c_desc.format_desc.blocking.strides[batch ? 1 : 0];
             if (batch) {
-                auto &bstride = desc_.c_desc.format_desc.blocking.strides[0];
+                auto &bstride = c_desc.format_desc.blocking.strides[0];
                 bstride = (bstride / ostride) * unroll_n_ * ld;
             }
             ostride = unroll_n_ * ld;
         }
         packed_c_ = true;
-    } else if (c_mdw.matches_one_of_tag(c_packed_tag, ab, abc) == undef)
+    } else if (!c_mdw.matches_tag(c_packed_tag)
+            && !is_md_gemm_compatible_plain_format(&c_desc, true))
         return false;
 
     packed_a_ = packed_a_ || a_mdw.matches_tag(a_packed_tag);
@@ -417,13 +419,6 @@ status_t xe_hp_systolic_gemm_t::init(engine_t *engine) {
         default: co_kind_ = 'N'; break;
     }
 
-    if (get_verbose() >= 2) {
-        char tag_s[2] = {pd()->kernel_tag(), 0};
-        printf("onednn_verbose,info,gpu,gemm,kernel:%dx%d,%s,new:%c\n",
-                pd()->unroll_m(), pd()->unroll_n(), tag_s,
-                pd()->use_new_kernels() ? 'Y' : 'N');
-    }
-
     // Initialize compute kernels (assembly)
     {
         auto status = pd()->use_new_kernels() ? init_compute_new(engine)
@@ -453,6 +448,13 @@ status_t xe_hp_systolic_gemm_t::init(engine_t *engine) {
         }
     }
 
+    if (get_verbose() >= 2) {
+        printf("onednn_verbose,info,gpu,gemm,kernel:%dx%d,%dx%dx%d,new:%c\n",
+                pd()->unroll_m(), pd()->unroll_n(), compute_info_.wg[LoopM],
+                compute_info_.wg[LoopN], compute_info_.wg[LoopK],
+                pd()->use_new_kernels() ? 'Y' : 'N');
+    }
+
     return status::success;
 }
 
@@ -484,6 +486,9 @@ status_t xe_hp_systolic_gemm_t::init_compute_old(engine_t *engine) {
     cfg.tile_n = pd()->unroll_n();
     cfg.global_3x_buf = (cfg.tile_n == 32);
 
+    auto compute_engine = utils::downcast<compute::compute_engine_t *>(engine);
+    cfg.eu_count = compute_engine->device_info()->eu_count();
+
     if (pd()->with_c_zero_points())
         cfg.co_type = cfg.c_type;
     else if (pd()->with_bias()) {
@@ -574,12 +579,15 @@ status_t xe_hp_systolic_gemm_t::init_compute_old(engine_t *engine) {
 }
 
 status_t xe_hp_systolic_gemm_t::init_compute_new(engine_t *engine) {
-    using kernel_t = gen_gemm_xe_systolic_kernel_t;
-    using offset_t = kernel_t::offset_t;
+    using kernel_t = gen_gemm_kernel_t;
+    using kd_t = gen_gemm_xe_systolic_kernel_desc_t;
+    using offset_t = kd_t::offset_t;
 
-    auto a_type = pd()->desc()->a_type();
-    auto b_type = pd()->desc()->b_type();
-    auto c_type = pd()->desc()->c_type();
+    const auto d = pd()->desc();
+
+    auto a_type = d->a_type();
+    auto b_type = d->b_type();
+    auto c_type = d->c_type();
     auto co_type = pd()->impl_co_type();
     auto acc_type = pd()->impl_acc_type();
 
@@ -590,7 +598,7 @@ status_t xe_hp_systolic_gemm_t::init_compute_new(engine_t *engine) {
     offset_t bias_offset
             = pd()->with_bias() ? offset_t::runtime : offset_t::none;
 
-    bool may_k_block = (pd()->desc()->k() > kernel_t::min_block_k(a_type));
+    bool may_k_block = (d->k() > kd_t::min_block_k(a_type));
     bool got_info = false;
 
     bool with_eltwise
@@ -618,21 +626,24 @@ status_t xe_hp_systolic_gemm_t::init_compute_new(engine_t *engine) {
                     this_post_ops = &no_post_ops;
                 }
 
-                kernel_t kernel;
+                kd_t kd;
 
-                auto status = kernel.init(arch_, pd()->with_batch(), ab_offset,
+                auto status = kd.select_kernel(arch_, eu_count_,
+                        pd()->with_batch(), pd()->packed_c(), ab_offset,
                         ab_offset, this_c_offset, bias_offset, pd()->alpha(),
                         this_beta, *this_post_ops, a_type, b_type, c_type,
-                        co_type, acc_type, pd()->packed_c(), pd()->unroll_m(),
-                        pd()->unroll_n(), pd()->kernel_tag());
+                        co_type, acc_type, d->m(), d->n(), d->k(), d->batch(),
+                        pd()->unroll_m(), pd()->unroll_n(), pd()->alt());
 
                 if (status != status::success) return status;
 
                 if (!got_info) {
-                    compute_info_ = kernel.driver_info();
+                    compute_info_ = *kd.driver_info();
                     got_info = true;
                 }
 
+                kernel_t kernel(kd);
+
                 create_kernel(
                         engine, &kernel_[first_k_block][last_k_block], &kernel);
 
@@ -732,10 +743,13 @@ xe_hp_systolic_gemm_t::get_blocking() const {
     // Decide on k blocking.
     int64_t block_k = compute_info_.blocking[LoopK];
     int64_t nblock_k = utils::div_up(k, block_k);
+    nblock_k = nstl::max<int64_t>(nblock_k, 1);
     block_k = utils::div_up(k, nblock_k);
+    block_k = nstl::max<dim_t>(block_k, 1);
     block_k = utils::rnd_up(
             (pd()->desc()->acc_type != pd()->desc()->c_type()) ? k : block_k,
             unroll_k);
+    block_k = nstl::max<dim_t>(block_k, 1);
 
     return std::make_tuple(block_m, block_n, block_k);
 }
@@ -1008,7 +1022,7 @@ status_t xe_hp_systolic_gemm_t::execute(const gemm_exec_ctx_t &ctx) const {
         if (status) return status;
     }
 
-    for (int64_t Bk = 0; Bk < k; Bk += block_k) {
+    for (int64_t Bk = 0; Bk < nstl::max<dim_t>(k, 1); Bk += block_k) {
         int64_t size_k = k - Bk;
         bool first_k_block = (Bk == 0);
         bool last_k_block = (size_k <= block_k);
