@@ -36,10 +36,11 @@ int jit_eltwise_injector_f32<hw>::min_scratch_regs() {
             case eltwise_exp:
             case eltwise_exp_use_dst_for_bwd: return 0;
             case eltwise_gelu_erf: return 4;
+            case eltwise_hardsigmoid: return 0;
             case eltwise_hardswish: return 1;
             case eltwise_log: return 0;
             case eltwise_logsigmoid: return 1;
-            case eltwise_mish: return 2;
+            case eltwise_mish: return 4;
             case eltwise_pow: return 1;
             case eltwise_relu:
             case eltwise_relu_use_dst_for_bwd: return (alpha_ == 0.f) ? 0 : 1;
@@ -50,7 +51,7 @@ int jit_eltwise_injector_f32<hw>::min_scratch_regs() {
             case eltwise_square: return 0;
             case eltwise_swish: return 1;
             case eltwise_tanh:
-            case eltwise_tanh_use_dst_for_bwd: return 1;
+            case eltwise_tanh_use_dst_for_bwd: return 2;
             case eltwise_round: return 0;
             case eltwise_linear: return 0;
             case eltwise_bounded_relu:
@@ -123,8 +124,8 @@ int jit_eltwise_injector_f32<hw>::max_batch_size() {
             case eltwise_logsigmoid:
             case eltwise_pow:
             case eltwise_soft_relu:
-            case eltwise_tanh:
             case eltwise_swish: return ss;
+            case eltwise_tanh:
             case eltwise_mish:
             case eltwise_gelu_erf: return ss / min_scratch_regs();
             case eltwise_gelu_tanh: return ss & ~1;
@@ -151,6 +152,7 @@ int jit_eltwise_injector_f32<hw>::phase_count(alg_kind_t alg) {
             case eltwise_exp:
             case eltwise_exp_use_dst_for_bwd: return 2;
             case eltwise_gelu_erf: return 25;
+            case eltwise_hardsigmoid: return 4;
             case eltwise_hardswish: return 3;
             case eltwise_log: return 2;
             case eltwise_logsigmoid:
@@ -164,7 +166,8 @@ int jit_eltwise_injector_f32<hw>::phase_count(alg_kind_t alg) {
             case eltwise_soft_relu: return 9;
             case eltwise_swish: return 5;
             case eltwise_tanh:
-            case eltwise_tanh_use_dst_for_bwd: return 7;
+            case eltwise_tanh_use_dst_for_bwd:
+                return (use_tanh_compat()) ? 9 : 6;
             case eltwise_linear: return 2;
             case eltwise_bounded_relu:
             case eltwise_clip:
@@ -253,17 +256,43 @@ void jit_eltwise_injector_f32<hw>::square_compute_fwd(
 
 template <gpu_gen_t hw>
 void jit_eltwise_injector_f32<hw>::tanh_compute_fwd(
-        int simd, const ngen::GRF &r, int phase, int off) {
-    const float log2e = 1.442695f; // log_2(e)
-    auto a = scratch_[off].f();
+        int simd, const ngen::GRF &r, int phase, int off, int batch) {
+    const float log2e = 1.44269502162933349609375f; // log_2(e)
+    auto one_half = scratch_[0].f(7);
+    auto a = scratch_[off + batch].f();
     switch (phase) {
-        case 0: h->mul(simd, a, abs(r), 2 * log2e); break;
+        case 0: h->mul(simd, a, abs(r), 2.f * log2e); break;
         case 1: h->exp(simd, a, a); break;
-        case 2: h->add(simd, a, a, 1.f); break;
+        case 2: h->mad(simd, a, one_half, a, one_half); break;
         case 3: h->inv(simd, a, a); break;
-        case 4: h->mul(simd, a, a, 2.f); break;
-        case 5: h->add(simd, a, -a, 1.f); break;
-        case 6: h->csel(simd | ge | f0[0], r, a, -a, r); break;
+        case 4: h->add(simd, a, -a, 1.f); break;
+        case 5: h->csel(simd | ge | f0[0], r, a, -a, r); break;
+        default: assert(!"invalid phase");
+    }
+}
+
+template <gpu_gen_t hw>
+void jit_eltwise_injector_f32<hw>::tanh_compute_fwd_compat(
+        int simd, const ngen::GRF &r, int phase, int off, int batch) {
+    // This approximation of tanh(x) does not use the math.exp instruction
+    // that seems to be faulty on DG2-128; the exact formula is as follows:
+    // R = max(min(0.0519867*x*((x^2 + k)^2 + l)/((x^2 + m)^2 + n), 1), -1)
+    // Both absolute and relative errors are <7*10^-5 \forall x \in \mathbb R
+    auto k = scratch_[0].f(4);
+    auto l = scratch_[0].f(5);
+    auto m = scratch_[0].f(6);
+    auto n = scratch_[0].f(7);
+    auto a = scratch_[off + batch].f();
+    switch (phase) {
+        case 0: h->mad(simd, a, m, r, r); break;
+        case 1: h->mad(simd, a, n, a, a); break;
+        case 2: h->inv(simd, a, a); break;
+        case 3: h->mul(simd, a, a, r); break;
+        case 4: h->mad(simd, r, k, r, r); break;
+        case 5: h->mad(simd, r, l, r, r); break;
+        case 6: h->mul(simd, r, r, 0.0519867f); break; // 0.051986694f
+        case 7: h->mul(simd | sat, r, r, abs(a)); break;
+        case 8: h->csel(simd | ge | f0[0], r, r, -r, a); break;
         default: assert(!"invalid phase");
     }
 }
@@ -383,6 +412,24 @@ void jit_eltwise_injector_f32<hw>::clip_prepare_bwd() {
     h->mov(1, pos_inf, pos_inf_imm);
 }
 
+template <gpu_gen_t hw>
+void jit_eltwise_injector_f32<hw>::tanh_prepare_fwd() {
+    auto one_half = scratch_[0].f(7);
+    h->mov(1, one_half, 0.5f);
+}
+
+template <gpu_gen_t hw>
+void jit_eltwise_injector_f32<hw>::tanh_prepare_fwd_compat() {
+    auto k = scratch_[0].f(4);
+    auto l = scratch_[0].f(5);
+    auto m = scratch_[0].f(6);
+    auto n = scratch_[0].f(7);
+    h->mov(1, k, 77.0954f); //  77.095392909578f
+    h->mov(1, l, -4435.55f); // -4435.54623970169f
+    h->mov(1, m, 17.06396f); //  17.06396485f
+    h->mov(1, n, -212.7724f); // -212.772646402036f
+}
+
 template <gpu_gen_t hw>
 void jit_eltwise_injector_f32<hw>::abs_compute_bwd(
         int simd, const ngen::GRF &r, int phase) {
@@ -529,6 +576,18 @@ void jit_eltwise_injector_f32<hw>::gelu_erf_compute_fwd(
     }
 }
 
+template <gpu_gen_t hw>
+void jit_eltwise_injector_f32<hw>::hardsigmoid_compute_fwd(
+        int simd, const ngen::GRF &r, int phase, int off) {
+    switch (phase) {
+        case 0: h->mul(simd, r, r, alpha_); break;
+        case 1: h->add(simd, r, r, beta_); break;
+        case 2: h->min_(simd, r, r, 1.f); break;
+        case 3: h->max_(simd, r, r, 0.f); break;
+        default: assert(!"invalid phase");
+    }
+}
+
 template <gpu_gen_t hw>
 void jit_eltwise_injector_f32<hw>::hardswish_compute_fwd(
         int simd, const ngen::GRF &r, int phase, int off) {
@@ -566,15 +625,20 @@ void jit_eltwise_injector_f32<hw>::logsigmoid_compute_fwd(
 template <gpu_gen_t hw>
 void jit_eltwise_injector_f32<hw>::mish_compute_fwd(
         int simd, const ngen::GRF &r, int phase, int off, int batch) {
-    auto temp = scratch_[off].f();
-    auto temp2 = scratch_[off + batch].f();
+    auto temp = scratch_[off + batch].f();
+    auto temp2 = scratch_[off + 2 * batch].f();
     const int srelu_phases = phase_count(alg_kind::eltwise_soft_relu);
     const int tanh_phases = phase_count(alg_kind::eltwise_tanh);
-    // note tanh_compute_fwd will trash temp
+    // note tanh_compute_fwd_* clobbers scratch_[off] and scratch_[off + batch]
     if (phase < srelu_phases)
         soft_relu_compute_fwd_inner(simd, r, temp, temp2, phase, off);
-    if (phase >= srelu_phases && phase < srelu_phases + tanh_phases)
-        tanh_compute_fwd(simd, temp2, phase - srelu_phases, off);
+    if (phase >= srelu_phases && phase < srelu_phases + tanh_phases) {
+        if (use_tanh_compat())
+            tanh_compute_fwd_compat(
+                    simd, temp2, phase - srelu_phases, off, batch);
+        else
+            tanh_compute_fwd(simd, temp2, phase - srelu_phases, off, batch);
+    }
     if (phase == srelu_phases + tanh_phases) h->mul(simd, r, r, temp2);
     if (phase > srelu_phases + tanh_phases) assert(!"invalid phase");
 }
@@ -632,6 +696,9 @@ void jit_eltwise_injector_f32<hw>::compute(const ngen::GRFRange &regs) {
                         case eltwise_gelu_erf:
                             gelu_erf_compute_fwd(simd, base, phase, ii, batch);
                             break;
+                        case eltwise_hardsigmoid:
+                            hardsigmoid_compute_fwd(simd, base, phase, ii);
+                            break;
                         case eltwise_hardswish:
                             hardswish_compute_fwd(simd, base, phase, ii);
                             break;
@@ -670,7 +737,11 @@ void jit_eltwise_injector_f32<hw>::compute(const ngen::GRFRange &regs) {
                             break;
                         case eltwise_tanh:
                         case eltwise_tanh_use_dst_for_bwd:
-                            tanh_compute_fwd(simd, base, phase, ii);
+                            if (use_tanh_compat())
+                                tanh_compute_fwd_compat(
+                                        simd, base, phase, ii, batch);
+                            else
+                                tanh_compute_fwd(simd, base, phase, ii, batch);
                             break;
                         case eltwise_round:
                             round_compute_fwd(simd, base);
@@ -735,7 +806,16 @@ void jit_eltwise_injector_f32<hw>::prepare() {
     assert(scratch_.getLen() >= min_scratch_regs());
 
     if (is_fwd_) {
-        /* nothing to do */
+        switch (alg_) {
+            case eltwise_mish:
+            case eltwise_tanh:
+                if (use_tanh_compat())
+                    tanh_prepare_fwd_compat();
+                else
+                    tanh_prepare_fwd();
+                break;
+            default: break;
+        }
     } else {
         switch (alg_) {
             case eltwise_relu: relu_prepare_bwd(); break;
