@@ -29,9 +29,10 @@
 
 #include "binary/binary.hpp"
 #include "bnorm/bnorm.hpp"
+#include "brgemm/brgemm.hpp"
 #include "concat/concat.hpp"
 #include "conv/conv.hpp"
-#include "conv/deconv.hpp"
+#include "deconv/deconv.hpp"
 #include "eltwise/eltwise.hpp"
 #include "ip/ip.hpp"
 #include "lnorm/lnorm.hpp"
@@ -55,7 +56,7 @@ bool mem_check {true};
 std::string skip_impl;
 bench_mode_t bench_mode {CORR};
 stat_t benchdnn_stat {0};
-const char *driver_name = "";
+std::string driver_name;
 
 double max_ms_per_prb {3e3};
 int min_times_per_prb {5};
@@ -127,16 +128,18 @@ int main(int argc, char **argv) {
         reduction::bench(--argc, ++argv);
     } else if (!strcmp("--zeropad", argv[0])) {
         zeropad::bench(--argc, ++argv);
+    } else if (!strcmp("--brgemm", argv[0])) {
+        brgemm::bench(--argc, ++argv);
     } else {
         fprintf(stderr, "err: unknown driver\n");
     }
 
-    printf("tests:%d passed:%d "
-           "skipped:%d mistrusted:%d unimplemented:%d "
-           "failed:%d listed:%d\n",
+    printf("tests:%d passed:%d skipped:%d mistrusted:%d unimplemented:%d "
+           "invalid_arguments:%d failed:%d listed:%d\n",
             benchdnn_stat.tests, benchdnn_stat.passed, benchdnn_stat.skipped,
             benchdnn_stat.mistrusted, benchdnn_stat.unimplemented,
-            benchdnn_stat.failed, benchdnn_stat.listed);
+            benchdnn_stat.invalid_arguments, benchdnn_stat.failed,
+            benchdnn_stat.listed);
     if (is_bench_mode(PERF)) {
         printf("total perf: min(ms):%g avg(ms):%g\n",
                 benchdnn_stat.ms[timer::timer_t::min],
