@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2017-2022 Intel Corporation
+* Copyright 2017-2023 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -105,19 +105,26 @@ const char *prop2str(dnnl_prop_kind_t prop) {
 const char *data_kind2str(data_kind_t kind) {
     switch (kind) {
         case SRC: return "SRC";
+        case SRC_1: return "SRC_ADD";
         case WEI: return "WEI";
         case BIA: return "BIA";
         case DST: return "DST";
         case ACC: return "ACC";
-        case DATA: return "DATA";
         case MEAN: return "MEAN";
         case VAR: return "VAR";
         case SS: return "SS";
         case SC: return "SC";
         case SH: return "SH";
-        case GWEI: return "GWEI";
+        case DST_ITER: return "DST_ITER";
+        case DST_ITER_C: return "DST_ITER_C";
+        case AUGRU_ATTENTION: return "AUGRU_ATTENTION";
+        case SRC_ITER: return "SRC_ITER";
+        case SRC_ITER_C: return "SRC_ITER_C";
+        case WEI_ITER: return "WEI_ITER";
+        case WEI_PEEPHOLE: return "WEI_PEEPHOLE";
+        case WEI_PROJECTION: return "WEI_PROJECTION";
+        default: assert(!"incorrect data kind");
     }
-    assert(!"incorrect data kind");
     return "incorrect data kind";
 }
 
@@ -132,9 +139,27 @@ static int str2arg(const std::string &str) {
     for (const auto &arg : supported_args)
         for (const auto &s : arg.second)
             if (str.compare(s) == 0) return arg.first;
+    // multiple srcs
+    if (str.compare(0, 3, "msrc")) {
+        const auto &str_index = str.substr(4);
+        const auto index = stoul(str_index);
+        return DNNL_ARG_MULTIPLE_SRC + index;
+    }
     return BENCHDNN_DNNL_ARG_UNDEF;
 }
 
+static std::string arg2str(int arg) {
+    if (supported_args.find(arg) != supported_args.end())
+        return std::string(supported_args.at(arg)[0]);
+    if (arg & DNNL_ARG_MULTIPLE_SRC) {
+        std::string msrc("msrc");
+        const int index = arg - DNNL_ARG_MULTIPLE_SRC;
+        return msrc + std::to_string(index);
+    }
+    assert(!"unknown argument");
+    return "unknown argument";
+}
+
 policy_t attr_t::str2policy(const std::string &str) {
     std::string s(str);
     // s.compare is lexicographical, case matters
@@ -320,6 +345,8 @@ static po_table_entry_t kind_table[] = {
                 dnnl_eltwise_gelu_erf},
         {pk_t::GELU_TANH, {"gelu_tanh", "eltwise_gelu_tanh"},
                 dnnl_eltwise_gelu_tanh},
+        {pk_t::HARDSIGMOID, {"hardsigmoid", "eltwise_hardsigmoid"},
+                dnnl_eltwise_hardsigmoid},
         {pk_t::HARDSWISH, {"hardswish", "eltwise_hardswish"},
                 dnnl_eltwise_hardswish},
         {pk_t::LINEAR, {"linear", "eltwise_linear"}, dnnl_eltwise_linear},
@@ -343,6 +370,8 @@ static po_table_entry_t kind_table[] = {
         {pk_t::SQUARE, {"square", "eltwise_square"}, dnnl_eltwise_square},
         {pk_t::SRELU, {"soft_relu", "eltwise_soft_relu", "srelu"},
                 dnnl_eltwise_soft_relu},
+        {pk_t::SRELU_V2, {"soft_relu_v2", "eltwise_soft_relu_v2", "srelu_v2"},
+                dnnl_eltwise_soft_relu_v2},
         {pk_t::SWISH, {"swish", "eltwise_swish"}, dnnl_eltwise_swish},
         {pk_t::TANH, {"tanh", "eltwise_tanh"}, dnnl_eltwise_tanh},
         {pk_t::TANH_DST, {"tanh_dst", "eltwise_tanh_use_dst_for_bwd"},
@@ -422,19 +451,6 @@ std::vector<std::pair<int, int>> attr_t::post_ops_t::get_po_masks() const {
     return v_masks;
 }
 
-std::vector<int> attr_t::post_ops_t::get_binary_po_masks() const {
-    std::vector<int> v_masks;
-    for (int idx = 0; idx < len(); ++idx) {
-        const auto &e = this->entry[idx];
-        if (!e.is_binary_kind()) continue;
-
-        const auto policy = e.binary.policy;
-        const auto mask = attr_t::get_default_mask(policy);
-        v_masks.push_back(mask);
-    }
-    return v_masks;
-}
-
 int attr_t::post_ops_t::from_str(const std::string &s) {
     *this = post_ops_t();
     if (s.empty()) return OK;
@@ -537,10 +553,12 @@ int attr_t::post_ops_t::from_str(const std::string &s) {
     return OK;
 }
 
-bool attr_t::is_def() const {
+bool attr_t::is_def(bool skip_fpmath) const {
     return oscale.is_def() && scales.is_def() && zero_points.is_def()
             && post_ops.is_def()
-            && scratchpad_mode == dnnl_scratchpad_mode_library;
+            && scratchpad_mode == dnnl_scratchpad_mode_library
+            && IMPLICATION(
+                    !skip_fpmath, fpmath_mode == dnnl_fpmath_mode_strict);
 }
 
 int attr_t::post_ops_t::find(pk_t kind, int start, int stop) const {
@@ -595,70 +613,6 @@ int attr_t::post_ops_t::prelu_index() const {
     return -1;
 }
 
-int str2attr(attr_t *attr, const char *str) {
-    if (attr == nullptr || str == nullptr) return FAIL;
-
-    *attr = attr_t();
-    std::string s(str), entry_name;
-    size_t start_pos = 0, end_pos = start_pos;
-
-    entry_name = "oscale=";
-    start_pos = s.find(entry_name);
-    if (start_pos != std::string::npos) {
-        start_pos += entry_name.size();
-        // ';' is an attribute delimeter
-        auto status
-                = attr->oscale.from_str(parser::get_substr(s, start_pos, ';'));
-        if (status != OK) return status;
-    }
-
-    entry_name = "zero_points=";
-    start_pos = s.find(entry_name);
-    if (start_pos != std::string::npos) {
-        start_pos += entry_name.size();
-        auto status = attr->zero_points.from_str(
-                parser::get_substr(s, start_pos, ';'));
-        if (status != OK) return status;
-    }
-
-    entry_name = "scales=";
-    start_pos = s.find(entry_name);
-    if (start_pos != std::string::npos) {
-        start_pos += entry_name.size();
-        auto status
-                = attr->scales.from_str(parser::get_substr(s, start_pos, ';'));
-        if (status != OK) return status;
-    }
-
-    entry_name = "post_ops=";
-    start_pos = s.find(entry_name);
-    if (start_pos != std::string::npos) {
-        start_pos += entry_name.size();
-        // "\'\'" separates post-ops from everything else
-        end_pos = s.find_first_of('\'', s.find_first_of('\'', start_pos) + 1);
-        // +1 to include '\'' as an essential part of post-ops
-        auto parse_str = s.substr(start_pos, end_pos - start_pos + 1);
-        auto status = attr->post_ops.from_str(parse_str);
-        if (status != OK) return status;
-    }
-
-    return OK;
-}
-
-void handle_legacy_attr(attr_t &attr, const attr_t &legacy_attr) {
-    if (legacy_attr.is_def()) return;
-
-    if (!attr.is_def()) {
-        BENCHDNN_PRINT(0, "%s\n",
-                "ERROR: both `--attr=` and one of `--attr-post-ops=`, "
-                "`--attr-scales=`, `--attr-zero-points=` or `--attr-oscale=` "
-                "options are specified, please use latter options.");
-        SAFE_V(FAIL);
-    }
-
-    attr = legacy_attr;
-}
-
 std::ostream &operator<<(std::ostream &s, const policy_t &policy) {
     s << attr_t::policy2str(policy);
     return s;
@@ -675,8 +629,8 @@ std::ostream &operator<<(
     const char *delim = "";
     for (const auto &point : zero_points.points) {
         s << delim;
-        s << supported_args.at(point.first)[0] << ":" << point.second.policy
-          << ":" << point.second.value;
+        s << arg2str(point.first) << ":" << point.second.policy << ":"
+          << point.second.value;
         if (point.second.runtime) s << '*';
         delim = "+";
     }
@@ -689,7 +643,7 @@ std::ostream &operator<<(std::ostream &s, const attr_t::arg_scales_t &scales) {
     for (const auto &v : scales.scales) {
         if (!v.second.is_def()) {
             s << delim;
-            s << supported_args.at(v.first)[0] << ":" << v.second;
+            s << arg2str(v.first) << ":" << v.second;
             delim = "+";
         }
     }
@@ -757,6 +711,11 @@ std::ostream &operator<<(std::ostream &s, dnnl_scratchpad_mode_t sm) {
     return s;
 }
 
+std::ostream &operator<<(std::ostream &s, dnnl_fpmath_mode_t fm) {
+    s << fpmath_mode2str(fm);
+    return s;
+}
+
 std::ostream &operator<<(std::ostream &s, const attr_t &attr) {
     if (!attr.is_def()) {
         if (!attr.oscale.is_def()) s << "--attr-oscale=" << attr.oscale << " ";
@@ -767,6 +726,8 @@ std::ostream &operator<<(std::ostream &s, const attr_t &attr) {
             s << "--attr-post-ops=" << attr.post_ops << " ";
         if (attr.scratchpad_mode != dnnl_scratchpad_mode_library)
             s << "--attr-scratchpad=" << attr.scratchpad_mode << " ";
+        if (attr.fpmath_mode != dnnl_fpmath_mode_strict)
+            s << "--attr-fpmath=" << attr.fpmath_mode << " ";
     }
     return s;
 }
@@ -781,6 +742,17 @@ std::ostream &operator<<(std::ostream &s, bench_mode_t mode) {
     return s;
 }
 
+std::ostream &operator<<(std::ostream &s, memory_kind_ext_t memory_kind) {
+    switch (memory_kind) {
+        case memory_kind_ext_t::usm: s << "usm"; break;
+        case memory_kind_ext_t::buffer: s << "buffer"; break;
+        case memory_kind_ext_t::usm_device: s << "usm_device"; break;
+        case memory_kind_ext_t::usm_shared: s << "usm_shared"; break;
+        default: assert(!"unexpected"); break;
+    }
+    return s;
+}
+
 std::ostream &dump_global_params(std::ostream &s) {
     s << "--" << driver_name << " ";
     if (canonical) s << "--canonical=" << bool2str(canonical) << " ";
@@ -801,6 +773,10 @@ std::ostream &dump_global_params(std::ostream &s) {
     if (canonical || bench_mode != CORR) s << "--mode=" << bench_mode << " ";
     if (canonical || attr_same_pd_check != false)
         s << "--attr-same-pd-check=" << bool2str(attr_same_pd_check) << " ";
+#if defined(DNNL_WITH_SYCL) || DNNL_GPU_RUNTIME == DNNL_RUNTIME_OCL
+    if (canonical || memory_kind != default_memory_kind)
+        s << "--memory-kind=" << memory_kind << " ";
+#endif
 
     return s;
 }
@@ -829,6 +805,31 @@ dnnl_scratchpad_mode_t str2scratchpad_mode(const char *str) {
     return dnnl_scratchpad_mode_library;
 }
 
+dnnl_fpmath_mode_t str2fpmath_mode(const char *str) {
+    if (std::strcmp(str, "") == 0) {
+        dnnl_fpmath_mode_t ret;
+        dnnl_get_default_fpmath_mode(&ret);
+        return ret;
+    }
+
+#define CASE(fpm) \
+    param = #fpm; \
+    if (!strncasecmp(param, str, strlen(param))) return dnnl_fpmath_mode_##fpm;
+
+    const char *param;
+
+    CASE(strict);
+    CASE(bf16);
+    CASE(f16);
+    CASE(tf32);
+    CASE(any);
+
+    assert(!"not expected");
+    return dnnl_fpmath_mode_strict;
+
+#undef CASE
+}
+
 void attr_args_t::prepare_output_scales(
         const attr_t &attr, const void *vals, int64_t count, int mask) {
     insert(DNNL_ARG_ATTR_OUTPUT_SCALES, vals, count, mask, attr.oscale.runtime);
@@ -876,9 +877,8 @@ int attr_args_t::prepare_post_ops_mds(
                 rhs_tensor_dims[d] = (!(mask & (1 << d))) ? 1 : dims[d];
 
             dnnl_memory_desc_t rhs_tensor_desc;
-            SAFE(init_md(&rhs_tensor_desc, ndims, rhs_tensor_dims,
-                         po_rhs_tensor_entry.dt, po_rhs_tensor_entry.tag),
-                    WARN);
+            rhs_tensor_desc = dnn_mem_t::init_md(ndims, rhs_tensor_dims,
+                    po_rhs_tensor_entry.dt, po_rhs_tensor_entry.tag);
             mds.emplace((DNNL_ARG_ATTR_MULTIPLE_POST_OP(idx)
                                 | po_rhs_tensor_entry.arg_attr_mask),
                     rhs_tensor_desc);
@@ -919,12 +919,12 @@ dnnl_primitive_attr_t create_dnnl_attr(
         DNN_SAFE_V(dnnl_primitive_attr_set_output_scales(
                 dnnl_attr, count, mask, scales));
     } else if (!attr.scales.is_def()) {
-        for (const auto &arg : supported_args) {
-            const auto arg_name = arg.first;
-            const auto &as = attr.scales;
+        const auto &as = attr.scales;
+        for (const auto &arg : as.scales) {
+            const int arg_name = arg.first;
             if (as.is_def(arg_name)) continue;
 
-            const auto &e = as.get(arg_name);
+            const auto &e = arg.second;
             // Only common policy is supported in the library at this point
             int64_t count = 1;
             int mask = attr_t::get_default_mask(e.policy);
@@ -936,12 +936,12 @@ dnnl_primitive_attr_t create_dnnl_attr(
     }
 
     if (!attr.zero_points.is_def()) {
-        for (const auto &arg : supported_args) {
+        const auto &zp = attr.zero_points;
+        for (const auto &arg : zp.points) {
             const auto arg_name = arg.first;
-            const auto &zp = attr.zero_points;
             if (zp.is_def(arg_name)) continue;
 
-            const auto &e = zp.get(arg_name);
+            const auto &e = arg.second;
             // Only common policy/single RT value are supported in the library
             // at this point
             int64_t count = 1;
@@ -997,9 +997,7 @@ dnnl_primitive_attr_t create_dnnl_attr(
             }
         }
         DNN_SAFE_V(dnnl_primitive_attr_set_post_ops(dnnl_attr, ops));
-
-        const_dnnl_post_ops_t c_ops;
-        DNN_SAFE_V(dnnl_primitive_attr_get_post_ops(dnnl_attr, &c_ops));
+        auto c_ops = query_post_ops(dnnl_attr);
         SAFE_V(dnnl_post_ops_len(c_ops) == po.len() ? OK : FAIL);
 
         DNN_SAFE_V(dnnl_post_ops_destroy(ops));
@@ -1008,6 +1006,9 @@ dnnl_primitive_attr_t create_dnnl_attr(
     DNN_SAFE_V(dnnl_primitive_attr_set_scratchpad_mode(
             dnnl_attr, attr.scratchpad_mode));
 
+    DNN_SAFE_V(
+            dnnl_primitive_attr_set_fpmath_mode(dnnl_attr, attr.fpmath_mode));
+
     return dnnl_attr;
 }
 
@@ -1078,10 +1079,9 @@ int check_abc_tag(const std::string &tag_, bool check_enum_tags_only) {
         // Check rules related to lowercase/uppercase/block order.
         int dim_idx = std::tolower(c) - 'a';
         dim_state_t prev_state = dim_states[dim_idx];
-        dim_state_t cur_state = is_upper
-                ? dim_state_t::upper
-                : block != 0 ? dim_state_t::lower_with_block
-                             : dim_state_t::lower;
+        dim_state_t cur_state = is_upper ? dim_state_t::upper
+                : block != 0             ? dim_state_t::lower_with_block
+                                         : dim_state_t::lower;
 
         switch (cur_state) {
             case dim_state_t::upper:
@@ -1326,6 +1326,7 @@ float compute_eltwise_fwd(
         case pk_t::LINEAR: return scale * linear_fwd(src, alpha, beta);
         case pk_t::BRELU: return scale * bounded_relu_fwd(src, alpha);
         case pk_t::SRELU: return scale * soft_relu_fwd(src);
+        case pk_t::SRELU_V2: return scale * soft_relu_v2_fwd(src, alpha);
         case pk_t::LOGSIGMOID: return scale * logsigmoid_fwd(src);
         case pk_t::MISH: return scale * mish_fwd(src);
         case pk_t::LOGISTIC: return scale * logistic_fwd(src);
@@ -1339,6 +1340,8 @@ float compute_eltwise_fwd(
         case pk_t::GELU_ERF: return scale * gelu_erf_fwd(src);
         case pk_t::ROUND: return scale * round_fwd(src);
         case pk_t::HARDSWISH: return scale * hardswish_fwd(src);
+        case pk_t::HARDSIGMOID:
+            return scale * hardsigmoid_fwd(src, alpha, beta);
         case pk_t::RELU_DST: return scale * relu_fwd(src, alpha);
         case pk_t::TANH_DST: return scale * tanh_fwd(src);
         case pk_t::ELU_DST: return scale * elu_fwd(src, alpha);
@@ -1366,6 +1369,7 @@ float compute_eltwise_bwd(
         case pk_t::LINEAR: return linear_bwd(d_dst, src, alpha, beta);
         case pk_t::BRELU: return bounded_relu_bwd(d_dst, src, alpha);
         case pk_t::SRELU: return soft_relu_bwd(d_dst, src);
+        case pk_t::SRELU_V2: return soft_relu_v2_bwd(d_dst, src, alpha);
         case pk_t::LOGSIGMOID: return logsigmoid_bwd(d_dst, src);
         case pk_t::MISH: return mish_bwd(d_dst, src);
         case pk_t::LOGISTIC: return logistic_bwd(d_dst, src);
@@ -1378,6 +1382,7 @@ float compute_eltwise_bwd(
         case pk_t::POW: return pow_bwd(d_dst, src, alpha, beta);
         case pk_t::GELU_ERF: return gelu_erf_bwd(d_dst, src);
         case pk_t::HARDSWISH: return hardswish_bwd(d_dst, src);
+        case pk_t::HARDSIGMOID: return hardsigmoid_bwd(d_dst, src, alpha, beta);
 
         case pk_t::RELU_DST: return relu_bwd_use_dst(d_dst, src, alpha);
         case pk_t::TANH_DST: return tanh_bwd_use_dst(d_dst, src);
@@ -1465,6 +1470,7 @@ void update_cpu_ref_attrs(attr_t &attr, dnnl_data_type_t new_dt) {
         if (!e.is_binary_kind()) continue;
 
         e.binary.src1_dt = new_dt;
+        e.binary.tag = tag::abx; // Hardcoded in setup_binary_po as well.
     }
 }
 
