@@ -1,5 +1,5 @@
 /*******************************************************************************
-* Copyright 2017-2022 Intel Corporation
+* Copyright 2017-2023 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
@@ -72,21 +72,30 @@ std::ostream &operator<<(std::ostream &s, const std::vector<T> &v) {
     return s;
 }
 
-typedef int data_kind_t;
-enum {
+enum data_kind_t {
     SRC = 0,
     WEI,
     BIA,
     DST,
     ACC,
-    DATA,
+    // bnorm, lnorm
+    SRC_1,
     MEAN,
     VAR,
     SS,
     SC,
     SH,
-    GWEI,
-    DAT_TOTAL
+    // rnn
+    DST_ITER,
+    DST_ITER_C,
+    AUGRU_ATTENTION,
+    SRC_ITER,
+    SRC_ITER_C,
+    WEI_ITER,
+    WEI_PEEPHOLE,
+    WEI_PROJECTION,
+
+    DAT_TOTAL,
 };
 const char *data_kind2str(data_kind_t kind);
 
@@ -216,6 +225,7 @@ struct attr_t {
             EXP_DST,
             GELU_ERF,
             GELU_TANH,
+            HARDSIGMOID,
             HARDSWISH,
             LINEAR,
             LOG,
@@ -231,6 +241,7 @@ struct attr_t {
             SQRT_DST,
             SQUARE,
             SRELU,
+            SRELU_V2,
             SWISH,
             TANH,
             TANH_DST,
@@ -327,27 +338,34 @@ struct attr_t {
 
         std::vector<std::pair<int, int>> get_po_masks() const;
 
-        // TODO: remove
-        std::vector<int> get_binary_po_masks() const;
-
         std::vector<entry_t> entry;
     };
 
-    attr_t() : scratchpad_mode(dnnl_scratchpad_mode_library) {}
+    attr_t()
+        : scratchpad_mode(dnnl_scratchpad_mode_library)
+        , fpmath_mode(dnnl_fpmath_mode_strict) {}
+
+    template <typename First, typename... Rest>
+    void insert(const First &first, const Rest &...rest) {
+        this->insert(first);
+        if (sizeof...(rest) > 0) this->insert(rest...);
+    }
 
     void insert(const scale_t &s) { this->oscale = s; }
     void insert(const arg_scales_t &as) { this->scales = as; }
     void insert(const zero_points_t &zp) { this->zero_points = zp; }
     void insert(const post_ops_t &po) { this->post_ops = po; }
     void insert(dnnl_scratchpad_mode_t sm) { this->scratchpad_mode = sm; }
+    void insert(dnnl_fpmath_mode_t fpm) { this->fpmath_mode = fpm; }
 
     scale_t oscale;
     arg_scales_t scales;
     zero_points_t zero_points;
     post_ops_t post_ops;
     dnnl_scratchpad_mode_t scratchpad_mode;
+    dnnl_fpmath_mode_t fpmath_mode;
 
-    bool is_def() const;
+    bool is_def(bool skip_fpmath = false) const;
 };
 
 struct isa_hints_t {
@@ -391,9 +409,6 @@ struct isa_hints_t {
 
 using policy_t = attr_t::policy_t;
 
-void handle_legacy_attr(attr_t &attr, const attr_t &legacy_attr);
-
-int str2attr(attr_t *attr, const char *str);
 std::ostream &operator<<(std::ostream &s, const policy_t &policy);
 std::ostream &operator<<(std::ostream &s, const attr_t::scale_t &scale);
 std::ostream &operator<<(
@@ -402,6 +417,7 @@ std::ostream &operator<<(std::ostream &s, const attr_t::arg_scales_t &scales);
 std::ostream &operator<<(std::ostream &s, const attr_t::post_ops_t::kind_t &k);
 std::ostream &operator<<(std::ostream &s, const attr_t::post_ops_t &post_ops);
 std::ostream &operator<<(std::ostream &s, dnnl_scratchpad_mode_t sm);
+std::ostream &operator<<(std::ostream &s, dnnl_fpmath_mode_t fm);
 std::ostream &operator<<(std::ostream &s, const attr_t &attr);
 
 // A container for additional data and info, not available from user's input at
@@ -504,6 +520,7 @@ dnnl_primitive_attr_t create_dnnl_attr(
 
 dnnl_engine_kind_t str2engine_kind(const char *str);
 dnnl_scratchpad_mode_t str2scratchpad_mode(const char *str);
+dnnl_fpmath_mode_t str2fpmath_mode(const char *str);
 
 void maybe_oscale(
         const attr_t &attr, float &d, const float *scales, int64_t oc);
