@@ -14,6 +14,8 @@
 * limitations under the License.
 *******************************************************************************/
 
+#include <cctype>
+
 #include "utils/parser.hpp"
 
 #include "dnnl_common.hpp"
@@ -74,9 +76,15 @@ bool parse_multi_dt(std::vector<std::vector<dnnl_data_type_t>> &dt,
         const std::vector<std::vector<dnnl_data_type_t>> &def_dt,
         const char *str, const std::string &option_name /* = "sdt"*/) {
     static const std::string help
-            = "DT0:DT1[:DTi]    (Default: `f32` for all)\n    Specifies data "
-              "type `DTi` for source i.\n    `DT` values can be `f32`, `bf16`, "
-              "`f16`, `s32`, `s8`, `u8`.\n";
+            = "DT0:DT1[:DTi]    (Default: `f32` for all)\n    When the driver "
+              "supports the notion of multiple sources, the option specifies a "
+              "data type `DTi` for a source `i`.\n    When the driver supports "
+              "the notion of source, weights (optional), and destination, the "
+              "option specifies data types for source, weights (optional) and "
+              "destination correspondently.\n    The option may support "
+              "broadcast semantics (check the driver online documentation), "
+              "when a single value will be used for all inputs.\n    `DT` "
+              "values can be `f32`, `bf16`, `f16`, `s32`, `s8`, `u8`.\n";
     return parse_multivector_option(dt, def_dt, str2dt, str, option_name, help);
 }
 
@@ -136,24 +144,6 @@ bool parse_mb(std::vector<int64_t> &mb, const std::vector<int64_t> &def_mb,
     return parse_vector_option(mb, def_mb, atoi, str, option_name, help);
 }
 
-bool parse_attr(attr_t &attr, const char *str,
-        const std::string &option_name /* = "attr"*/) {
-    const std::string pattern = parser_utils::get_pattern(option_name);
-    if (pattern.find(str, 0, pattern.size()) == eol) return false;
-    static bool notice_printed = false;
-    if (!notice_printed) {
-        BENCHDNN_PRINT(0, "%s\n",
-                "WARNING (DEPRECATION NOTICE): `--attr` option is deprecated. "
-                "Please use one of `--attr-oscale`, `--attr-post-ops`, "
-                "`--attr-scales` or `--attr-zero-points` to specify attributes "
-                "specific part for a problem. New options support mixing and "
-                "will iterate over all possible combinations.");
-        notice_printed = true;
-    }
-    SAFE(str2attr(&attr, str + pattern.size()), CRIT);
-    return true;
-}
-
 bool parse_attr_oscale(std::vector<attr_t::scale_t> &oscale, const char *str,
         const std::string &option_name /* = "attr-oscale"*/) {
     static const std::string help
@@ -212,6 +202,16 @@ bool parse_attr_scratchpad_mode(
             str2scratchpad_mode, str, option_name, help);
 }
 
+bool parse_attr_fpmath_mode(std::vector<dnnl_fpmath_mode_t> &fpmath_mode,
+        const std::vector<dnnl_fpmath_mode_t> &def_fpmath_mode, const char *str,
+        const std::string &option_name /* = "attr-fpmath"*/) {
+    static const std::string help
+            = "MODE    (Default: `strict`)\n    Specifies fpmath_mode "
+              "attribute. `MODE` values can be `strict` or `bf16`.\n";
+    return parse_vector_option(fpmath_mode, def_fpmath_mode, str2fpmath_mode,
+            str, option_name, help);
+}
+
 bool parse_axis(std::vector<int> &axis, const std::vector<int> &def_axis,
         const char *str, const std::string &option_name /* = "axis"*/) {
     static const std::string help
@@ -359,15 +359,26 @@ bool parse_main_help(
 }
 
 // prb_dims_t type
-void parse_prb_vdims(prb_vdims_t &prb_vdims, const std::string &str) {
+void parse_prb_vdims(
+        prb_vdims_t &prb_vdims, const std::string &str, size_t min_inputs) {
+    assert(!str.empty());
+
     size_t start_pos = 0;
     // `n` is an indicator for a name supplied with dims_t object.
     std::string vdims_str = get_substr(str, start_pos, 'n');
+    // Sanity check that dims start with a digit.
+    if (!std::isdigit(vdims_str[0])) {
+        BENCHDNN_PRINT(0, "%s\n%s \'%s\'\n",
+                "ERROR: dims are expected to start with an integer value.",
+                "Given input:", str.c_str());
+        exit(1);
+    }
+
     parse_multivector_str(
             prb_vdims.vdims, {dims_t()}, atoi, vdims_str, ':', 'x');
 
     // Expect at least two inputs provided
-    SAFE_V(prb_vdims.vdims.size() > 1 ? OK : FAIL);
+    SAFE_V(prb_vdims.vdims.size() >= min_inputs ? OK : FAIL);
 
     prb_vdims.ndims = static_cast<int>(prb_vdims.vdims[0].size());
     // If second and consecutive inputs are provided with less dimensions
@@ -381,10 +392,14 @@ void parse_prb_vdims(prb_vdims_t &prb_vdims, const std::string &str) {
             prb_vdims.vdims[i].resize(prb_vdims.ndims, 1);
 
     prb_vdims.dst_dims = prb_vdims.vdims[0];
-    for (int i = 1; i < prb_vdims.n_inputs(); i++)
-        for (int d = 0; d < prb_vdims.ndims; ++d)
-            prb_vdims.dst_dims[d]
-                    = std::max(prb_vdims.dst_dims[d], prb_vdims.vdims[i][d]);
+    for_(int i = 1; i < prb_vdims.n_inputs(); i++)
+    for (int d = 0; d < prb_vdims.ndims; ++d) {
+        bool has_zero_dim
+                = prb_vdims.dst_dims[d] == 0 || prb_vdims.vdims[i][d] == 0;
+        prb_vdims.dst_dims[d] = has_zero_dim
+                ? 0
+                : std::max(prb_vdims.dst_dims[d], prb_vdims.vdims[i][d]);
+    }
 
     if (start_pos != eol) prb_vdims.name = str.substr(start_pos);
 }
@@ -399,7 +414,7 @@ void parse_prb_dims(prb_dims_t &prb_dims, const std::string &str) {
         try {
             value = std::stoll(s_);
         } catch (const std::invalid_argument &) {
-            BENCHDNN_PRINT(0, "%s\n%s \'%s\';\n",
+            BENCHDNN_PRINT(0, "%s\n%s \'%s\'\n",
                     "Error: dims value is expected to be an integer value.",
                     "Given input:", s_.c_str());
             exit(1);
@@ -503,7 +518,7 @@ static bool parse_fast_ref_gpu(
         fprintf(stderr,
                 "%s driver: WARNING: option `fast_ref_gpu` is not supported "
                 "for GPU only configurations.\n",
-                driver_name);
+                driver_name.c_str());
     }
 #endif
     return parsed;
@@ -580,11 +595,25 @@ static bool parse_mode(
     const auto str2bench_mode = [](const std::string &_str) {
         bench_mode_t mode;
         for (size_t i = 0; i < _str.size(); i++) {
-            if (_str[i] == 'r' || _str[i] == 'R') mode |= RUN;
-            if (_str[i] == 'c' || _str[i] == 'C') mode |= CORR;
-            if (_str[i] == 'p' || _str[i] == 'P') mode |= PERF;
-            if (_str[i] == 'l' || _str[i] == 'L') mode |= LIST;
-            if (_str[i] == 'o' || _str[i] == 'O') mode |= PROF;
+            switch (_str[i]) {
+                case 'r':
+                case 'R': mode |= RUN; break;
+                case 'c':
+                case 'C': mode |= CORR; break;
+                case 'p':
+                case 'P': mode |= PERF; break;
+                case 'l':
+                case 'L': mode |= LIST; break;
+                case 'o':
+                case 'O': mode |= PROF; break;
+                default: {
+                    fprintf(stderr,
+                            "ERROR: Unsupported character for `--mode` "
+                            "option: `%c`.\n",
+                            _str[i]);
+                    exit(2);
+                }
+            }
         }
         if (!(mode & LIST).none() && mode.count() > 1) {
             fprintf(stderr,
@@ -608,9 +637,9 @@ static bool parse_skip_impl(
         const char *str, const std::string &option_name = "skip-impl") {
     static const std::string help
             = "STRING    (Default: not specified)\n    Instructs the driver to "
-              "skip implementations which names matching `STRING`.\n    "
-              "`STRING` is a string literal with no spaces.\n    When empty, "
-              "option has no effect.\n";
+              "iterate over implementations when fetched implementation name "
+              "matching `STRING`.\n    `STRING` is a string literal with no "
+              "spaces.\n    When empty, option has no effect.\n";
     const auto chars2chars = [](const char *str) { return str; };
     bool parsed = parse_single_value_option(
             skip_impl, std::string(), chars2chars, str, option_name, help);
@@ -681,13 +710,13 @@ bool parse_bench_settings(const char *str) {
 
     // Last condition makes this help message to be triggered once driver_name
     // is already known.
-    if (!parsed && !end_msg && !std::string(driver_name).empty()) {
+    if (!parsed && !end_msg && !driver_name.empty()) {
         help_ss << "===================\n";
         help_ss << "= Driver options: =\n";
         help_ss << "===================\n";
         help_ss << "(More technical details available at "
                    "https://github.com/oneapi-src/oneDNN/blob/master/tests/"
-                   "benchdnn/doc/knobs_"
+                   "benchdnn/doc/driver_"
                 << driver_name << ".md)\n\n";
         end_msg = true;
     }
@@ -698,10 +727,19 @@ bool parse_bench_settings(const char *str) {
 void catch_unknown_options(const char *str) {
     last_parsed_is_problem = true; // if reached, means problem parsing
 
-    const std::string pattern = "--";
+    std::string pattern = "--";
     if (pattern.find(str, 0, pattern.size()) != eol) {
-        fprintf(stderr, "%s driver: ERROR: unknown option: `%s`, exiting...\n",
-                driver_name, str);
+        BENCHDNN_PRINT(0, "%s %s \'%s\'\n",
+                "driver: ERROR: unknown option:", driver_name.c_str(), str);
+        exit(2);
+    }
+
+    // Must stay after `--` check.
+    pattern = "-";
+    if (pattern.find(str, 0, pattern.size()) != eol) {
+        BENCHDNN_PRINT(0, "%s\n%s \'%s\'\n",
+                "ERROR: options should be passed with `--` prefix.",
+                "Given input:", str);
         exit(2);
     }
 }
@@ -710,7 +748,7 @@ int parse_last_argument() {
     if (!last_parsed_is_problem)
         fprintf(stderr,
                 "%s driver: WARNING: No problem found for a given option!\n",
-                driver_name);
+                driver_name.c_str());
     return OK;
 }
 
